/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib;

import org.globus.security.authorization.PDPInterceptor;
import org.globus.security.authorization.Decision;
import org.globus.security.authorization.RequestEntities;
import org.globus.security.authorization.NonRequestEntities;
import org.globus.security.authorization.AuthorizationException;
import org.globus.security.authorization.ChainConfig;
import org.globus.security.authorization.InitializeException;
import org.globus.security.authorization.CloseException;
import org.globus.security.authorization.EntityAttributes;
import org.globus.security.authorization.Attribute;
import org.globus.security.authorization.AttributeException;
import org.globus.gridshib.gt.authorization.PDP;
import org.globus.gridshib.gt.authorization.VomsConstants;
import org.globus.gridshib.gt.authorization.GridShibConfigParams;
import org.globus.gridshib.gt.authorization.PDPDecision;
import org.globus.wsrf.impl.security.util.AuthzUtil;
import org.globus.wsrf.impl.security.util.AttributeUtil;
import org.globus.wsrf.config.ConfigException;
import org.globus.wsrf.security.SecureContainerConfig;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.security.auth.Subject;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.namespace.QName;
import java.util.HashMap;
import java.util.Iterator;

public abstract class BasePDP implements PDPInterceptor {

    static Log logger = LogFactory.getLog(BasePDP.class.getName());

    protected abstract PDP pdpImpl();

    public void initialize(String chainName,
                           String prefix,
                           ChainConfig config) throws InitializeException {

        assert config != null;

        logger.info("PDP initializing for authz chain " + chainName +
                    " (scope: " + prefix + ")");

        try {
            AuthzUtil.parseNameValueParam(prefix, config);
        } catch (ConfigException e) {
            throw new InitializeException("problem parsing configuration",e);
        }

        HashMap configs = new HashMap();
        String[] keys = GridShibConfigParams.ALL_CONFIG_KEYS;
        for (int i = 0; i < keys.length; i++) {
            Object o = config.getProperty(prefix, keys[i]);
            if (o != null) {
                logger.debug("Config parameter " + keys[i] + ": " +
                             o.toString());
                configs.put(keys[i], o);
            }
        }

        // see note in org.globus.gridshib.gt.authorization.VomsConstants
        keys = VomsConstants.ALL_CONFIG_KEYS;
        for (int i = 0; i < keys.length; i++) {
            Object o = config.getProperty(prefix, keys[i]);
            if (o != null) {
                logger.debug("Config parameter " + keys[i] + ": " +
                             o.toString());
                configs.put(keys[i], o);
            }
        }

        try {
            pdpImpl().initialize(configs, chainName);
        } catch (Exception e) {
            throw new InitializeException("",e);
        }
    }

    public Decision canAccess(RequestEntities requestEntities,
                              NonRequestEntities nonRequestEntities)
            throws AuthorizationException {

        MessageContext ctx =  AttributeUtil.getMessageContext(requestEntities,
                                                                      null);

        if (ctx == null) {
            // for now we assume we are running in container
            throw new AttributeException("null message context");
        }

        return isPermitted(requestEntities, ctx);
    }

    public Decision canAdminister(RequestEntities requestEntities,
                                  NonRequestEntities nonRequestEntities)
            throws AuthorizationException {
        return null;
    }

    public void close() throws CloseException {
    }

    private Decision isPermitted(RequestEntities requestAttrs,
                                 MessageContext ctx)
            throws AuthorizationException {

        /* container is the issuer of this decision */
        EntityAttributes issuerEntity =
            SecureContainerConfig.getSecurityDescriptor().getContainerEntity();

        /**
         * requestAttributes contains Requestor, Environment, Action,
         * and Resource bags of attributes.  The Decision object needs
         * Requestor.
         */
        EntityAttributes reqEntity = requestAttrs.getRequestor();

        /* requester information for impl */
        Subject peer = AttributeUtil.getPeerSubject(reqEntity);

        // not used in new base PDP api
        //String peerIdentity = CredentialUtil.getIdentity(peer);

        /* operation name for impl */
        EntityAttributes actionEntity = requestAttrs.getAction();
        Attribute opAttr = AttributeUtil
                .getAttribute(actionEntity.getIdentityAttributes(),
                          AttributeUtil.getOperationAttrIdentifier());
        Iterator it = opAttr.getAttributeValueSet().iterator();
        QName qname = null;
        if (it.hasNext()) {
            qname = (QName)it.next();
        }

        int dec = PDPDecision.INDETERMINATE;
        try {
            dec = this.pdpImpl().isPermitted(peer, ctx, qname);
        } catch (IllegalArgumentException e) {
            throw new AuthorizationException("Illegal argument", e);
        } catch (Exception e) {
            logger.error(e);
        }

        switch (dec) {
            case PDPDecision.PERMIT:
                return new Decision(issuerEntity, reqEntity,
                            Decision.PERMIT, null, null);
            case PDPDecision.DENY:
                return new Decision(issuerEntity, reqEntity,
                            Decision.DENY, null, null);
            case PDPDecision.NOT_APPLICABLE:
                return new Decision(issuerEntity, reqEntity,
                            Decision.NOT_APPLICABLE, null, null);
            default:
                return new Decision(issuerEntity, reqEntity,
                            Decision.INDETERMINATE, null, null);
        }
    }
}
