/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Vector;

import javax.security.auth.Subject;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.common.mapper.GridShibEntityMapper;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.SAMLPrincipal;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;
import org.globus.gridshib.security.BasicAttribute;
import org.globus.gridshib.security.SAMLAuthnContext;
import org.globus.gridshib.security.SAMLIdentity;

import org.globus.opensaml11.saml.SAMLNameIdentifier;

/**
 * @see org.globus.gridshib.AttributeAcceptancePIP
 */
public class AttributeAcceptancePIPImpl extends BasePIP
                                     implements GridShibConfigParams {

    static Log logger =
        LogFactory.getLog(AttributeAcceptancePIPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "AttributeAcceptancePIPImpl initialization complete";

    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("AttributeAcceptancePIPImpl initializing in scope " +
                    name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("AttributeAcceptancePIPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
    }

    public void collectAttributes(Subject peerSubject,
                                  MessageContext context) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("AttributeAcceptancePIPImpl collecting attributes " +
                    "for identity " + peerIdentity);

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);

        /* In the following issuer checks, an entity is trusted
         * if the service is configured with metadata for that
         * entity.  In a later version of GridShib for GT, we
         * will implement true Attribute Acceptance Policy,
         * which is independent of metadata.
         */

        // check issuer of identity(ies):
        SAMLIdentity[] identities = secCtx.getSAMLIdentities();
        assert (identities != null);
        logger.debug("Found " + identities.length + " identit" +
                     ((identities.length == 1) ? "y" : "ies"));
        for (int i = 0; i < identities.length; i++) {
            if (identities[i].isTrusted()) {
                logger.debug("SAML identity is already trusted: " +
                             identities[i].toString());
            } else {
                String entityID = identities[i].getIssuer();
                if (GridShibEntityMapper.hasMetadata(entityID)) {
                    logger.debug("Trusted SAML authority: " + entityID);
                    //secCtx.getSAMLIdentities()[i].setTrusted(true);
                    identities[i].setTrusted(true);
                    logger.debug("Trusted SAML identity: " +
                                 identities[i].toString());
                    // TODO: push the following code into SAMLIdentity class
                    SAMLNameIdentifier nameid = identities[i].getNameID();
                    String name = nameid.getName();
                    if (nameid.getNameQualifier() == null) {
                        if (secCtx.addPrincipal(new SAMLPrincipal(name))) {
                            logger.debug("New principal added: " + name);
                        } else {
                            logger.warn("Unable to add new principal: " + name);
                        }
                    } else {
                        String msg = "Principal (" + name + ") not added " +
                                     "because the name is qualified: " +
                                     nameid.getNameQualifier();
                        logger.debug(msg);
                    }
                } else {
                    logger.info("Untrusted SAML authority: " + entityID);
                    logger.info("Untrusted SAML identity: " +
                                identities[i].toString());
                    logger.info("Removing untrusted identity");
                    secCtx.removeSAMLIdentity(identities[i]);
                }
            }
        }

        // check issuer of authn context(s):
        SAMLAuthnContext[] authnContexts = secCtx.getSAMLAuthnContexts();
        assert (authnContexts != null);
        logger.debug("Found " + authnContexts.length + " authn context" +
                     ((authnContexts.length == 1) ? "" : "s"));
        for (int i = 0; i < authnContexts.length; i++) {
            if (authnContexts[i].isTrusted()) {
                logger.debug("SAML authnContext is already trusted: " +
                             authnContexts[i].toString());
            } else {
                String entityID = authnContexts[i].getIssuer();
                if (GridShibEntityMapper.hasMetadata(entityID)) {
                    logger.debug("Trusted SAML authority: " + entityID);
                    //secCtx.getSAMLAuthnContexts()[i].setTrusted(true);
                    authnContexts[i].setTrusted(true);
                    logger.debug("Trusted SAML authnContext: " +
                                 authnContexts[i].toString());
                } else {
                    logger.info("Untrusted SAML authority: " + entityID);
                    logger.info("Untrusted SAML authnContext: " +
                                authnContexts[i].toString());
                    logger.info("Removing untrusted authnContext");
                    secCtx.removeSAMLAuthnContext(authnContexts[i]);
                }
            }
        }

        // check issuer of attributes(s):
        BasicAttribute[] attributes = secCtx.getAttributes();
        assert (attributes != null);
        logger.debug("Found " + attributes.length + " attribute" +
                     ((attributes.length == 1) ? "" : "s"));
        for (int i = 0; i < attributes.length; i++) {
            if (attributes[i].isTrusted()) {
                logger.debug("SAML attribute is already trusted: " +
                             attributes[i].toString());
            } else {
                String entityID = attributes[i].getIssuer();
                if (GridShibEntityMapper.hasMetadata(entityID)) {
                    logger.debug("Trusted SAML authority: " + entityID);
                    //secCtx.getAttributes()[i].setTrusted(true);
                    attributes[i].setTrusted(true);
                    logger.debug("Trusted SAML attribute: " +
                                 attributes[i].toString());
                } else {
                    logger.info("Untrusted SAML authority: " + entityID);
                    logger.info("Untrusted SAML attribute: " +
                                attributes[i].toString());
                    logger.info("Removing untrusted attribute");
                    secCtx.removeAttribute(attributes[i]);
                }
            }
        }

        // check issuer of legacy attribute information:
        Vector attrs =
            new Vector(Arrays.asList(secCtx.getLegacyAttributes()));
        logger.debug("Found " + attrs.size() + " attribute" +
                     ((attrs.size() == 1) ? "" : "s"));
        for (int i = 0; i < attrs.size(); i++) {
            SAMLAttribute attr = (SAMLAttribute)(attrs.elementAt(i));
            logger.debug("Checking attribute " + attr.getName());
            String entityID = attr.getIssuer().getAAProviderId();
            if (GridShibEntityMapper.hasMetadata(entityID)) {
                logger.debug("Trusted SAML authority: " + entityID);
                logger.debug("Trusted legacy attribute " + attr.getName());
            } else {
                logger.info("Untrusted SAML authority: " + entityID);
                logger.info("Untrusted legacy attribute " + attr.getName());
                logger.info("Removing untrusted legacy attribute");
                secCtx.removeLegacyAttribute(attr);
            }
        }

        logger.info("AttributeAcceptancePIPImpl attribute collection " +
                    "complete");
    }
}
