/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import java.io.File;
import java.util.HashMap;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.attributes.policy.AttributeMapping;
import org.globus.gridshib.gt.authorization.attributes.policy.SimpleSAMLAttributePolicy;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;

/**
 * A PDP implementation that backs <code>SAMLAttributePDP</code>.
 *
 * @see org.globus.gridshib.SAMLAttributePDP
 */
public class SAMLAttributePDPImpl extends BasePDP
                               implements GridShibConfigParams {

    static Log logger =
        LogFactory.getLog(SAMLAttributePDPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "SAMLAttributePDPImpl initialization complete";
    private static final String COMPLETION_MSG =
        "SAMLAttributePDPImpl decision making complete";

    private boolean isAuthzEnabled;
    private ShibbolethPDPPolicy authzPolicy = null;
    private File policyFile = null;
    private long policyFileLastLoaded = 0;

    boolean isAttributeAuthzEnabled() {
        return this.isAuthzEnabled;
    }

    /**
     * Loads a policy file that is used for access control
     * decisions.  If no policy file is configured, this
     * method disables attribute-based authorization.
     *
     * @param configs a (non-null) map of configuration
     *        parameters
     * @param name the scope prefix associated with this
     *        PDP implementation
     *
     * @exception java.lang.IllegalArgumentException
     *            If the <code>configs</code> argument is null
     * @exception java.lang.RuntimeException
     *            If no policy file is configured, the path to
     *            the configured policy file is not absolute, or
     *            the policy file can not be parsed
     */
    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("SAMLAttributePDPImpl initializing in scope " + name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("SAMLAttributePDPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        /* This block implements the following logic:
         *
         * 1. Check for an object-based authz policy instance
         *    (SHIB_PDP_POLICY_KEY).  If one exists, persist
         *    the instance and skip to step 3; otherwise
         *    continue with the next step.
         *
         * 2. Check for an authz policy file (ATTR_FILE_KEY).
         *    If one exists, parse the file, instantiate
         *    an instance, persist the instance, and continue
         *    with the next step.
         *
         * 3. If a policy instance was found, set a config
         *    parameter (ENABLE_ATTRIBUTE_AUTHZ_KEY to true;
         *    otherwise, set it to false.
         */
        Object authzPolicyObj = configs.get(SHIB_PDP_POLICY_KEY);
        if (authzPolicyObj != null) {
            logger.debug("Found GridShib policy instance (" +
                         SHIB_PDP_POLICY_KEY + ")");
            try {
                this.authzPolicy = (ShibbolethPDPPolicy)authzPolicyObj;
            } catch (ClassCastException e) {
                String msg = "GridShib policy instance is not " +
                             "the correct type (programmer error)";
                logger.error(msg, e);
                throw new RuntimeException(msg, e);
            }
            logger.debug("Using object-based authz policy: " +
                         this.authzPolicy.toString());
        } else {
            // expecting policy file:
            String policyPath = (String)configs.get(ATTR_FILE_KEY);
            if (policyPath == null) {
                String msg = "No path to GridShib policy file found";
                logger.debug(msg);
            } else if (policyPath.equals("")) {
                String msg = "Path to GridShib policy file is the " +
                             "empty string, please reconfigure " +
                             "parameter " + ATTR_FILE_KEY;
                logger.error(msg);
                throw new RuntimeException(msg);
            } else {
                logger.debug("Found path to GridShib policy file (" +
                             policyPath + ")");

                this.policyFile = new File(policyPath);
                assert (this.policyFile != null);
                if (!this.policyFile.isAbsolute()) {
                    String msg = "Path to GridShib policy file is not " +
                                 "absolute (" + policyPath + ")";
                    logger.error(msg);
                    throw new RuntimeException(msg);
                }

                // an array of attributes and an array of usernames:
                AttributeMapping[] attributeMapping;
                try {
                    this.policyFileLastLoaded = this.policyFile.lastModified();
                    attributeMapping =
                        SimpleSAMLAttributePolicy.parseAttrFile(this.policyFile);
                } catch (Exception e) {
                    String msg = "Unable to parse GridShib policy file " +
                                 "(" + policyPath + ")";
                    logger.error(msg, e);
                    throw new RuntimeException(msg, e);
                }
                logger.info("Using GridShib policy file (" + policyPath + ")");

                this.authzPolicy = new ShibbolethPDPPolicy(attributeMapping);
                logger.debug("Using file-based authz policy: " +
                             this.authzPolicy.toString());
            }
        }

        if (this.authzPolicy == null) {
            this.isAuthzEnabled = false;
            configs.put(ENABLE_ATTRIBUTE_AUTHZ_KEY, "false");
            logger.info("Attribute-based authz is implicitly disabled");
        } else {
            this.isAuthzEnabled = true;
            configs.put(ENABLE_ATTRIBUTE_AUTHZ_KEY, "true");
            logger.info("Attribute-based authz is implicitly enabled");
        }

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
    }

    /**
     * Note that this method <strong>never</strong> returns
     * <code>PDPDecision.DENY</code>.
     * <p>
     * A GridShib policy file is a dynamic resource in the
     * following sense.  Every time a policy decision is
     * required, this method determines if the policy file
     * was modified since it was last loaded.  If the file
     * is dirty, this method reloads the policy file before
     * making an access control decision.  If the policy
     * file can not be reloaded (because of a syntax error,
     * e.g.), this method logs an error message and returns
     * <code>PDPDecision.INDETERMINATE</code>.
     *
     * @param peerSubject the authenticated subject
     * @param context the message context
     * @param operation the requested operation
     *
     * @exception java.lang.IllegalArgumentException
     *            If either the <code>peerSubject</code> argument
     *            or the <code>context</code> argument is null
     */
    public int isPermitted(Subject peerSubject,
                           MessageContext context,
                           QName operation) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());
        logger.debug("operation = " + operation.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("SAMLAttributePDPImpl making decision " +
                    "for identity " + peerIdentity);

        // if attribute authorization is disabled, short-circuit this method:
        if (this.isAuthzEnabled) {
            logger.info("Attribute-based authorization is enabled");
        } else {
            logger.info("Attribute-based authorization is disabled");
            return PDPDecision.NOT_APPLICABLE;
        }
        assert (this.isAuthzEnabled);

        // reload policy file if necessary:
        try {
            reloadPolicyFile();
        } catch (Exception e) {
            String msg = "Unable to reload GridShib policy file, " +
                         "SAMLAttributePDPImpl returned INDETERMINATE " +
                         "decision";
            logger.error(msg, e);
            return PDPDecision.INDETERMINATE;
        }

        boolean attrPassed =
            ShibbolethPDPUtil.isAttributePermitted(peerSubject,
                                                   this.authzPolicy);

        /* http://bugzilla.globus.org/globus/show_bug.cgi?id=2287
         *
         * if 'return false' based on attribute, the exception returned to
         * client by ServiceAuthorizationChain claims DN is denied.
         *
         * One possible workaround, AuthorizationException, implies error in
         * policy processing, so something along these lines is actually not
         * a viable alternative:
         *
         * throw new AuthorizationException("primary attribute denied: " +
         *                                  primaryAttr);
         */

        int decisionCode;
        if (attrPassed) {
            logger.info("SAMLAttributePDPImpl returned PERMIT decision");
            decisionCode = PDPDecision.PERMIT;
        } else {
            logger.info("SAMLAttributePDPImpl returned DENY decision");
            decisionCode = PDPDecision.DENY;
        }
        logger.info(COMPLETION_MSG);

        return decisionCode;
    }

    /**
     * If policy is file-based, check to see if the policy
     * file has been modified since it was last loaded.  If
     * so, reload it.  If policy is object-based (or more
     * precisely, if no policy file is available), do nothing.
     *
     * @exception java.lang.SecurityException
     *            If unable to determine the "last modified"
     *            date on the policy file
     * @exception java.lang.Exception
     *            If unable to parse the policy file
     */
    private void reloadPolicyFile() throws Exception {

        assert this.authzPolicy != null;

        if (this.policyFile != null) {
            // this may throw java.lang.SecurityException:
            long lastModified = this.policyFile.lastModified();

            logger.debug("GridShib policy file (" +
                         this.policyFile.getPath() +
                         ") last loaded (" +
                         this.policyFileLastLoaded +
                         "), last modified (" + lastModified + ")");

            // if policy file has been modified, reload it:
            if (this.policyFileLastLoaded < lastModified) {
                logger.info("Reloading GridShib policy file (" +
                            this.policyFile.getPath() + ")");
                // this may throw java.lang.Exception:
                AttributeMapping[] attributeMapping =
                    SimpleSAMLAttributePolicy.parseAttrFile(this.policyFile);
                this.authzPolicy.replaceAttrs(attributeMapping);
                this.policyFileLastLoaded = lastModified;
            }
        }
    }
}
