/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import java.io.File;
import java.security.Principal;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Vector;
import javax.security.auth.Subject;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.attributes.policy.AttributeMapping;
import org.globus.gridshib.gt.authorization.attributes.policy.SimpleSAMLAttributePolicy;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.UserNamePrincipal;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;

/**
 * Reads a configuration file containing mappings from
 * attributes to usernames. Adds all mapped usernames
 * to the subject's security context.
 */
public class SAMLMapPIPImpl extends BasePIP
                         implements GridShibConfigParams {

    static Log logger = LogFactory.getLog(SAMLMapPIPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "SAMLMapPIPImpl initialization complete";
    private static final String COMPLETION_MSG =
        "SAMLMapPIPImpl attribute collection complete";

    private boolean isMappingEnabled;
    private ShibbolethPDPPolicy mapPolicy = null;
    private File policyFile = null;
    private long policyFileLastLoaded;

    private String[] cachedUsernames = null;
    private boolean cacheAttributes = false;

    boolean isAttributeMappingEnabled() {
        return this.isMappingEnabled;
    }

    /**
     * This method attempts to load
     * a map policy file (<code>MAP_FILE_KEY</code>).  If
     * no such file is configured, the method attempts to
     * load an authorization policy file
     * (<code>ATTR_FILE_KEY</code>) instead.  If neither
     * of these policy files are configured, this method
     * disables attribute-based mapping.
     *
     * @param configs a (non-null) map of configuration
     *        parameters
     * @param name the scope prefix associated with this
     *        PIP implementation
     *
     * @exception java.lang.IllegalArgumentException
     *            If the <code>configs</code> argument is null
     */
    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("SAMLMapPIPImpl initializing in scope " + name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("SAMLMapPIPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        /* This complicated block implements the following logic:
         *
         * 1. Check for an object-based map policy instance
         *    (SAML_MAP_POLICY_KEY).  If one exists, persist
         *    the instance and skip to step 5; otherwise
         *    continue with the next step.
         *
         * 2. Check for a map policy file (MAP_FILE_KEY).
         *    If one exists, parse the file, instantiate
         *    an instance, persist the instance, and skip to
         *    step 5; otherwise continue with the next step.
         *
         * 3. Check for an object-based authz policy instance
         *    (SHIB_PDP_POLICY_KEY).  If one exists, persist
         *    the instance and skip to step 5; otherwise
         *    continue with the next step.
         *
         * 4. Check for an authz policy file (ATTR_FILE_KEY).
         *    If one exists, parse the file, instantiate
         *    an instance, persist the instance, and continue
         *    with the next step.
         *
         * 5. If a policy instance was found, set a config
         *    parameter (ENABLE_ATTRIBUTE_MAPPING_KEY to true;
         *    otherwise, set it to false.
         */
        Object policyObj, stringObj;
        if ((policyObj = configs.get(SAML_MAP_POLICY_KEY)) != null) {
            logger.debug("Found GridShib policy instance (" +
                         SAML_MAP_POLICY_KEY + ")");

            try {
                this.mapPolicy = (ShibbolethPDPPolicy)policyObj;
            } catch (ClassCastException e) {
                String msg = "GridShib policy instance is not " +
                             "the correct type (programmer error)";
                logger.error(msg, e);
                throw new RuntimeException(msg, e);
            }
            logger.debug("Using object-based map policy: " +
                         this.mapPolicy.toString());
        } else if ((stringObj = configs.get(MAP_FILE_KEY)) != null) {

            String policyPath = (String)stringObj;
            if (policyPath.equals("")) {
                String msg = "Path to GridShib policy file is the " +
                             "empty string, please reconfigure " +
                             "parameter " + MAP_FILE_KEY;
                logger.error(msg);
                throw new RuntimeException(msg);
            } else {
                logger.debug("Found path to GridShib policy file (" +
                             policyPath + ")");

                this.policyFile = new File(policyPath);
                assert (this.policyFile != null);
                if (!this.policyFile.isAbsolute()) {
                    String msg = "Path to GridShib policy file is not " +
                                 "absolute (" + policyPath + ")";
                    logger.error(msg);
                    throw new RuntimeException(msg);
                }

                // an array of attributes and an array of usernames:
                AttributeMapping[] attributeMapping;
                try {
                    this.policyFileLastLoaded = this.policyFile.lastModified();
                    attributeMapping =
                        SimpleSAMLAttributePolicy.parseAttrFile(this.policyFile);
                } catch (Exception e) {
                    String msg = "Unable to parse GridShib policy file " +
                                 "(" + policyPath + ")";
                    logger.error(msg, e);
                    throw new RuntimeException(msg, e);
                }
                logger.info("Using GridShib policy file (" + policyPath + ")");

                this.mapPolicy = new ShibbolethPDPPolicy(attributeMapping);
                logger.debug("Using file-based map policy: " +
                             this.mapPolicy.toString());
            }
        } else if ((policyObj = configs.get(SHIB_PDP_POLICY_KEY)) != null) {
            // no map policy, fall back on authz policy:
            logger.debug("Found GridShib policy instance (" +
                         SHIB_PDP_POLICY_KEY + ")");

            try {
                this.mapPolicy = (ShibbolethPDPPolicy)policyObj;
            } catch (ClassCastException e) {
                String msg = "GridShib policy instance is not " +
                             "the correct type (programmer error)";
                logger.error(msg, e);
                throw new RuntimeException(msg, e);
            }
            logger.debug("Using object-based authz policy: " +
                         this.mapPolicy.toString());
        } else if ((stringObj = configs.get(ATTR_FILE_KEY)) != null) {

            String policyPath = (String)stringObj;
            if (policyPath.equals("")) {
                String msg = "Path to GridShib policy file is the " +
                             "empty string, please reconfigure " +
                             "parameter " + ATTR_FILE_KEY;
                logger.error(msg);
                throw new RuntimeException(msg);
            } else {
                logger.debug("Found path to GridShib policy file (" +
                             policyPath + ")");

                this.policyFile = new File(policyPath);
                assert (this.policyFile != null);
                if (!this.policyFile.isAbsolute()) {
                    String msg = "Path to GridShib policy file is not " +
                                 "absolute (" + policyPath + ")";
                    logger.error(msg);
                    throw new RuntimeException(msg);
                }

                // an array of attributes and an array of usernames:
                AttributeMapping[] attributeMapping;
                try {
                    this.policyFileLastLoaded = this.policyFile.lastModified();
                    attributeMapping =
                        SimpleSAMLAttributePolicy.parseAttrFile(this.policyFile);
                } catch (Exception e) {
                    String msg = "Unable to parse GridShib policy file " +
                                 "(" + policyPath + ")";
                    logger.error(msg, e);
                    throw new RuntimeException(msg, e);
                }
                logger.info("Using GridShib policy file (" + policyPath + ")");

                this.mapPolicy = new ShibbolethPDPPolicy(attributeMapping);
                logger.debug("Using file-based authz policy: " +
                             this.mapPolicy.toString());
                // an optimization:
                configs.put(SHIB_PDP_POLICY_KEY, this.mapPolicy);
            }
        }

        // set a config parameter:
        if (this.mapPolicy == null) {
            this.isMappingEnabled = false;
            configs.put(ENABLE_ATTRIBUTE_MAPPING_KEY, "false");
            logger.info("Attribute-based mapping is implicitly disabled");
        } else {
            this.isMappingEnabled = true;
            configs.put(ENABLE_ATTRIBUTE_MAPPING_KEY, "true");
            logger.info("Attribute-based mapping is implicitly enabled");
        }

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
        return;

        /* The following block is dead code due to a bug:
           http://bugzilla.globus.org/globus/show_bug.cgi?id=6016

        this.cacheAttributes = false;
        try {
            Object cacheflag = configs.get(CACHE_KEY);

            if (cacheflag == null) {
                logger.debug("no cache configuration, default is false");
            } else {
                String cachestr = (String) cacheflag;
                if (cachestr.trim().equalsIgnoreCase("true")) {
                    logger.debug("caching activated");
                    this.cacheAttributes = true;
                    // every authz chain scope has its own cache scope
                } else {
                    logger.debug("cache flag configured, but does " +
                            "not match 'true'");
                }
            }
        } catch (ClassCastException e) {
            logger.warn("caching configuration is present but it is " +
                    "not the proper class, programmer error");
        } catch (Exception e) {
            logger.warn(e.getMessage());
        }
        */
    }

    /**
     * TBD
     *
     * @param peerSubject the authenticated subject
     * @param context the message context
     *
     * @exception java.lang.IllegalArgumentException
     *            If either the <code>peerSubject</code> argument
     *            or the <code>context</code> argument is null
     */
    public void collectAttributes(Subject peerSubject,
                                  MessageContext context) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("SAMLMapPIPImpl collecting attributes " +
                    "for identity " + peerIdentity);

        // if attribute mapping is disabled, short-circuit this method:
        if (this.isMappingEnabled) {
            logger.info("Attribute-based username mapping is enabled");
        } else {
            logger.info("Attribute-based username mapping is disabled");
            return;
        }
        assert (this.isMappingEnabled);

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);

        // check if file has been updated, if not pick up from cache
        if (!isFileUpdated()) {
            if (this.cachedUsernames != null) {
                //addUsernames(peerSubject, this.cachedUsernames);
                secCtx.addPrincipals(toPrincipalArray(this.cachedUsernames));
            }
        }

        Vector peerAttributes =
            new Vector(Arrays.asList(secCtx.getLegacyAttributes()));

        if (peerAttributes == null) {
            logger.warn("cannot retrieve attribute info from Subject");
            return;
        }

        // update to latest code
        try {
            updateConfigFile();
        } catch (Exception exp) {
            logger.warn("Exception updating map file " + exp.getMessage());
        }

        // Get mapping info, mapPolicy cannot be null.
        AttributeMapping[] policyMappings =
                                this.mapPolicy.getAttrMappingsArray();

        String[] newUsernames =
            ShibbolethPDPUtil.getUsernameMappings(policyMappings,
                                                  peerAttributes.iterator());

        secCtx.addPrincipals(toPrincipalArray(newUsernames));

        if (cacheAttributes) {
            this.cachedUsernames = newUsernames;
        }

        logger.info(COMPLETION_MSG);
    }

    private Principal[] toPrincipalArray(String[] usernames) {

        if (usernames == null) return null;

        Principal[] principals = new UserNamePrincipal[usernames.length];
        for (int i = 0; i < usernames.length; i++) {
            principals[i] = new UserNamePrincipal(usernames[i]);
        }
        return principals;
    }

    private void updateConfigFile() throws Exception {

        if (isFileUpdated()) {

            AttributeMapping[] map = SimpleSAMLAttributePolicy
                .parseAttrFile(this.policyFile);
            this.mapPolicy = new ShibbolethPDPPolicy(map);

            this.policyFileLastLoaded = this.policyFile.lastModified();
        }
    }

    private boolean isFileUpdated() {

        if (this.policyFile == null) {
            return false;
        } else {
            return (this.policyFileLastLoaded <
                    this.policyFile.lastModified());
        }
    }
}
