/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.gridshib.gt.authorization.attributes.AttributeInformation;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.globus.gridshib.gt.authorization.attributes.policy.AttributeMapping;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.security.auth.Subject;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;
import java.util.Set;
import java.util.HashSet;


class TextFile extends ArrayList {
    //via Eckel
    TextFile(String fileName) throws IOException {
        super(Arrays.asList(ShibbolethPDPUtil.gRead(fileName).split("\n")));
    }
}

/**
 * Contains static methods for common Shib PDP tasks
 */
public class ShibbolethPDPUtil {

    private static Log logger =
        LogFactory.getLog(ShibbolethPDPUtil.class.getName());

    /**
     * Read file as Java properties file.
     * @param authzFile config file
     * @return java properties object
     * @throws IOException io
     */
    static Properties readConfigFileAsProps(File authzFile)
            throws IOException {

        if (authzFile == null) {
            throw new FileNotFoundException("authzFile null");
        }

        Properties properties = new Properties();
        if (authzFile.exists()) {
            properties.load(new FileInputStream(authzFile));
        } else {
            throw new FileNotFoundException("can not find: "
                                            + authzFile.getAbsolutePath());
        }
        return properties;
    }


    static String gRead(String fileName) throws IOException {
        StringBuffer sb = new StringBuffer();
        BufferedReader in =
        new BufferedReader(new FileReader(fileName));
        String s;
        while((s = in.readLine()) != null) {
            sb.append(s);
            sb.append("\n");
        }
        in.close();
        return sb.toString();
    }

    static String[] readConfigFile(File authzFile) throws IOException {
        TextFile file = new TextFile(authzFile.getAbsolutePath());
        String[] result = new String[file.size()];

        for (int i = 0; i < file.size(); i++) {

            result[i] = (String) file.get(i);
        }
        if (logger.isDebugEnabled()) {
            for (int i = 0; i < result.length; i++) {
                logger.debug("result[" + i + "] = " + result[i]);
            }
        }

        return result;
    }

    /**
     * Checks the configured attributes against the subject's
     * issued attributes (i.e., those attributes stored in the
     * subject's security context).
     *
     * @param peerSubject a subject that has a security context
     * @param policy configured attribute-based policy
     *
     * @return true if at least one attribute configured in policy
     *         is also in the subject's security context
     */
    public static boolean isAttributePermitted(Subject peerSubject,
                                               ShibbolethPDPPolicy policy) {

        Iterator ACL = policy.getAttrMappings();
        if (ACL == null || !ACL.hasNext()) {
            logger.debug("No configured attributes " +
                         "(and so policy is trivially satisfied)");
            return true;
        }

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);

        Vector peerAttributes =
            new Vector(Arrays.asList(secCtx.getLegacyAttributes()));
        if (peerAttributes.isEmpty()) {
            logger.debug("No issued attributes " +
                         "(and so policy can't possibly be satisfied)");
            return false;
        }
        logger.debug("Obtained " + peerAttributes.size() +
                     " issued attributes from security context");

        int i = 0;
        Vector samlAttributes;
        while (ACL.hasNext()) {
            logger.debug("Checking configured attribute set " + i);
            samlAttributes =
                ((AttributeMapping)ACL.next()).getSAMLAttributes();
            logger.debug("There are " + samlAttributes.size() +
                         " configured attributes in set " + i);
            boolean match = attributeSetsMatch(samlAttributes, peerAttributes);
            if (match) {
                logger.debug("Configured attribute set " + i + " matched");
                return true;
            }
            i++;
        }

        logger.debug("No configured attribute set matched");
        return false;
    }

    /**
     * Checks the configured attributes against the given attributes.
     *
     * @param attributes a provided array of attributes to check against
     * @param policy configured attribute-based policy
     *
     * @return true if at least one attribute configured in policy
     *         is also in the given array of attributes
     */
    public static boolean isAttributePermitted(SAMLAttribute[] attributes,
                                               ShibbolethPDPPolicy policy) {

        Iterator ACL = policy.getAttrMappings();
        if (ACL == null || !ACL.hasNext()) {
            logger.debug("No configured attributes " +
                         "(and so policy is trivially satisfied)");
            return true;
        }

        if ((attributes == null) || (attributes.length == 0)) {
            logger.debug("No issued attributes " +
                         "(and so policy can't possibly be satisfied)");
            return false;
        }
        logger.debug("Obtained " + attributes.length +
                     " issued attributes from caller");

        int i = 0;
        Vector samlAttributes;
        while (ACL.hasNext()) {
            logger.debug("Checking configured attribute set " + i);
            samlAttributes =
                ((AttributeMapping)ACL.next()).getSAMLAttributes();
            logger.debug("There are " + samlAttributes.size() +
                         " configured attributes in set " + i);
            boolean match = attributeSetsMatch(samlAttributes, attributes);
            if (match) {
                logger.debug("Configured attribute set " + i + " matched");
                return true;
            }
            i++;
        }

        logger.debug("No configured attribute set matched");
        return false;
    }


    /**
     * Checks for any previously obtained attribute information about
     * the subject. If re-implemented, should stay encapsulated beginning here.
     * @param peerIdentity caller
     *        This is intended to be the string generated from
     *        AuthUtil.getIdentity(peer), but can be any unique string.
     * @param scope intended to be the pdp chain scope, but could be
     *        used in other ways
     * @return If any problem or nothing found, returns null
     */
    public static AttributeInformation
        queryCachedAttributeInformation(String peerIdentity, String scope) {

        if (peerIdentity == null) {
            logger.debug("no peer, no cache");
            return null;
        } else if (peerIdentity.length() < 1) {
            logger.debug("zero length peer, no cache allowed");
            return null;
        }

        //use JNDI
        try {
            return queryJNDICachedAttributeInformation(peerIdentity,scope);
        } catch (Exception e) {
            logger.debug("Problem with attribute store: " + e.getMessage());
            return null;
        }

    }


    /**
     *
     * @param peerIdentity caller
     *        This is intended to be the string generated from
     *        AuthUtil.getIdentity(peer), but can be any unique string.
     * @param scope intended to be the pdp chain scope, but could be
     *        used in  other ways
     * @param info attrs
     * @return true/false if succeeded or not
     */
    public static boolean
        setCachedAttributeInformation(String peerIdentity, String scope,
                                      AttributeInformation info) {

        if (peerIdentity == null) {
            logger.debug("no peer, no cache");
            return false;
        } else if (peerIdentity.length() < 1) {
            logger.debug("zero length peer, no cache allowed");
            return false;
        }

        //use JNDI
        try {
            setJNDICachedAttributeInformation(peerIdentity,scope,info);
        } catch (Exception e) {
            logger.debug("Problem with attribute store: " + e.getMessage());
            return false;
        }

        return true;
    }

    private static synchronized void
        setJNDICachedAttributeInformation(String peerIdentity, String scope,
                                          AttributeInformation info)
        throws Exception {

        // allow null AttributeInformation, even though proper semantics is
        // to remove entry entirely

        HashMap infomap = queryJNDIgetInfomap(queryGetJNDIScope(scope));

        if (infomap == null) {
            throw new Exception("no cache");
        }

        Object o = infomap.put(peerIdentity, info);

        if (logger.isDebugEnabled()) {

            if (o != null) {
                logger.debug("previous value associated with key");
            } else {
                logger.debug("no previous value associated with key " +
                        "(or previous value was null)");
            }

        }

    }


    private static AttributeInformation
        queryJNDICachedAttributeInformation(String peer, String scope)
            throws Exception {

        AttributeInformation subjectinformation = null;

        HashMap infomap = queryJNDIgetInfomap(queryGetJNDIScope(scope));

        if (infomap == null) {
            return null;
        } else if (infomap.isEmpty()) {
            return null;
        }

        try {
            subjectinformation = (AttributeInformation) infomap.get(peer);
        } catch (ClassCastException e) {
            logger.debug("object stored in cache is not of type "
                         + "AttributeInformation");
        } catch (Exception e) {
            logger.debug("exception retrieving from cache: " + e.getMessage());
        }

        return subjectinformation;
    }

    private static HashMap queryJNDIgetInfomap(String scope)
            throws Exception {
        Object o;
        HashMap infomap;

        Context context = new InitialContext();

        try {
            o = context.lookup(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                               ShibbolethConstants.ATTRIBUTE_MAP + scope
                               + ShibbolethConstants.ATTRIBUTE_MAP_KEY);
            if (o instanceof HashMap) {
                logger.debug("found attribute cache");
                infomap = (HashMap) o;
            } else {
                throw new Exception();
            }

        } catch (Exception e) {
            logger.debug("lookup failed: " + e.getMessage());
            logger.debug("creating new cache for scope: "
                         + scope.substring(1));

            infomap = new HashMap();
            try {
                context.createSubcontext(org.globus.wsrf.Constants
                                         .JNDI_BASE_NAME +
                                         ShibbolethConstants.ATTRIBUTE_MAP
                                         + scope);
                logger.debug("created subcontext: "
                             + org.globus.wsrf.Constants.JNDI_BASE_NAME +
                             ShibbolethConstants.ATTRIBUTE_MAP + scope);
            } catch (Exception exp) {
                logger.debug("create subcontext failed, likely first cache "
                             + "store: " + exp.getMessage());

                try {
                    context.createSubcontext(org.globus.wsrf.Constants
                                             .JNDI_BASE_NAME +
                            ShibbolethConstants.ATTRIBUTE_MAP);
                    logger.debug("created subcontext: "
                                 + org.globus.wsrf.Constants.JNDI_BASE_NAME +
                                 ShibbolethConstants.ATTRIBUTE_MAP);
                    context.createSubcontext(org.globus.wsrf.Constants
                                             .JNDI_BASE_NAME +
                                             ShibbolethConstants.ATTRIBUTE_MAP
                                             + scope);
                } catch (Exception exxxp) {
                    logger.debug("creating subcontexts fails entirely: ",
                                 exxxp);
                    throw exxxp;
                }
            }
            try {
                context.bind(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                             ShibbolethConstants.ATTRIBUTE_MAP + scope
                             + ShibbolethConstants.ATTRIBUTE_MAP_KEY, infomap);
            } catch (Exception exp2) {
                logger.debug("bind failed: " + exp2.getMessage());
                //couldn't lookup but bind fails? last resort:
                context.rebind(org.globus.wsrf.Constants.JNDI_BASE_NAME +
                               ShibbolethConstants.ATTRIBUTE_MAP + scope
                               + ShibbolethConstants.ATTRIBUTE_MAP_KEY,
                               infomap);
            }
        }

        return infomap;
    }


    private static String queryGetJNDIScope(String scope)
            throws Exception {

        if (scope == null) {
            logger.debug("no scope given, looking for default cache");
            scope = ShibbolethConstants.CACHE_DEFAULT;
        } else if (scope.length() > 0) {
            scope = "/" + scope;
        } else {
            logger.debug("empty scope given, looking for default cache");
            scope = ShibbolethConstants.CACHE_DEFAULT;
        }
        return scope;
    }

    public static boolean attributeSetsMatch(Vector samlAttributes,
                                             Vector peerAttributes) {

        if (samlAttributes == null || samlAttributes.isEmpty()) {
            logger.debug("No configured attributes " +
                         "(and so policy is trivially satisfied)");
            return true;
        }
        logger.debug("Obtained " + samlAttributes.size() +
                     " configured attributes from caller");

        if (peerAttributes == null || peerAttributes.isEmpty()) {
            logger.debug("No issued attributes " +
                         "(and so policy can't possibly be satisfied)");
            return false;
        }
        logger.debug("Obtained " + peerAttributes.size() +
                     " issued attributes from caller");

        for (int i = 0; i < samlAttributes.size(); i++) {
            logger.debug("Checking configured attribute " + i);
            SAMLAttribute sa = (SAMLAttribute)samlAttributes.get(i);
            if (peerAttributes.contains(sa)) {
                logger.debug("Configured attribute was issued: " + sa);
                return true;
            }
            logger.debug("Configured attribute was NOT issued: " + sa);
        }

        logger.debug("No configured attribute in this set was issued");
        return false;
    }

    public static boolean attributeSetsMatch(Vector samlAttributes,
                                             SAMLAttribute[] attributes) {

        if (samlAttributes == null || samlAttributes.isEmpty()) {
            logger.debug("No configured attributes " +
                         "(and so policy is trivially satisfied)");
            return true;
        }
        logger.debug("Obtained " + samlAttributes.size() +
                     " configured attributes from caller");

        if (attributes == null || attributes.length == 0) {
            logger.debug("No issued attributes " +
                         "(and so policy can't possibly be satisfied)");
            return false;
        }
        logger.debug("Obtained " + attributes.length +
                     " issued attributes from caller");

        for (int i = 0; i < samlAttributes.size(); i++) {
            logger.debug("Checking configured attribute " + i);
            SAMLAttribute sa = (SAMLAttribute)samlAttributes.get(i);
            boolean found = false;
            for (int j = 0; j < attributes.length; j++) {
                if (attributes[j].equals(sa)) {
                    found = true;
                    break;
                }
            }
            if (found) {
                logger.debug("Configured attribute was issued: " + sa);
                return true;
            }
            logger.debug("Configured attribute was NOT issued: " + sa);
        }

        logger.debug("No configured attribute in this set was issued");
        return false;
    }

    public static String[] getUsernameMappings(AttributeMapping[] mappings,
                                               Iterator peerAttributes) {
        if (mappings == null) {
            return null;
        }

        if (peerAttributes == null) {
            return null;
        }

        // use set so that the add method filters out duplicates for us
        Set usernamesSet = new HashSet();

        // for each of the peer's attributes, look through all the attribute
        // mappings for matches
        while (peerAttributes.hasNext()) {

            SAMLAttribute aPeerAttr = (SAMLAttribute)peerAttributes.next();

            for (int i = 0; i < mappings.length; i++) {

                Vector mapAttrs = mappings[i].getSAMLAttributes();

                if (mapAttrs.contains(aPeerAttr)) {

                    String[] thisMappingUsernames = mappings[i].getUsernames();

                    if (thisMappingUsernames != null) {
                        for (int j = 0; j < thisMappingUsernames.length; j++) {
                            usernamesSet.add(thisMappingUsernames[j]);
                        }
                    }
                }
            }
        }

        String[] returnMappings =
                        (String[]) usernamesSet.toArray(new String[0]);

        if (returnMappings.length < 1) {

            logger.debug("No user name mapping");

            return null;

        } else {

            if (logger.isDebugEnabled()) {
                for (int i = 0; i < returnMappings.length; i++) {
                    logger.debug("Username " + i + " " + returnMappings[i]);
                }
            }

            return returnMappings;
        }
    }
}

