/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.attributes;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

public class AttributeAuthority {

    static Log logger = LogFactory.getLog(AttributeAuthority.class.getName());

    private String AAurl = null;
    private String IdPproviderId = null;
    private String nameQualifier = null;
    private SAMLAttribute[] availableAttributes = null;

    // SAML2 Metadata can note the distinction between a key's use
    // for encryption or signing and list multiple keys.  Right now,
    // we only assume one key in the RoleDescriptor and assume it's
    // used for both.
    private X509Certificate key = null;

    // We do, however, support multiple authorities
    // (X509Certificate collection)
    private Collection authorities = null;

    public AttributeAuthority(String idPproviderId) {
        this.IdPproviderId = idPproviderId;
        this.authorities = new ArrayList();
    }

    public AttributeAuthority(String AAurl,
                              String idPproviderId,
                              String nameQualifier,
                              SAMLAttribute[] availableAttributes,
                              X509Certificate key,
                              Collection authorities) {

        this.AAurl = AAurl;
        this.IdPproviderId = idPproviderId;
        this.nameQualifier = nameQualifier;
        this.availableAttributes = availableAttributes;
        this.key = key;
        if (authorities != null) {
            this.authorities = authorities;
        } else {
            this.authorities = new ArrayList();
        }
    }

    /**
     * providerId of AA.
     * @return IdPproviderId
     */
    public String getAAProviderId() {
        return IdPproviderId;
    }

    /**
     * URL of AA for this request.  Returns String because it
     * may be used in different ways (not just as URL instance).
     * @return AAurl
     */
    public String getAAURL() {
        return AAurl;
    }

    /**
     * nameQualifier attribute of the NameIdentifier element in the
     * AttributeQuery element for queries to this Attribute Authority
     * @return nameQualifier
     */
    public String getNameQualifier() {
        return nameQualifier;
    }

    /**
     * The potential attributes available at this AA
     * @return array of attributes or null if none
     */
    public SAMLAttribute[] getAvailableAttributes() {
        return availableAttributes;
    }

    /**
     * @return public key of AA or null
     */
    public X509Certificate getKey() {
        return key;
    }

    /**
     * Collection of X509Certificate objects
     * @return public keys of AA's authority (never null)
     */
    public Iterator getAuthorities() {
        return authorities.iterator();
    }

    /**
     * This does not account for differences in AAUrl, nameQualifier,
     * or keys, only IdPproviderId.  Currently onus is on developer to
     * watch for conflicts.
     *
     * @param o
     * @return true if IdPproviderId's are equal (or one is null!)
     */
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final AttributeAuthority that = (AttributeAuthority) o;

        // we treat null as a wildcard!
        if ((this.IdPproviderId == null) || (that.IdPproviderId == null)) {
            return true;
        }

        return (this.IdPproviderId.equals(that.IdPproviderId));
    }

    public int hashCode() {
        return (IdPproviderId != null ? IdPproviderId.hashCode() : 0);
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();

        toStringCommon(buf);

        buf.append("\n=== key ===\n");
        if (key == null) {
            buf.append("none\n");
        } else {
            buf.append("   Subject: ")
               .append(key.getSubjectDN())
               .append("\n   Issuer : ")
               .append(key.getIssuerDN())
               .append("\n");
        }

        buf.append("\n=== key authorities ===\n");
        if (authorities == null) {
            buf.append("none\n");
        } else {
            Iterator authorities = this.getAuthorities();
            int count = 1;
            while (authorities.hasNext()) {
                X509Certificate authority =
                        (X509Certificate) authorities.next();
                buf.append("\n#")
                   .append(count)
                   .append(":\n   Subject: ")
                   .append(authority.getSubjectDN())
                   .append("\n   Issuer : ")
                   .append(authority.getIssuerDN())
                   .append("\n");
                count += 1;
            }
        }

        return buf.toString();
    }

    private void toStringCommon(StringBuffer buf) {
        buf.append("AAurl='").append(AAurl).append("'\n")
           .append("IdPproviderId='").append(IdPproviderId).append("'\n")
           .append("nameQualifier='").append(IdPproviderId).append("'\n");

        buf.append("#attributes=");
        if (availableAttributes == null) {
            buf.append("0");
        } else {
            buf.append(availableAttributes.length);
        }
        buf.append("\n");
    }

    public String toStringDebug() {
        StringBuffer buf = new StringBuffer();

        toStringCommon(buf);

        buf.append("\n========= availableAttributes =========");

        if (availableAttributes == null) {
            buf.append("null\n");
        } else {
            buf.append("\n");
            for (int i = 0; i < availableAttributes.length; i++) {
                buf.append(availableAttributes[i])
                   .append("\n");
            }
        }

        buf.append("\n========= key =========\n");
        buf.append(key);

        buf.append("\n\n========= key authorities =========\n");
        if (authorities == null) {
            buf.append("none\n");
        } else {
            Iterator auths = this.getAuthorities();
            int count = 1;
            while (auths.hasNext()) {
                buf.append("\n\n ----------- #")
                   .append(count)
                   .append(" -----------\n");
                buf.append(auths.next());
                count += 1;
            }
        }


        return buf.toString();
    }

    /* Static helper methods: */

    /**
     * Helper method to pick a particular AA.  Does not trim strings,
     * idPproviderId must be exactly equal.
     *
     * @param idPproviderId
     * @param AAs
     * @return returns first AA w/ matching providerId, no duplicate awareness
     */
    public static AttributeAuthority findAA(String idPproviderId,
                                            AttributeAuthority[] AAs) {
        if ((idPproviderId == null) || (AAs == null)) {
            return null;
        }

        for (int i = 0; i < AAs.length; i++) {
            if (AAs[i].getAAProviderId() != null) {
                if (AAs[i].getAAProviderId().equals(idPproviderId)) {
                    return AAs[i];
                }
            }
        }

        return null;
    }

    /**
     * Helper method to pick a particular AA.  Does not trim strings,
     * idPproviderId must be exactly equal.
     *
     * @param idPproviderId
     * @param AAs must contain only AttributeAuthority objects
     * @return returns first AA w/ matching providerId, no duplicate awareness
     */
    public static AttributeAuthority findAA(String idPproviderId,
                                            Collection AAs) {
        if ((idPproviderId == null) || (AAs == null)) {
            return null;
        }

        Iterator aas = AAs.iterator();

        while (aas.hasNext()) {
            try {
                AttributeAuthority aa = (AttributeAuthority) aas.next();

                if (aa == null) {
                    logger.error("Collection contains null " +
                        "AttributeAuthority object");
                    return null;
                }

                if (aa.getAAProviderId() != null) {
                    if (aa.getAAProviderId().equals(idPproviderId)) {
                        return aa;
                    }
                }

            } catch (ClassCastException e) {
                logger.error("Collection does not contain only " +
                        "AttributeAuthority objects", e);
                return null;
            }

        }

        return null;
    }
}
