/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.samlpiputils;

import org.globus.gsi.CertUtil;
import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TimeZone;

public class CertificateSetupRun {

    public static final Integer NOTSET;
    public static final Integer STATUS;
    public static final Integer INSTALL;
    //public static final Integer FORCE;
    public static final Integer REMOVE;

    static {
        NOTSET = new Integer(-1);
        STATUS = new Integer(0);
        INSTALL = new Integer(1);
        //FORCE = new Integer(2);
        REMOVE = new Integer(3);
    }

    /**
     * @param client
     * @throws Exception
     */
    public static void run(CertificateSetup client,
                           PrintStream out,
                           PrintStream debug) throws Exception {

        AttributeAuthority[] aas;

        if (client.getMdDirPath() != null) {
            aas = MetadataUtil.loadMetadataDir(client.getMdDirPath(),true);
        } else if (client.getMdFilePath() != null) {
            aas = MetadataUtil.loadMetadata(client.getMdFilePath(),true);
        } else {
            throw new Exception("must be given file or directory path");
        }

        Cert[] trustedCerts =
               CertificateSetupUtil.loadCerts(client.getTrustedCertDirPaths());

        //Get CRLs:
        //CertificateRevocationLists defaultCrls =
        //   CertificateRevocationLists.getDefaultCertificateRevocationLists();

        if (client.getAction() == STATUS) {
            if (debug != null) {
                printTrustedCerts(out, trustedCerts, true);
            } else {
                printTrustedCerts(out, trustedCerts, false);
            }
        } else if (debug != null) {
            printTrustedCerts(debug, trustedCerts, true);
        }

        if ((aas == null) || (aas.length == 0)) {
            if (client.getAction() == STATUS) {
                out.println("There are zero attribute authorities under" +
                        " consideration, done.");
                return;
            } else {
                throw new Exception("Found no Attribute Authorities");
            }
        }

        if (aas.length == 1) {
            out.println("There is one attribute authority " +
                    "under consideration.");
        } else {
            out.println("There are " + aas.length + " attribute " +
                "authorities under consideration.");
        }

        if (client.getAction() == STATUS) {
            if (debug != null) {
                printAA(out, aas, true);
            } else {
                printAA(out, aas, false);
            }
        } else if (debug != null) {
            printAA(debug, aas, true);
        }

        Diff diff = diffAlreadyInstalled(trustedCerts, aas, debug);

        if (client.getAction() == INSTALL) {
            out.println("\n*** INSTALL:\n");
            addCerts(diff.newCerts, client, out, debug, false);
        } else if (client.getAction() == STATUS) {
            out.println("\n*** STATUS: Dryrun of the install subcommand:\n");
            addCerts(diff.newCerts, client, out, debug, true);
        }

        if (client.getAction() == REMOVE) {
            out.println("\n*** REMOVE:\n");
            removeCerts(diff.existingAACerts, client, out, debug, false);
        } else if (client.getAction() == STATUS) {
            out.println("\n*** STATUS: Dryrun of the remove subcommand:\n");
            removeCerts(diff.existingAACerts, client, out, debug, true);
        }
    }

    private static void addCerts(X509Certificate[] newCerts,
                                 CertificateSetup client,
                                 PrintStream out,
                                 PrintStream debug,
                                 boolean dryrun) {

        if (newCerts.length == 0) {
            out.println("There are no new certificates to add, done.");
            return;
        } else if (newCerts.length == 1) {
            out.println("There is one new certificate to add.");
        } else {
            out.println("There are " + newCerts.length +
                                        " new certificates to add.");
        }

        PrintStream x;
        if (client.getAction() == STATUS) {
            x = out;
        } else {
            x = debug;
        }
        if (x != null) {
            x.println("Going to try to add these certificates: ");
            for (int i = 0; i < newCerts.length; i++) {
                x.println("  --  " + newCerts[i].getSubjectDN().toString());
            }
        }

        if (client.getTrustedCertDirPaths().length == 0) {
            out.println("Certificates to add but no " +
                    "directory to add them to, done.");
            return;
        }

        // find first writable dir
        String[] dirs = client.getTrustedCertDirPaths();
        File dir = null;
        for (int i = 0; i < dirs.length; i++) {
            dir = new File(dirs[i]);
            if (!dir.canRead()) {
                out.println("Can't read '" + dir.getAbsolutePath() + "'");
                dir = null;
                continue;
            }
            if (!dir.canWrite()) {
                out.println("Can't write to '" + dir.getAbsolutePath() + "'");
                dir = null;
            }
        }

        if (dir == null) {
            out.println("Found no valid directory to " +
                                        "place new certificates, done.");
            return;
        }

        for (int i = 0; i < newCerts.length; i++) {

            out.println("\n----------------- New Cert #" + (i+1)
                                    + "-----------------");
            String base =
                    CertificateSetupUtil.nameHash(newCerts[i].getSubjectDN());

            if (base == null) {
                out.println("Could not add '"
                        + newCerts[i].getSubjectDN().toString()
                        + "': cannot generate CA hashed name");
                continue;
            }

            boolean added = false;
            int counter = 0;
            String path = dir.getAbsolutePath() + File.separator
                                + base + ".";

            while ((!added) && (counter < 10)) {
                try {
                    addAttempt(newCerts[i], path, counter, out, dryrun);
                    added = true;
                } catch (CollisionException e) {
                    counter += 1;
                } catch (Exception e) {
                    out.println("Could not add '"
                            + newCerts[i].getSubjectDN().toString()
                            + "': "
                            + e.getClass()
                            + ": " +
                            e.getMessage());
                    if (debug != null) {
                        e.printStackTrace(debug);
                    }
                    break;
                }
            }

            if (!added) {
                out.println("Could not add '"
                               + newCerts[i].getSubjectDN().toString() + "'");
            } else {
                SimpleDateFormat formatter =
                    new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
                formatter.setTimeZone(TimeZone.getTimeZone("GMT"));

                out.println("    Subject DN : " + newCerts[i].getSubjectDN());
                out.println("    Issuer DN  : " + newCerts[i].getIssuerDN());
                out.print("    Not Before : ");
                out.println(newCerts[i].getNotBefore() != null ?
                      formatter.format(newCerts[i].getNotBefore()): null);
                out.print("    Not After  : ");
                out.println(newCerts[i].getNotAfter() != null ?
                       formatter.format(newCerts[i].getNotAfter()): null);
            }
        }
    }

    private static void addAttempt(X509Certificate newCert,
                                   String base,
                                   int counter,
                                   PrintStream out,
                                   boolean dryrun)
                    throws IOException,
                           CollisionException,
                           CertificateEncodingException {

        String path = base + counter;
        File newFile = new File(path);
        if (newFile.isFile()) {
            out.println("Hash collision (rare, " +
                    "but possible for certs from same CA): " +
                        newFile.getAbsolutePath());
            throw new CollisionException();
        }

        if (!dryrun) {
            FileOutputStream fop = new FileOutputStream(newFile);
            CertUtil.writeCertificate(fop, newCert);
            fop.flush();
            fop.close();
            out.println("Wrote certificate to '" + path + "'");
        } else {
            out.println("Would have written this cert to '" + path + "': ");
            CertUtil.writeCertificate(out, newCert);
            out.println();
        }

    }

    private static void removeCerts(Cert[] existing,
                                    CertificateSetup client,
                                    PrintStream out,
                                    PrintStream debug,
                                    boolean dryrun) {

        if (existing.length == 0) {
            out.println("There are no certificates to remove, done.");
            return;
        } else if (existing.length == 1) {
            out.println("There is one certificate to remove.");
        } else {
            out.println("There are " + existing.length +
                                        " certificates to remove.");
        }

        PrintStream x;
        if (client.getAction() == STATUS) {
            x = out;
        } else {
            x = debug;
        }
        if (x != null) {
            x.println("Going to try to remove these certificates: ");
            for (int i = 0; i < existing.length; i++) {
                x.println("  --  " +
                        existing[i].getCert().getSubjectDN().toString());
                x.println("      @ " + existing[i].getAbsolutePath());
            }
        }

        if (dryrun) {
            out.println("Dryrun, ending.");
            return;
        }

        // TODO: renameTo method does not work across partitions in Unix...

        File newLocationDir = new File(client.getRemoveDirPath());
        for (int i = 0; i < existing.length; i++) {
            File removing = new File(existing[i].getAbsolutePath());
            File newLocation = new File(newLocationDir.getAbsolutePath()
                                       + File.separator + removing.getName());
            if (!removing.renameTo(newLocation)) {
                out.println("Could not move file: "
                    + removing.getAbsolutePath() + "' to '"
                    + newLocation.getAbsolutePath() + "'");
            } else {
                out.println("Moved : "
                        + existing[i].getCert().getSubjectDN().toString());
                out.println("    @ : " + removing.getAbsolutePath());
                out.println("    to: " + newLocation.getAbsolutePath());
            }
        }


    }

    private static Diff diffAlreadyInstalled(Cert[] trustedCerts,
                                             AttributeAuthority[] aas,
                                             PrintStream debug) {

        ArrayList different = new ArrayList();
        ArrayList existing = new ArrayList();

        for (int i = 0; i < aas.length; i++) {
            if (debug != null) {
                debug.println("Checking if authority cert(s) for '"
                        + aas[i].getAAProviderId()
                                 + "' already installed: ");
            }
            Iterator authorities = aas[i].getAuthorities();
            while (authorities.hasNext()) {
                X509Certificate authority =
                        (X509Certificate)authorities.next();
                boolean found = false;

                if (debug != null) {
                    debug.println("-- Authority: '" + authority.getSubjectDN()
                                        + "'");
                }

                for (int j = 0; j < trustedCerts.length; j++) {
                    if (trustedCerts[j].getCert().equals(authority)) {
                        found = true;
                        if (debug != null) {
                            debug.println("   ==> yes: " +
                                    trustedCerts[j].getAbsolutePath());
                        }
                        // there can be identical trusted certs from
                        // different AAs
                        if (!existing.contains(trustedCerts[j])) {
                            existing.add(trustedCerts[j]);
                        }
                    }
                }

                if (!found) {
                    if (debug != null) {
                        debug.println("   ==> no.");
                    }
                    // there can be identical trusted certs from different AAs
                    if (!different.contains(authority)) {
                        different.add(authority);
                    }
                }
            }
            if (debug != null) {
                debug.println();
            }
        }

        Diff ret = new Diff();
        ret.newCerts =
                (X509Certificate[]) different.toArray(new X509Certificate[0]);
        ret.existingAACerts = (Cert[]) existing.toArray(new Cert[0]);
        return ret;
    }

    private static void printAA(PrintStream out,
                                AttributeAuthority[] aas,
                                boolean printAll) {
        for (int i = 0; i < aas.length; i++) {
            out.println(new StringBuffer().
                    append("\n----------------- Attribute Authority #").
                    append(i + 1).
                    append(": -------------------").
                    toString());
            if (printAll) {
                out.println(aas[i].toStringDebug());
            } else {
                out.println(aas[i]);
            }
            out.println("----------------------------------------------" +
                "---------------\n");
        }
    }

    private static void printTrustedCerts(PrintStream out,
                                          Cert[] tc,
                                          boolean printAll) {

        SimpleDateFormat formatter =
                new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
        formatter.setTimeZone(TimeZone.getTimeZone("GMT"));

        out.println("\n-------------- Current trusted certificates:" +
                " ----------------");
        for (int i = 0; i < tc.length; i++) {
            out.println(new StringBuffer().append("Certificate #").
                            append(i + 1).append(": ").
                            append(CertificateSetupUtil.nameHash(
                                    (tc[i].getCert().getSubjectDN()))).toString());
            out.println("    Path       : " + tc[i].getAbsolutePath());
            out.println("    Subject DN : " + tc[i].getCert().getSubjectDN());
            out.println("    Issuer DN  : " + tc[i].getCert().getIssuerDN());
            if (printAll) {
                out.print("    Not Before : ");
                out.println(tc[i].getCert().getNotBefore() != null ?
                      formatter.format(tc[i].getCert().getNotBefore()): null);
                out.print("    Not After  : ");
                out.println(tc[i].getCert().getNotAfter() != null ?
                       formatter.format(tc[i].getCert().getNotAfter()): null);
            }
        }
        out.println("----------------------------------------------" +
                "---------------\n");
    }

    private static class CollisionException extends Exception {
    }

    private static class Diff {
        X509Certificate[] newCerts;
        Cert[] existingAACerts;
    }
}
