/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.samlpiputils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.asn1.DEROutputStream;
import org.bouncycastle.asn1.x509.X509Name;
import org.globus.gsi.CertUtil;
import org.globus.gsi.TrustedCertificates;
//import sun.security.x509.X500Name;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.Principal;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import javax.security.auth.x500.X500Principal;

public class CertificateSetupUtil {

    static Log logger =
            LogFactory.getLog(CertificateSetupUtil.class.getName());

    private static MessageDigest md5 = null;

    private static void init() {
        if (md5 == null) {
            try {
                md5 = MessageDigest.getInstance ("MD5");
            } catch(NoSuchAlgorithmException e) {
                logger.error("",e);
            }
        }
    }
    /**
     * @param paths can be dirs - or a mix of dirs and files
     * @return never null, len 0 or more, all exceptions squelched.
     */
    public static Cert[] loadCerts(String[] paths) {
        ArrayList allcerts = new ArrayList();
        for (int i = 0; i < paths.length; i++) {
            try {
                allcerts.addAll(load(paths[i]));
            } catch (IOException e) {
                logger.error("",e);
            } catch (GeneralSecurityException e) {
                logger.error("",e);
            }
        }
        return (Cert[]) allcerts.toArray(new Cert[0]);
    }

    /**
     * @param path file or dir
     * @return len 0+ list or exception, never null
     * @throws IOException only if path can't be read
     * @throws GeneralSecurityException only if path is a file and a problem
     */
    public static Cert[] loadCerts(String path)
                            throws IOException, GeneralSecurityException {
        return (Cert[]) load(path).toArray(new Cert[0]);
    }


    /**
     * @param path file or dir
     * @return len 0+ list or exception, never null
     * @throws IOException only if path can't be read
     * @throws GeneralSecurityException only if path is a file and a problem
     */
    private static ArrayList load(String path)
            throws IOException, GeneralSecurityException {
        ArrayList certs = new ArrayList();
        File f = new File(path);

        if (!f.canRead()) {
            throw new IOException("can't read " + f.getAbsolutePath());
        }

        if (!f.isDirectory()) {
            try {
                X509Certificate cert =
                        CertUtil.loadCertificate(
                                        f.getAbsolutePath());
                certs.add(new Cert(f.getAbsolutePath(), cert));
            } catch (GeneralSecurityException e) {
                logger.error("",e);
                throw e;
            }
        } else {
            String[] cas = f.list(TrustedCertificates.getCertFilter());
            if (cas != null) {
                for (int i = 0; i < cas.length; i++) {
                    String caFilename = f.getPath() +
                        File.separatorChar + cas[i];
                    File caFile = new File(caFilename);
                    if (caFile.canRead()) {
                        try {
                            X509Certificate cert = CertUtil.loadCertificate(
                                                     caFile.getAbsolutePath());
                            certs.add(new Cert(caFile.getAbsolutePath(),cert));
                        } catch (GeneralSecurityException e) {
                            logger.error("",e);
                        }
                    }
                }
            }
        }

        return certs;
    }

    /**
     * Returns equivalent of:
     * openssl x509 -in "cert-file" -hash -noout
     * @param subjectDN
     * @return hash for certificate names
     */
    public static String nameHash(Principal subjectDN) {
        try {
            return hash(encodePrincipal(subjectDN));
        } catch (Exception e) {
            logger.error("",e);
            return null;
        }
    }

    public static byte[] encodePrincipal(Principal subject) throws IOException {
        if (subject instanceof X500Principal) {
            return ((X500Principal)subject).getEncoded();
        //} else if (subject instanceof X500Name) {
        //    return ((X500Name)subject).getEncoded();
        } else if (subject instanceof X509Name) {
            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            DEROutputStream der = new DEROutputStream(bout);
            X509Name nm = (X509Name)subject;
            der.writeObject(nm.getDERObject());
            return bout.toByteArray();
        } else {
            throw new ClassCastException("unsupported input class: "
                                + subject.getClass().toString());
        }
    }

    // bit twiddling solution from the jglobus Attic gets it right
    // http://www.cogkit.org/viewcvs/viewcvs.cgi/src/jglobus/src/org/globus/security/Attic/HashUtil.java?rev=HEAD&content-type=text/vnd.viewcvs-markup
    private static String hash(byte [] data) {
        init();
        if (md5 == null) return null;

        md5.reset();
        md5.update (data);

        byte[] md = md5.digest();

        long ret = ( fixByte(md[0]) | (fixByte(md[1])<<8L) |
             fixByte(md[2])<<16L | fixByte(md[3])<<24L )&0xffffffffL;

        return Long.toHexString(ret);
    }

    private static long fixByte(byte b) {
        return (b<0) ? (long)(b+256) : (long)b;
    }

}







