/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.security;

import java.security.Principal;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.security.auth.Subject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.samlpiputils.SAMLutil;
import org.globus.gridshib.security.BasicAttribute;
import org.globus.gridshib.security.SAMLAuthnContext;
import org.globus.gridshib.security.SAMLIdentity;
import org.globus.gridshib.security.SAMLSecurityContext;
import org.globus.gridshib.security.SecurityContext;
import org.globus.gridshib.security.SecurityContextFactory;

import org.globus.gsi.jaas.GlobusPrincipal;
import org.globus.gsi.jaas.UserNamePrincipal;

import org.globus.opensaml11.saml.SAMLAttribute;
import org.globus.opensaml11.saml.SAMLSubjectAssertion;

/**
 * @see org.globus.gridshib.security.SAMLSecurityContext
 */
public final class GS4GTSecurityContext extends SAMLSecurityContext {

    private static final String CLASSNAME =
        GS4GTSecurityContext.class.getName();

    static Log logger = LogFactory.getLog(CLASSNAME);

    public static void init() {
        try {
            SecurityContextFactory.setSecurityContextImpl(CLASSNAME);
        } catch (ClassNotFoundException e) {
            String msg = "Class not found: " + CLASSNAME;
            throw new RuntimeException(msg, e);
        }
    }

    /**
     * A convenience method to get the one and only
     * <code>GS4GTSecurityContext</code> instance
     * associated with the given <code>Subject</code>.
     */
    public static GS4GTSecurityContext getGS4GTSecurityContext(Subject subject) {

        SecurityContext secCtx = null;
        try {
            secCtx = SecurityContextFactory.getInstance(subject, CLASSNAME);
        } catch (ClassNotFoundException e) {
            String msg = "Class not found: " + CLASSNAME;
            throw new RuntimeException(msg, e);
        }
        assert (secCtx != null);

        return (GS4GTSecurityContext)secCtx;
    }

    private List legacyAttributes;

    public GS4GTSecurityContext(Subject subject) {
        super(subject);
        this.legacyAttributes = new ArrayList();
    }

    /**
     * @deprecated use #getAttributes instead
     */
    public org.globus.gridshib.gt.authorization.attributes.SAMLAttribute[] getLegacyAttributes() {
        logger.debug("Found " + this.legacyAttributes.size() +
                     " attribute(s)");
        Object[] o = this.legacyAttributes.toArray(new org.globus.gridshib.gt.authorization.attributes.SAMLAttribute[0]);
        return (org.globus.gridshib.gt.authorization.attributes.SAMLAttribute[])o;
    }

    public Principal getGlobusPrincipal() {

        Principal[] principals = this.getPrincipals();
        assert (principals != null);

        for (int i = 0; i < principals.length; i++) {
            if (principals[i] instanceof GlobusPrincipal) {
                return principals[i];
            }
        }
        return null;
    }

    // TODO: move getSAMLPrincipals() into the superclass
    public Principal[] getSAMLPrincipals() {

        Principal[] principals = this.getPrincipals();
        assert (principals != null);
        if (principals.length == 0) return new Principal[]{};

        List samlPrincipals = new ArrayList();
        for (int i = 0; i < principals.length; i++) {
            if (principals[i] instanceof SAMLPrincipal) {
                samlPrincipals.add(principals[i]);
            }
        }
        Object[] o = samlPrincipals.toArray(new Principal[0]);
        return (Principal[])o;
    }

    public Principal[] getUserNamePrincipals() {

        Principal[] principals = this.getPrincipals();
        assert (principals != null);
        if (principals.length == 0) return new Principal[]{};

        List userNamePrincipals = new ArrayList();
        for (int i = 0; i < principals.length; i++) {
            if (principals[i] instanceof UserNamePrincipal) {
                userNamePrincipals.add(principals[i]);
            }
        }
        Object[] o = userNamePrincipals.toArray(new Principal[0]);
        return (Principal[])o;
    }

    /**
     * @return true if this security context changed as a result of the call
     *
     * @deprecated use #removeAttribute instead
     */
    public boolean removeLegacyAttribute(org.globus.gridshib.gt.authorization.attributes.SAMLAttribute attribute) {
        return this.legacyAttributes.remove(attribute);
    }

    /**
     * @return true if this security context changed as a result of the call
     */
    public boolean addSAMLAttributes(String id,
                                     String issuer,
                                     Iterator attributes) {

        assert (issuer != null && attributes != null);

        boolean result = false;
        while (attributes.hasNext()) {
            SAMLAttribute attribute = (SAMLAttribute)attributes.next();
            if (addSAMLAttribute(id, issuer, attribute)) { result = true; }

            // update legacy attribute information (bypassing AAP):
            try {
                AttributeAuthority aa = new AttributeAuthority(issuer);
                this.legacyAttributes.add(SAMLutil.parseAttribute(
                                            attribute,
                                            null,  // notBefore
                                            null,  // notOnOrAfter
                                            aa));
            } catch (Exception e) {
                logger.error("Unable to parse SAML attribute: " +
                             attribute.toString());
                logger.warn("Skipping SAML attribute");
            }
        }
        return result;
    }

    // TODO: move portions of toString(boolean) to superclass
    public String toString(boolean verbose) {

        StringBuffer buf = new StringBuffer();

        // buffer SAML assertions:
        if (verbose) {
            SAMLSubjectAssertion[] assertions = this.getSAMLAssertions();
            assert (assertions != null);
            for (int i = 0; i < assertions.length; i++) {
                buf.append("SAMLAssertion ");
                buf.append("{\n  ").append(assertions[i].toString());
                buf.append("\n}\n");
            }
        }

        // buffer Globus principal:
        Principal principal = this.getGlobusPrincipal();
        if (principal != null) {
            buf.append("Principal ");
            buf.append("{\n  name='").append(principal.getName());
            buf.append("'\n  type='Globus");
            buf.append("'\n}");
            buf.append("\n");
        }

        // buffer SAML identities:
        SAMLIdentity[] identities = this.getSAMLIdentities();
        assert (identities != null);
        for (int i = 0; i < identities.length; i++) {
            buf.append(identities[i].toString());
            buf.append("\n");
        }

        // buffer SAML principals:
        Principal[] samlPrincipals = this.getSAMLPrincipals();
        for (int i = 0; i < samlPrincipals.length; i++) {
            buf.append("Principal ");
            buf.append("{\n  name='").append(samlPrincipals[i].getName());
            buf.append("'\n  type='SAML");
            buf.append("'\n}");
            buf.append("\n");
        }

        // buffer SAML authn contexts:
        SAMLAuthnContext[] authnContexts = this.getSAMLAuthnContexts();
        assert (authnContexts != null);
        for (int i = 0; i < authnContexts.length; i++) {
            buf.append(authnContexts[i].toString());
            buf.append("\n");
        }

        // buffer basic attributes:
        BasicAttribute[] attributes = this.getAttributes();
        assert (attributes != null);
        for (int i = 0; i < attributes.length; i++) {
            buf.append(attributes[i].toString());
            buf.append("\n");
        }

        // buffer UserName principals:
        Principal[] userNamePrincipals = this.getUserNamePrincipals();
        for (int i = 0; i < userNamePrincipals.length; i++) {
            buf.append("Principal ");
            buf.append("{\n  name='").append(userNamePrincipals[i].getName());
            buf.append("'\n  type='UserName");
            buf.append("'\n}");
            buf.append("\n");
        }

        // get legacy attributes:
        if (verbose) {
            org.globus.gridshib.gt.authorization.attributes.SAMLAttribute[] attrs;
            attrs = getLegacyAttributes();

            // buffer legacy attributes:
            for (int i = 0; i < attrs.length; i++) {
                buf.append(attrs[i].toString());
                buf.append("\n");
            }
        }

        return buf.toString();
    }
}
