/*
 * Copyright 1999-2007 University of Chicago
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib;

import java.util.HashMap;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.PDPDecision;
import org.globus.gridshib.gt.authorization.GridShibConfigParams;
import org.globus.gridshib.gt.authorization.VomsConstants;

import org.globus.security.gridmap.GridMap;
import org.globus.wsrf.config.ConfigException;
import org.globus.wsrf.impl.security.authorization.exceptions.AuthorizationException;
import org.globus.wsrf.impl.security.authorization.exceptions.CloseException;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.impl.security.authorization.exceptions.InvalidPolicyException;
import org.globus.wsrf.impl.security.descriptor.SecurityPropertiesHelper;
import org.globus.wsrf.security.authorization.PDP;
import org.globus.wsrf.security.authorization.PDPConfig;

import org.w3c.dom.Node;

/**
 * GT4.0.x compatible proxy to the GridShib authorization code.
 * This is for the master PDP, typically the only configured interceptor.
 *
 * @see org.globus.wsrf.security.authorization.PDP
 */
public abstract class BasePDP implements PDP {

    static Log logger = LogFactory.getLog(BasePDP.class.getName());

    protected String scope;
    protected String serviceId;
    protected abstract org.globus.gridshib.gt.authorization.PDP pdpImpl();

    public void initialize(PDPConfig config,
                           String name,
                           String id) throws InitializeException {

        assert config != null;

        /* The following block of code works around a Globus bug:
         *
         * http://bugzilla.globus.org/globus/show_bug.cgi?id=5544
         *
         * A GS4GT PDP implementation keeps track of calls to its
         * initialize method.
         */
        if (pdpImpl().isInitialized(name)) {
            logger.info("PDP previously initialized for service " + id +
                        " (scope: " + name + ")");
            return;
        }

        logger.info("PDP initializing for service " + id +
                    " (scope: " + name + ")");

        HashMap configs = new HashMap();
        String[] keys = GridShibConfigParams.ALL_CONFIG_KEYS;
        for (int i = 0; i < keys.length; i++) {
            Object o = config.getProperty(name, keys[i]);
            if (o != null) {
                logger.debug("Config parameter " + keys[i] + ": " +
                             o.toString());
                configs.put(keys[i], o);
            }
        }

        // see note in org.globus.gridshib.gt.authorization.VomsConstants
        keys = VomsConstants.ALL_CONFIG_KEYS;
        for (int i = 0; i < keys.length; i++) {
            Object o = config.getProperty(name, keys[i]);
            if (o != null) {
                logger.debug("Config parameter " + keys[i] + ": " +
                             o.toString());
                configs.put(keys[i], o);
            }
        }

        // programmatic, alternative use of the PDP could have resulted in
        // another default gridmap to use.  This makes gridmap retrieval
        // unnecessary

        Object o =
            config.getProperty(name, GridShibConfigParams.DEFAULT_GRIDMAP);
        if (o == null) {
            // getGridMap method is 4.0.x specific
            try {
                GridMap gridmap = SecurityPropertiesHelper.
                                                   getGridMap(id, null);
                if (gridmap != null) {
                    configs.put(GridShibConfigParams.DEFAULT_GRIDMAP, gridmap);
                } else {
                    logger.info("Default gridmap is null");
                }
            } catch (ConfigException e) {
                logger.error(e.getMessage());
            }
        }

        try {
            pdpImpl().initialize(configs, name);
        } catch (Exception e) {
            throw new InitializeException(e.getMessage(), e);
        }

        this.scope = name;
        this.serviceId = id;
    }

    /**
     *
     * @return true if and only if the implementation returns
     *         <code>PDPDecision.PERMIT</code>
     */
    public boolean isPermitted(Subject peer,
                               MessageContext context,
                               QName op) throws AuthorizationException {

        assert (pdpImpl() != null);

        logger.info("PDP making access control decision for service " +
                    this.serviceId + " (scope: " + this.scope + ")...");

        int dec = PDPDecision.INDETERMINATE;
        try {
            dec = pdpImpl().isPermitted(peer, context, op);
        } catch (IllegalArgumentException e) {
            throw new AuthorizationException("Illegal argument", e);
        } catch (Exception e) {
            logger.error(e);
        }

        return (dec == PDPDecision.PERMIT);
    }

    public String[] getPolicyNames() {
        return null;
    }

    public Node getPolicy(Node node)
            throws InvalidPolicyException {
        return null;
    }

    public Node setPolicy(Node node)
            throws InvalidPolicyException {
        return null;
    }

    public void close() throws CloseException {
    }

}
