/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib;

import java.util.HashMap;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.GridShibConfigParams;
import org.globus.gridshib.gt.security.GS4GTSecurityContextLogger;

import org.globus.wsrf.impl.security.authorization.exceptions.AttributeException;
import org.globus.wsrf.impl.security.authorization.exceptions.CloseException;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.security.authorization.PDPConfig;
import org.globus.wsrf.security.authorization.PIP;

/**
 * An implementation of the GT&nbsp;4.0 <code>PIP</code>
 * interface.  All exposed PIPs should extend this
 * abstract class.
 * <p>
 * This implementation is backed by GridShib for GT
 * interceptor, that is, an interceptor that implements
 * <code>org.globus.gridshib.gt.authorization.PIP</code>.
 *
 * @see org.globus.wsrf.security.authorization.PIP
 */
public abstract class BasePIP implements PIP {

    static Log logger = LogFactory.getLog(BasePIP.class.getName());

    protected String scope;
    protected String serviceId;

    /**
     * A subclass MUST implement this method.
     */
    protected abstract org.globus.gridshib.gt.authorization.PIP pipImpl();

    public void initialize(PDPConfig config,
                           String name,
                           String id) throws InitializeException {

        assert config != null;

        /* The following block of code works around a Globus bug:
         *
         * http://bugzilla.globus.org/globus/show_bug.cgi?id=5544
         *
         * A GS4GT PIP implementation keeps track of calls to its
         * initialize method.
         */
        if (pipImpl().isInitialized(name)) {
            logger.info("PIP previously initialized for service " + id +
                        " (scope: " + name + ")");
            return;
        }

        logger.info("PIP initializing for service " + id +
                    " (scope: " + name + ")");

        HashMap configs = new HashMap();
        String[] keys = GridShibConfigParams.ALL_CONFIG_KEYS;
        for (int i = 0; i < keys.length; i++) {
            Object o = config.getProperty(name, keys[i]);
            if (o != null) {
                logger.debug("Config parameter " + keys[i] + ": " +
                             o.toString());
                configs.put(keys[i], o);
            }
        }

        try {
            pipImpl().initialize(configs, name);
        } catch (Exception e) {
            throw new InitializeException(e.getMessage(), e);
        }

        this.scope = name;
        this.serviceId = id;
    }

    public void collectAttributes(Subject peerSubject,
                                  MessageContext context,
                                  QName op) throws AttributeException {

        assert (pipImpl() != null);

        logger.info("PIP collecting attributes for service " +
                    this.serviceId + " (scope: " + this.scope + ")...");
        try {
            pipImpl().collectAttributes(peerSubject, context);
        } catch (Exception e) {
            throw new AttributeException(e.getMessage(), e);
        }

        GS4GTSecurityContextLogger.log(peerSubject);
    }

    public void close() throws CloseException {
    }
}
