/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;

import org.globus.gridshib.gt.authorization.PDP;
import org.globus.gridshib.gt.authorization.PDPDecision;
import org.globus.gridshib.gt.authorization.SAMLBlacklistPDPImpl;

import org.globus.wsrf.impl.security.authorization.exceptions.AuthorizationException;

/**
 * This PDP makes an access control decision based on
 * information from the user's security context.  If
 * an IP address or principal name identifier from the
 * security context is on a preconfigured blacklist,
 * the request is denied.
 *
 * @see org.globus.gridshib.BasePDP
 * @see org.globus.gridshib.gt.authorization.SAMLBlacklistPDPImpl
 */
public class SAMLBlacklistPDP extends BasePDP {

    /**
     * The instance of <code>SAMLBlacklistPDPImpl</code>
     * that backs this <code>SAMLBlacklistPDP</code> instance.
     */
    protected SAMLBlacklistPDPImpl impl = null;

    /**
     * Gets the instance of <code>SAMLBlacklistPDPImpl</code>
     * that backs this <code>SAMLBlacklistPDP</code> instance.
     * If no such instance exists (presumably because this is
     * the first time the method has been called), this method
     * creates a new instance of <code>SAMLBlacklistPDPImpl</code>.
     *
     * @return an instance of <code>SAMLBlacklistPDPImpl</code>
     */
    protected PDP pdpImpl() {
        if (this.impl == null) {
            this.impl = new SAMLBlacklistPDPImpl();
        }
        return this.impl;
    }

    /**
     * Overrides the corresponding method in the superclass.
     *
     * @param peer the authenticated subject
     * @param context the message context
     * @param op the requested operation
     *
     * @return false if and only if the implementation returns
     *         <code>PDPDecision.DENY</code>
     *
     * @exception org.globus.wsrf.impl.security.authorization.exceptions.AuthorizationException
     *            If the underlying implementation
     *            (<code>SAMLBlacklistPDPImpl</code>) throws an
     *            exception of any kind
     */
    public boolean isPermitted(Subject peer,
                               MessageContext context,
                               QName op) throws AuthorizationException {

        assert (pdpImpl() != null);

        logger.info("PDP making access control decision for service " +
                    this.serviceId + " (scope: " + this.scope + ")...");

        int dec;
        try {
            dec = pdpImpl().isPermitted(peer, context, op);
        } catch (Exception e) {
            throw new AuthorizationException("Authorization failed", e);
        }

        return (dec != PDPDecision.DENY);
    }
}
