/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;

import org.globus.gridshib.gt.authorization.GridShibConfigParams;
import org.globus.gridshib.gt.authorization.PIP;
import org.globus.security.authorization.PIPInterceptor;
import org.globus.security.authorization.ChainConfig;
import org.globus.security.authorization.InitializeException;
import org.globus.security.authorization.AttributeException;
import org.globus.security.authorization.EntityAttributes;
import org.globus.security.authorization.CloseException;
import org.globus.security.authorization.NonRequestEntities;
import org.globus.security.authorization.RequestEntities;

import org.globus.wsrf.config.ConfigException;
import org.globus.wsrf.impl.security.util.CredentialUtil;
import org.globus.wsrf.impl.security.util.AttributeUtil;
import org.globus.wsrf.impl.security.util.AuthzUtil;

import java.util.HashMap;

import javax.security.auth.Subject;
import javax.xml.rpc.handler.MessageContext;

/**
 * A GT4.1.2+ compatible proxy to the GridShib authorization code.
 * All exposed PIPs should extend this abstract class.
 */
public abstract class BasePIP implements PIPInterceptor {

    static Log logger = LogFactory.getLog(BasePIP.class.getName());

    protected abstract PIP pipImpl();

    public void initialize(String chainName,
                           String prefix,
                           ChainConfig config)
                    throws InitializeException {

        assert config != null;

        logger.info("PIP initializing for authz chain " + chainName +
                    " (scope: " + prefix + ")");

        try {
            AuthzUtil.parseNameValueParam(prefix, config);
        } catch (ConfigException e) {
            throw new InitializeException("problem parsing configuration",e);
        }

        HashMap configs = new HashMap();
        String[] keys = GridShibConfigParams.ALL_CONFIG_KEYS;
        for (int i = 0; i < keys.length; i++) {
            Object o = config.getProperty(prefix, keys[i]);
            if (o != null) {
                logger.debug("Config parameter " + keys[i] + ": " +
                             o.toString());
                configs.put(keys[i], o);
            }
        }

        try {
            pipImpl().initialize(configs, chainName);
        } catch (Exception e) {
            throw new InitializeException("",e);
        }

    }

    public NonRequestEntities collectAttributes(
                                        RequestEntities requestEntities)
            throws AttributeException {

        if (requestEntities == null) {
            throw new AttributeException("null requestEntities");
        }

        EntityAttributes requestorAttrs = requestEntities.getRequestor();
        Subject peerSubject = AttributeUtil.getPeerSubject(requestorAttrs);
        String peerIdentity = CredentialUtil.getIdentity(peerSubject);

        logger.debug("found peerSubject = '" + peerSubject + "'");
        logger.debug("found peerIdentity = '" + peerIdentity + "'");

        if (peerSubject == null) {
            throw new AttributeException("null peerSubject");
        }

        if (peerIdentity == null) {
            throw new AttributeException("null peerIdentity");
        }

        MessageContext ctx =  AttributeUtil.getMessageContext(requestEntities,
                                                              null);

        if (ctx == null) {
            // for now we assume we are running in container
            throw new AttributeException("null message context");
        }

        try {
            pipImpl().collectAttributes(peerSubject, ctx);
        } catch (Exception e) {
            throw new AttributeException("",e);
        }

        // not returning information about NonRequestEntities to the framework
        return null;
    }

    public void close() throws CloseException {
    }

}
