/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.attributes.AttributeInformation;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttributeInformation;
import org.globus.gridshib.gt.authorization.attributes.policy.AttributeMapping;
import org.globus.gridshib.gt.authorization.attributes.policy.SimpleSAMLAttributePolicy;
import org.globus.gridshib.gt.authorization.samlpiputils.SAMLutil;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.UserNamePrincipal;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;

import org.globus.wsrf.utils.PerformanceLog;
import org.globus.security.gridmap.GridMap;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;
import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import java.lang.reflect.Method;

public class ClassicGridShibPDPImpl extends BasePDP
                                 implements ShibbolethConstants {

    private ShibbolethPDPPolicy shibPolicy = null;
    private SAMLQueryPIPImpl PIP = null;
    private SAMLMapPIPImpl samlPIP = null;
    private boolean cacheAttributes = false;
    private String cacheScope = null;
    private boolean respectMDAttributes;

    private GridMap gridmap = null;

    private File attrConfigFile = null;
    private long attrConfigLastModified = 0;

    private static Log logger =
        LogFactory.getLog(ClassicGridShibPDPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "ClassicGridShibPDPImpl initialization complete";

    private static PerformanceLog performanceLogger;

    static {
        performanceLogger = new PerformanceLog(
                           ClassicGridShibPDPImpl.class.getName() + ".performance");
    }

    private boolean useVOMS = false;
    // Using reflection so we do not need VOMS classes for either compiling
    // or running GridShib;  see initialize()
    private Object vomsPIP = null;
    private Object vomsPDP = null;
    private static Class vomsPipClass = null;
    private static Class vomsPdpClass = null;
    private static Method vomsIsPermittedMethod = null;
    private static Method vomsCollectAttributesMethod = null;

    static {
        try {
            vomsPipClass = Class.forName(VOMS_PIP);
        } catch (Throwable t) {
            logger.debug("VOMS PIP class not available");
        }
        try {
            vomsPdpClass = Class.forName(VOMS_PDP);
        } catch (Throwable t) {
            logger.debug("VOMS PDP class not available");
        }
        if (vomsPipClass != null) {
            try {

                Class[] collectAttributesParams = new Class [] {Subject.class,
                            String.class, MessageContext.class};

                vomsCollectAttributesMethod = vomsPipClass.getMethod(
                                 "collectAttributes", collectAttributesParams);

            } catch (NoSuchMethodException e) {

                logger.error("VOMS PIP class is available but the " +
                        "collectAttributes method has a different signature " +
                        "than expected (or it does not exist at all)");
            }
        }
        if (vomsPdpClass != null) {
            try {

                Class[] isPermittedParams = new Class [] {Subject.class,
                              String.class, MessageContext.class, QName.class};

                vomsIsPermittedMethod =
                      vomsPdpClass.getMethod("isPermitted", isPermittedParams);

            } catch (NoSuchMethodException e) {

                logger.error("VOMS PDP class is available but the " +
                        "isPermitted method has a different signature than " +
                        "expected (or it does not exist at all)");
            }
        }
    }

    /**
     * Loads policies, configures caching and initializes PIP.
     * <p>
     * If configuration object contains a <code>ShibbolethPDPPolicy</code>,
     * keyed by <code>ShibbolethConstants.SHIB_PDP_POLICY_KEY</code>, DN and
     * attribute ACLs are not considered (even if settings for them are in the
     * configuration). The policy is either managed by ACLs or in memory (for
     * instance, managed by a service that configures custom policies for
     * itself or particular resources).
     * <p>
     * The configuration object is also used by the PIP that this PDP
     * insantiates and initializes.  See the PIP documentation for what
     * is required to initialize the PIP correctly {@link SAMLQueryPIPImpl}
     * <p>
     * Required configurations:
     * <ul>
     * <li> When using ACL based policy, ShibbolethConstants.ATTR_FILE_KEY is
     * required.
     * <li>When using in memory policy, ShibbolethConstants.SHIB_PDP_POLICY_KEY
     * is required.
     * </ul>
     *
     * Optional configurations:
     * <ul>
     * <li> When using ACL based policy, ShibbolethConstants.DN_FILE_KEY can be
     * configured.
     * <li>When using either kind of policy, the ShibbolethConstants.CACHE_KEY
     * can be set to "true" to enable attribute caching.
     * </ul>
     *
     * @param configs object containing configuration parameters
     * @param name String differentiates cache
     * @throws Exception problem
     * @see SAMLQueryPIPImpl#initialize(java.util.HashMap, String)
     * @see ShibbolethPDPPolicy
     * @see ShibbolethConstants
     */
    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("ClassicGridShibPDPImpl initializing in scope " + name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("ClassicGridShibPDPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        /* If assertions are enabled, DEV will be set to true */
        boolean DEV = false;
        assert DEV = true;

        if (DEV) {
            if (performanceLogger != null) {
                performanceLogger.start(); //PDP initialization
            }
        }

        /* Initialize the PIP to use with this PDP. The particular
           authorization chain logic we would like can not be
           implemented without creating a sub-chain in this PDP  */
        this.PIP = new SAMLQueryPIPImpl();
        try {
            this.PIP.initialize(configs, name);
        } catch (Exception e) {
            logger.error("Problem initializing PIP: " + e.getMessage());
            this.PIP = null;
            throw e;
        }

        /* Check if caching is desired */

        // any error for the cache flag is just logged
        try {
            Object cacheflag = configs.get(CACHE_KEY);

            if (cacheflag == null) {
                logger.debug("no cache configuration, default is false");
                // cache flag could have been set at previous configuration
                // (i.e., it is not always true that initialize is called
                //  only at object instantiation time)
                this.cacheAttributes = false;
            } else {
                String cachestr = (String) cacheflag;
                if (cachestr.trim().equalsIgnoreCase("true")) {
                    logger.debug("caching activated");
                    this.cacheAttributes = true;
                    /* every authz chain scope will have its own cache scope */
                    this.cacheScope = name;
                } else {
                    logger.debug("cache configured, but does not match "
                                 + "'true'");
                    //need to explicitly set to false, see above
                    this.cacheAttributes = false;
                }
            }
        } catch (ClassCastException e) {
            logger.warn("caching configuration is present but it is " +
                    "not the proper class, programmer error");
            this.cacheAttributes = false;
        } catch (Exception e) {
            logger.warn(e.getMessage());
            this.cacheAttributes = false;
        }


        /*  Identity based authorization, preempts attribute based
            authorization

           Any exceptions are just logged.  */

        try {
            Object useGridmap = configs.get(CONSULT_GRIDMAP_KEY);

            if (useGridmap == null) {
                logger.debug("no use-gridmap configuration");
            } else {
                String useGridmapVal = (String) useGridmap;
                if (useGridmapVal.trim().equalsIgnoreCase("true")) {
                    Object gm = configs.get(DEFAULT_GRIDMAP);
                    if (gm == null) {
                        logger.warn("gridmap is requested but there is " +
                                "no default gridmap available");
                    } else {
                        try {
                            this.gridmap = (GridMap) gm;
                            logger.info("gridmap enabled");
                        } catch (ClassCastException e) {
                            logger.warn("default gridmap configuration " +
                                  "is present but it is not the proper " +
                                  "class (GridMap), programmer error: " +
                                  e.getMessage());
                        }
                    }
                }
            }

        } catch (ClassCastException e) {
            logger.warn("consultDefaultGridmap configuration is present " +
                    "but it is not the proper class (String), " +
                    "programmer error: " + e.getMessage());
        } catch (Exception e) {
            logger.warn(e.getMessage());
        }

        // triggers an invocation to the VOMS modules (result of which can
        // be combined in an OR with DNs and SAML attrs)
        String useVOMSString = (String) configs.get(USE_VOMS);
        if (useVOMSString != null) {
            if (useVOMSString.trim().equalsIgnoreCase("true")) {
                this.useVOMS = true;
            }
        }

        if (this.useVOMS) {

            if (vomsPipClass == null) {
                throw new Exception("VOMS is configured but VOMS PIP class " +
                              "could not be instantiated, the VOMS package " +
                              "must be deployed to GLOBUS_LOCATION in order " +
                              "to use VOMS");
            }

            if (vomsPdpClass == null) {
                throw new Exception("VOMS is configured but VOMS PDP class " +
                              "could not be instantiated, the VOMS package " +
                              "must be deployed to GLOBUS_LOCATION in order " +
                              "to use VOMS");
            }

            this.vomsPIP = vomsPipClass.newInstance();
            this.vomsPDP = vomsPdpClass.newInstance();

            Method pipInit;
            Class[] initParams = new Class [] {HashMap.class,  String.class};
            Object[] args = new Object[] {configs, name};
            try {
                pipInit = vomsPipClass.getMethod("initialize", initParams);
                pipInit.invoke(this.vomsPIP, args);
            } catch (Exception e) {
                if (logger.isDebugEnabled()) {
                    logger.error("problem initializing VOMS PIP: " +
                                                e.getCause().getMessage(), e);
                } else {
                    logger.error("problem initializing VOMS PIP: " +
                                                e.getCause().getMessage());
                }
                throw new Exception("problem initializing VOMS PIP: " +
                                                e.getCause().getMessage(), e);
            }

            Method pdpInit;
            try {
                pdpInit = vomsPdpClass.getMethod("initialize", initParams);
                pdpInit.invoke(this.vomsPDP, args);
            } catch (Exception e) {
                if (logger.isDebugEnabled()) {
                    logger.error("problem initializing VOMS PDP: " +
                                                e.getCause().getMessage(), e);
                } else {
                    logger.error("problem initializing VOMS PDP: " +
                                                e.getCause().getMessage());
                }
                throw new Exception("problem initializing VOMS PDP: " +
                                                e.getCause().getMessage(), e);
            }
        }


        Object shibPolicyObj = configs.get(SHIB_PDP_POLICY_KEY);

        if (shibPolicyObj != null) {
            // provided programmatically
            try {
                this.shibPolicy = (ShibbolethPDPPolicy) shibPolicyObj;
                if (logger.isDebugEnabled()) {
                    logger.debug("using policy from configuration, not files");
                    logger.debug(this.shibPolicy);
                }

            } catch (ClassCastException e) {
                logger.debug("shibPDPPolicy is present but it is not " +
                        "the proper class, programmer error");
                throw new Exception("no shibPDPPolicy configuration");
            }
        } else {
            // via attribute ACL

            // contains an array of attributes and an array of
            // username mappings
            AttributeMapping[] attributeMapping;

            String configFile = (String) configs.get(ATTR_FILE_KEY);
            if (configFile == null) {
                logger.warn("no attribute authz file configuration");
                throw new Exception("No attribute authz file configuration");
            }

            this.attrConfigFile = new File(configFile);

            if (!this.attrConfigFile.isAbsolute()) {
                String msg = "Path is not absolute: \"" + configFile + "\"";
                logger.error(msg);
                throw new Exception(msg);
            }

            try {
                this.attrConfigLastModified =
                                            this.attrConfigFile.lastModified();

                attributeMapping = SimpleSAMLAttributePolicy.
                                            parseAttrFile(this.attrConfigFile);

            } catch (Exception e) {
                throw new Exception(e.getMessage(),e);
            }

            this.shibPolicy = new ShibbolethPDPPolicy(attributeMapping);
        }

        this.samlPIP = new SAMLMapPIPImpl();
        try {
            this.samlPIP.initialize(configs, name);
        } catch (Exception e) {
            logger.error("Problem initializing SAML Map PIP: "
                         + e.getMessage());
            throw e;
        }


        String respectAttributesConfig = (String) configs.get(RESPECT_KEY);

        this.respectMDAttributes = false;
        if (respectAttributesConfig != null) {
            if (respectAttributesConfig.trim().equalsIgnoreCase("true")) {
                this.respectMDAttributes = true;
            }
        }

        // matches beginning of initialize()
        if (DEV) {
            if (performanceLogger != null) {
                performanceLogger.stop("PDP initialization");
            }
        }

        logger.debug("broadcast:" +
                     "\n\n   *************************" +
                       "\n     END INITIALIZING PDP" +
                       "\n   *************************\n");

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
    }


    /**
     *  Runs the masterPDP logic, which:
     *  1. optionally checks the DN ACL
     *  2. optionally makes an AA query
     *  3. checks the Attribute ACL
     *
     *  Exceptions are only thrown in extreme circumstances. If misconfigured
     *  or if there are problems with the callout, exceptions are NOT generally
     *  thrown.
     *
     *  0. init
     *
     *  1. If a DN ACL is configured, the PDP first checks if
     *  the caller?s DN is in the ACL.  If so, it returns PERMIT
     *  to the container and the PDP is done.
     *
     *  2. If the DN ACL is not applicable, the cache flag is checked.
     *  If the cache flag is set to TRUE (the default is FALSE), assertions
     *  about the caller are queried.
     *
     *  3. If caching, if there are previously obtained assertions, a
     *  validity check is made on all assertions.   If any single one
     *  is invalid, they are all discarded.
     *
     *  4. If caching and there are valid assertions, use these attributes
     *  instead of calling the PIP.
     *
     *  5. If 'respecting metadata attributes' and the attributes in ACL
     *  can not be returned from the IdP, do not call the PIP.
     *
     *  6. If calling the PIP, do it now.  If successful and valid assertions
     *  come back, parse out the attributes.
     *
     *  7. Compare the current information about the subject to the attribute
     *  ACL and return PERMIT/DENY accordingly.
     *
     * @param peerSubject subject object
     * @param context msgCtx
     * @param operation operation
     * @return PDPDecision constant (int)
     */
    public int isPermitted(Subject peerSubject,
                           MessageContext context,
                           QName operation) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());
        logger.debug("operation = " + operation.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("ClassicGridShibPDPImpl making decision " +
                    "for identity " + peerIdentity);

        boolean ret = false;

        try {
            ret = isPermittedImpl(peerSubject, peerIdentity, context, operation);
        } catch (Exception e) {
            // catch all, log and return false (DENY) (per 4/28/05 meeting)
            logger.error(e);
        }

        logger.debug("broadcast:" +
                     "\n\n   ***************" +
                       "\n     ENDING PDP" +
                       "\n   ***************\n");

        logger.info("ClassicGridShibPDPImpl decision making complete");

        /* In the future, isPermittedImpl will be more expressive (for 4.1+) */
        if (ret) {
            return PDPDecision.PERMIT;
        } else {
            return PDPDecision.DENY;
        }
    }

    private boolean isPermittedImpl(Subject peer, String peerIdentity,
                                    MessageContext msgCtx, QName op)
                             throws Exception {

        /*
         *  0. init
         */

        // Figure out what policy is being used and update accordingly.
        // Policy can come from in-memory (for instance, if a service creates
        // a PDP instance for a WSRF resource) or from files.  Currently these
        // are the only two distinctions we make. If files, we need to check
        // on modification times.

        // A possible optimization would be to move the separate checks to only
        // be invoked if necessary.  For instance, if the DN policy returns
        // permit the attribute policy would not have needed to be updated.

        checkPolicy();

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peer);
        assert (secCtx != null);

        /*
         *  1. If a DN ACL is configured, the PDP first checks if the caller's
         *  DN is in the ACL.  If so, it returns PERMIT to the
         *  container and the PDP is done.
         */

        if (this.gridmap != null) {
            String[] accounts = this.gridmap.getUserIDs(peerIdentity);

            // gridmap semantics require at least one user mapping for
            // a valid entry
            if (accounts != null && accounts.length > 0) {

                for (int i = 0; i < accounts.length; i++) {
                    //peer.getPrincipals().add(
                    //               new UserNamePrincipal(accounts[i]));
                    secCtx.addPrincipal(new UserNamePrincipal(accounts[i]));
                }

                logger.info("ACCEPTED by DN: Operation " + op +
                    " called by subject: " + peerIdentity +
                    " --- Username mappings: " + accounts.toString());

                return true;
            } else {
                logger.info("DN authorization returned not applicable");
            }
        }

        /*
         *  2. If the DN ACL is not applicable, check if VOMS modules are
         *  configured.  This is also in an OR-wise combination with the
         *  SAML attrs: if this returns PERMIT, this PDP is done -- otherwise
         *  we continue on.
         */
        if (this.useVOMS) {
            if (this.vomsPIP == null) {
                throw new Exception("VOMS PIP was never instantiated? " +
                        "ShibboletPDP initialize never called?");
            }
            if (this.vomsPDP == null) {
                throw new Exception("VOMS PDP was never instantiated? " +
                        "ShibboletPDP initialize never called?");
            }

            if (vomsCollectAttributesMethod == null) {
                throw new Exception("voms PIP method wasn't" +
                        "initialized properly");
            }

            if (vomsIsPermittedMethod == null) {
                throw new Exception("voms PDP method wasn't" +
                        "initialized properly");
            }

            Object[] args = new Object[] {peer, peerIdentity, msgCtx};
            vomsCollectAttributesMethod.invoke(this.vomsPIP, args);

            Object[] args2 = new Object[] {peer, peerIdentity, msgCtx, op};
            Integer result = (Integer)
                         vomsIsPermittedMethod.invoke(this.vomsPDP, args2);
            if (result == null) {
                logger.fatal("result from VOMS was null? skipping");
            } else {
                if (result.intValue() == PDPDecision.PERMIT) {
                    logger.info("ACCEPTED by VOMS attribute: Operation " +
                            op + " called by subject: " + peerIdentity);
                    return true;
                } else {
                    logger.info("VOMS returned not applicable");
                }
            }

        }

        /*
         *  3. If the DN ACL is not applicable, the cache flag is checked.
         *  If the cache flag is set to TRUE (the default is FALSE), assertions
         *  about the caller are queried.
         */

        boolean callPIP = true;
        SAMLAttributeInformation samlinfo = null;

        if (this.cacheAttributes) {

            /*  4. If caching, if there are previously obtained assertions, a
            *  validity check is made on all assertions.   If any single one
            *  is invalid, they are all discarded.
            */

            // TODO: once AA to use can be designated by an assertion in the
            //       user credential, when caching is on and a DN with cached
            //       attributes comes along and invokes but now with a
            //       different AA to use, this code below will not pick this
            //       fact up!

            logger.debug("Looking up attribute cache for subject");
            AttributeInformation subjectinfo =
                    ShibbolethPDPUtil.queryCachedAttributeInformation(
                            peerIdentity,cacheScope);

            if (subjectinfo != null) {
                //if present and all valid, set callPIP = false
                //and assign samlinfo
                logger.debug("Found information in cache for subject: "
                        + subjectinfo.toString());
                boolean foundinvalid = false;
                try {
                    samlinfo = (SAMLAttributeInformation) subjectinfo;
                    Vector attrs = samlinfo.getAttrs();
                    for (int i = 0; i < attrs.size(); i++) {
                        SAMLAttribute attr = (SAMLAttribute)attrs.get(i);
                        if (!attr.isValid()) {
                            foundinvalid = true;
                            break;
                        }
                    }
                    if (!foundinvalid) {
                        logger.debug("all " + attrs.size() +
                                " attributes are still valid");
                    }

                } catch (ClassCastException e) {
                    logger.warn("AttributeInformation in cache is not " +
                            "SAMLAttributeInformation -- other " +
                            "system using your cache?",e);
                } catch (Exception e) {
                    logger.debug("problem, assuming bad cache/problem " +
                            "with attributes",e);
                    foundinvalid = true;
                }

                if (foundinvalid) {
                    // call PIP, don't use cached attributes.
                    // to save time, don't wipe the cache, PIP will replace
                    // them if it received valid assertions
                    samlinfo = null;
                } else {
                    callPIP = false;
                }
            }

        }

        // the attribute authority to call
        // don't look up immediately in case cache is used
        AttributeAuthority attributeAuthority = null;

        // if already using cache, don't even look at
        // respectMDattributes optimization
        if (callPIP) {

            /*  6. If respecting metadata attributes and the attributes
             *  in ACL can not be returned from the IdP, do not call the PIP.
             *
             *  (step 5 is below)
             */

            if (this.respectMDAttributes) {
                logger.debug("looking for matching attributes in IdP" +
                        " metadata and ACL");

                if (attributeAuthority == null) {
                    if (this.PIP != null) {
                        AttributeAuthority[] authorities =
                                    this.PIP.getAuthorities();

                        attributeAuthority =
                                SAMLutil.chooseAA(authorities, peer);
                                //SAMLQueryPIPImpl.chooseAA(authorities, peer);
                    }
                }


                if (attributeAuthority == null) {
                    logger.error("correct providerId not in " +
                            "metadata or explicitly configured");

                    // here we could return DENY as an optimization
                    // because we assume SAMLQueryPIPImpl will not
                    // collect attributes when this is null; but
                    // because it will not result in any remote
                    // call anyhow, let's not meddle for such a
                    // small savings (also, see autonomy comment below)
                } else {

                    SAMLAttribute[] idPattributes =
                            attributeAuthority.getAvailableAttributes();

                    if ((idPattributes == null) ||
                        (idPattributes.length == 0)) {

                        logger.debug("no attributes available" +
                                " in attribute authority information," +
                                " interpreting as wildcard");
                    } else {

                        if (!ShibbolethPDPUtil.isAttributePermitted(
                                idPattributes, this.shibPolicy)) {

                            logger.debug("no attribute in the ACL is in" +
                                         "the IdP metadata, NOT calling PIP");
                            callPIP = false;
                        }
                    }
                }
            }
        }


        if (callPIP) {
            /* 7. If calling the PIP, do it now.  If successful and
             * valid assertions come back, parse out the attributes.*/

            callPIP(peer, msgCtx);

            // retrieve SAMLAttributeInformation from message context
            // this is all kept this way (slower) so that PIP can maintain
            // autonomy

            Set credSet = peer.getPublicCredentials();
            Iterator creds = credSet.iterator();
            while (creds.hasNext()) {
                Object o = creds.next();
                if (o instanceof SAMLAttributeInformation) {
                    samlinfo = (SAMLAttributeInformation) o;
                    break;
                }
            }

            if (samlinfo == null) {
                logger.warn("cannot retrieve attribute info from Subject");
            }

        } else {
            /*  5. If caching and there are valid assertions, use these
             *    attributes instead of calling the PIP.*/
            /* Also can be disabled if analysis of metadata attributes
               for this IdPproviderId found no possible match w/ ACL */
            logger.debug("not calling PIP");
        }

        logger.debug("SAML PIP Collect Attributes");
        if (this.samlPIP != null) {
            try {
                this.samlPIP.collectAttributes(peer, msgCtx);
            } catch (Exception e) {
                logger.error("error collecting attributes: "
                             + e.getMessage());
            }
        }

        // at this point we have cached attributes, 0 or more
        // attributes from AA query (all considered valid), or
        // did not call PIP for some other reason and have zero
        // attributes.

        boolean attrPassed =
            ShibbolethPDPUtil.isAttributePermitted(peer, this.shibPolicy);

        /* see bugzilla #2287

           if 'return false' based on attribute, the exception returned to
           client by ServiceAuthorizationChain claims DN is denied.

           One possible workaround, AuthorizationException, implies error in
           policy processing, so something along these lines is actually not
           a viable alternative:

           throw new AuthorizationException("primary attribute denied: "
           + primaryAttr);
        */

        if (attrPassed) {
            logger.info("ACCEPTED by attribute: Operation " +
                    op + " called by subject: " + peerIdentity);
        } else {
            logger.info("DENIED: Operation " +
                    op + " called by subject: " + peerIdentity);
        }

        return attrPassed;
}


    /* ************************* */
    /* Non-static helper methods */
    /* ************************* */

    // (static helpers are in ShibbolethPDPUtil)

    private void callPIP(Subject peer,
                         MessageContext context) {

        if (this.PIP == null) {
            logger.error("no PIP instance configured");
            return;
        }

        try {
            this.PIP.collectAttributes(peer, context);
        } catch (Exception e) {
            logger.error("error calling PIP: " + e.getMessage());
        }

    }

    /*
     * Upon each invocation, it's possible the policy needs to be updated.
     * Currently, 2 types of policies are supported, in-memory and file based.
     * If file based, check modification times of files and reload policy if
     * needed.
     * A possible optimization would be to move the separate checks to only
     * be invoked if necessary.  For instance, if the DN policy returns permit
     * the attribute policy would not have been needed an update
     */
    private void checkPolicy() throws Exception {

        // this should be an impossible state, but could
        // exist if class used inappropriately, for instance
        // if PDP is not instantiated and invoked by the GT
        // authorization framework
        assert this.shibPolicy != null;

        if (this.gridmap != null) {
            this.gridmap.refresh();
        }

        if (this.attrConfigFile != null) {
            checkAttrFile();
        }
    }

    /**
     * Checks file for modifications, if present, update policy.
     * If called and attrConfigFile is null, programmer error,
     * because file is either configured OR policy in memory is
     * used.  (one reason this method is private)
     * @throws Exception problem parsing file
     */
    private void checkAttrFile() throws Exception {

        assert this.attrConfigFile != null;
        assert this.shibPolicy != null;

        logger.debug("Last modification time for " +
                this.attrConfigFile.getAbsolutePath() + ": " +
                this.attrConfigFile.lastModified());

        // If config file has been modified, re-read
        if (this.attrConfigLastModified < this.attrConfigFile.lastModified()) {
            logger.debug(this.attrConfigLastModified + " < " +
                    this.attrConfigFile.lastModified());

            this.shibPolicy.replaceAttrs(SimpleSAMLAttributePolicy.
                    parseAttrFile(this.attrConfigFile));

            this.attrConfigLastModified =
                    this.attrConfigFile.lastModified();
        }
    }

}
