/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import java.util.HashMap;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.GS4GTSecurityContextLogger;
import org.globus.gridshib.gt.security.UserNamePrincipal;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;

import org.globus.gridshib.security.SAMLIdentity;

import org.globus.security.gridmap.GridMap;

/**
 * A container for GridShib combined interceptors.
 * These interceptors share a common pool of GridShib
 * for GT PIPs and PDPs.
 */
public class GS4GT {

    private static Log logger =
        LogFactory.getLog(GS4GT.class.getName());

    private static final String PUSH_INITIALIZATION_MSG =
        "GS4GT.PushPDPImpl initialization complete";
    private static final String PULL_INITIALIZATION_MSG =
        "GS4GT.PullPDPImpl initialization complete";
    private static final String PUSH_PULL_INITIALIZATION_MSG =
        "GS4GT.PushPullPDPImpl initialization complete";
    private static final String PRE_POLICY_INITIALIZATION_MSG =
        "GS4GT.PrePolicyPDPImpl initialization complete";
    private static final String PUSH_COMPLETION_MSG =
        "GS4GT.PushPDPImpl decision making complete";
    private static final String PULL_COMPLETION_MSG =
        "GS4GT.PullPDPImpl decision making complete";
    private static final String PUSH_PULL_COMPLETION_MSG =
        "GS4GT.PushPullPDPImpl decision making complete";
    private static final String PRE_POLICY_COMPLETION_MSG =
        "GS4GT.PrePolicyPDPImpl decision making complete";

    private static AttributeAcceptancePIPImpl aapPIP = null;
    //private static GridMap gridmap = null;
    private static PrePolicyPDPImpl prePolicyPDP = null;
    private static PullPDPImpl pullPDP = null;
    private static SAMLAssertionPushPIPImpl pushPIP = null;
    private static SAMLAttributePDPImpl samlPDP = null;
    private static SAMLBlacklistPDPImpl blPDP = null;
    private static SAMLMapPIPImpl mapPIP = null;
    private static SAMLQueryPIPImpl queryPIP = null;

    static {
        aapPIP = new AttributeAcceptancePIPImpl();
        prePolicyPDP = new PrePolicyPDPImpl();
        pullPDP = new PullPDPImpl();
        pushPIP = new SAMLAssertionPushPIPImpl();
        samlPDP = new SAMLAttributePDPImpl();
        blPDP = new SAMLBlacklistPDPImpl();
        mapPIP = new SAMLMapPIPImpl();
        queryPIP = new SAMLQueryPIPImpl();
    }

    /**
     * Combined interceptor for GridShib attribute push.
     *
     * @see org.globus.gridshib.gt.authorization.BasePDP
     * @see org.globus.gridshib.gt.authorization.GridShibConfigParams
     */
    public static class PushPDPImpl extends BasePDP
                                 implements GridShibConfigParams {

        public void initialize(HashMap configs, String name) throws Exception {

            logger.info("GS4GT.PushPDPImpl initializing in scope " + name);

            if (configs == null) {
                String msg = "HashMap configs is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("configs = " + configs.toString());

            if (this.isInitialized(name)) {
                logger.info("GS4GT.PushPDPImpl previously initialized " +
                            "in scope " + name);
                logger.info(PUSH_INITIALIZATION_MSG);
                return;
            }

            try {
                prePolicyPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize GS4GT.PrePolicyPDPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                samlPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLAttributePDPImpl";
                logger.error(msg, e);
                throw e;
            }

            super.initialize(configs, name);
            logger.info(PUSH_INITIALIZATION_MSG);
        }

        /**
         * This PDP implementation orchestrates the
         * following specific workflow:
         * <ol>
         *   <li>Invoke <code>GS4GT.PrePolicyPDPImpl</code>
         *     <ul>
         *       <li>If <code>PERMIT</code> or <code>DENY</code>,
         *       return <code>PERMIT</code> or <code>DENY</code>
         *       (resp.); otherwise, continue</li>
         *     </ul>
         *   <li>Invoke <code>SAMLAttributePDPImpl</code>
         *   <li>Return SAMLAttributePDPImpl decision
         * </ol>
         * If this PDP reaches the end of the authorization chain
         * without returning <code>PERMIT</code> or <code>DENY</code>,
         * it returns whatever <code>SAMLAttributePDPImpl</code>
         * returns.
         */
        public int isPermitted(Subject peerSubject,
                               MessageContext context,
                               QName operation) throws Exception {

            if (peerSubject == null) {
                String msg = "Subject is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("peerSubject = " + peerSubject.toString());
            if (context == null) {
                String msg = "MessageContext is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("context = " + context.toString());
            logger.debug("operation = " + operation.toString());

            String peerIdentity =
                GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
            logger.info("GS4GT.PushPDPImpl making decision " +
                        "for identity " + peerIdentity);

            int decisionCode;

            // invoke GS4GT.PrePolicyPDPImpl:
            try {
                decisionCode =
                    prePolicyPDP.isPermitted(peerSubject, context, operation);
            } catch (Exception e) {
                String msg = "GS4GT.PrePolicyPDPImpl decision making failed";
                logger.error(msg, e);
                return PDPDecision.INDETERMINATE;
            }
            if (decisionCode == PDPDecision.PERMIT) {
                logger.info("GS4GT.PrePolicyPDPImpl returned PERMIT");
                logger.info("GS4GT.PushPDPImpl short-circuited");
                logger.info(PUSH_COMPLETION_MSG);
                return PDPDecision.PERMIT;
            } else if (decisionCode == PDPDecision.DENY) {
                logger.info("GS4GT.PrePolicyPDPImpl returned DENY");
                logger.info("GS4GT.PushPDPImpl short-circuited");
                logger.info(PUSH_COMPLETION_MSG);
                return PDPDecision.DENY;
            }

            // invoke SAMLAttributePDPImpl:
            try {
                decisionCode =
                    samlPDP.isPermitted(peerSubject, context, operation);
            } catch (Exception e) {
                String msg = "SAMLAttributePDPImpl decision making failed";
                logger.error(msg, e);
                return PDPDecision.INDETERMINATE;
            }

            logger.info("GS4GT.PushPDPImpl returned " +
                        PDPDecision.getString(decisionCode) + " decision");
            logger.info(PUSH_COMPLETION_MSG);

            return decisionCode;
        }
    }

    /**
     * Combined interceptor for GridShib attribute pull.
     *
     * @see org.globus.gridshib.gt.authorization.BasePDP
     * @see org.globus.gridshib.gt.authorization.GridShibConfigParams
     */
    public static class PullPDPImpl extends BasePDP
                                 implements GridShibConfigParams {

        public void initialize(HashMap configs, String name) throws Exception {

            logger.info("GS4GT.PullPDPImpl initializing in scope " + name);

            if (configs == null) {
                String msg = "HashMap configs is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("configs = " + configs.toString());

            if (this.isInitialized(name)) {
                logger.info("GS4GT.PullPDPImpl previously initialized " +
                            "in scope " + name);
                logger.info(PULL_INITIALIZATION_MSG);
                return;
            }

            try {
                queryPIP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLQueryPIPImpl";
                logger.error(msg, e);
                throw e;
            }

            // short-circuit if attribute query is disabled:
            if (!queryPIP.isAttributeQueryEnabled()) {
                logger.info("Attribute query is disabled");
                logger.info("GS4GT.PullPDPImpl short-circuited");
                return;
            }

            try {
                aapPIP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize AttributeAcceptancePIPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                blPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLBlacklistPDPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                mapPIP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLMapPIPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                samlPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLAttributePDPImpl";
                logger.error(msg, e);
                throw e;
            }

            // throw exception if attribute authz is disabled:
            if (!samlPDP.isAttributeAuthzEnabled()) {
                String msg = "Attribute-based authz is not enabled";
                logger.error(msg);
                throw new RuntimeException(msg);
            }

            super.initialize(configs, name);
            logger.info(PULL_INITIALIZATION_MSG);
        }

        /**
         * This PDP implementation orchestrates the
         * following specific workflow:
         * <ol>
         *   <li>
         *     For each SAML identity in the security context,
         *     do the following:
         *     <ul>
         *       <li>Invoke <code>SAMLQueryPIPImpl</code>
         *       <li>Invoke <code>AttributeAcceptancePIPImpl</code>
         *       <li>Invoke <code>SAMLBlacklistPDPImpl</code>
         *         <ul>
         *           <li>If <code>DENY</code>, return <code>DENY</code>;
         *           otherwise, continue iterating</li>
         *         </ul>
         *       <li>Invoke <code>SAMLAttributePDPImpl</code>
         *         <ul>
         *           <li>If <code>PERMIT</code>, break;
         *           otherwise, continue iterating</li>
         *         </ul>
         *     </ul>
         *   <li>Invoke <code>SAMLMapPIPImpl</code>
         *   <li>Return <code>SAMLAttributePDPImpl</code> decision
         * </ol>
         * <code>GS4GT.PullPDPImpl</code> returns whatever
         * <code>SAMLAttributePDPImpl</code> returned on the
         * last iteration through the loop.  If there are
         * no SAML identities in the security context (in
         * which case the loop is not executed), this PDP
         * returns <code>INDETERMINATE</code>.
         */
        public int isPermitted(Subject peerSubject,
                               MessageContext context,
                               QName operation) throws Exception {

            if (peerSubject == null) {
                String msg = "Subject is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("peerSubject = " + peerSubject.toString());
            if (context == null) {
                String msg = "MessageContext is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("context = " + context.toString());
            logger.debug("operation = " + operation.toString());

            String peerIdentity =
                GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
            logger.info("GS4GT.PullPDPImpl making decision " +
                        "for identity " + peerIdentity);

            // if attribute query is disabled, short-circuit this method:
            if (queryPIP.isAttributeQueryEnabled()) {
                logger.info("Attribute query is enabled");
            } else {
                logger.info("Attribute query is disabled");
                return PDPDecision.NOT_APPLICABLE;
            }

            // invoke a query sequence, perhaps multiple times:
            int decisionCode =
                isPermittedImpl(peerSubject, context, operation);

            logger.info("GS4GT.PullPDPImpl returned " +
                        PDPDecision.getString(decisionCode) + " decision");
            logger.info(PULL_COMPLETION_MSG);

            return decisionCode;
        }

        private int isPermittedImpl(Subject peerSubject,
                                    MessageContext context,
                                    QName op) {

            assert (peerSubject != null && context != null);

            int decisionCode = PDPDecision.INDETERMINATE;

            // get security context:
            GS4GTSecurityContext secCtx =
                GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
            assert (secCtx != null);

            SAMLIdentity savedIdentity = secCtx.getCurrentSAMLIdentity();
            SAMLIdentity[] identities = secCtx.getSAMLIdentities();
            for (int i = 0; i < identities.length + 1; i++) {

                if (i < identities.length) {
                    // the identity to query:
                    secCtx.setCurrentSAMLIdentity(identities[i]);
                } else {
                    secCtx.setCurrentSAMLIdentity(null);
                }

                // re-initialize interceptors?

                // TODO: the next two PIP invocations are an atomic operation

                // invoke SAMLQueryPIP:
                assert (queryPIP != null);
                try {
                    queryPIP.collectAttributes(peerSubject, context);
                } catch (Exception e) {
                    logger.error("SAMLQueryPIP attribute collection " +
                                 "failed", e);
                    //return PDPDecision.INDETERMINATE;
                    continue;
                }
                GS4GTSecurityContextLogger.log(peerSubject);

                // invoke AttributeAcceptancePIP:
                assert (aapPIP != null);
                try {
                    aapPIP.collectAttributes(peerSubject, context);
                } catch (Exception e) {
                    logger.error("AttributeAcceptancePIP attribute collection " +
                                 "failed", e);
                    //return PDPDecision.INDETERMINATE;
                }
                GS4GTSecurityContextLogger.log(peerSubject);

                // invoke SAMLBlacklistPDP:
                assert (blPDP != null);
                try {
                    decisionCode =
                        blPDP.isPermitted(peerSubject, context, op);
                } catch (Exception e) {
                    logger.error("SAMLBlacklistPDP decision making failed", e);
                    return PDPDecision.INDETERMINATE;
                }
                assert (decisionCode != PDPDecision.PERMIT);
                if (decisionCode == PDPDecision.DENY) {
                    logger.info("SAMLBlacklistPDP returned DENY");
                    logger.info("GS4GT.PullPDPImpl short-circuited");
                    return PDPDecision.DENY;
                }

                // invoke SAMLAttributePDP:
                try {
                    decisionCode =
                        samlPDP.isPermitted(peerSubject, context, op);
                } catch (Exception e) {
                    logger.error("SAMLAttributePDP decision making failed", e);
                    //return PDPDecision.INDETERMINATE;
                    continue;
                }
                //assert (decisionCode != PDPDecision.DENY);

                if (decisionCode == PDPDecision.PERMIT) {
                    logger.info("SAMLAttributePDP returned PERMIT");
                    logger.info("GS4GT.PullPDPImpl short-circuited");
                    break;
                }
            }
            secCtx.setCurrentSAMLIdentity(savedIdentity);

            // invoke SAMLMapPIP:
            assert (mapPIP != null);
            try {
                mapPIP.collectAttributes(peerSubject, context);
            } catch (Exception e) {
                logger.error("error collecting attributes", e);
            }
            GS4GTSecurityContextLogger.log(peerSubject);

            return decisionCode;
        }
    }

    /**
     * Combined interceptor for attribute push/pull
     * that implements the GridShib Attribute Aggregation Algorithm.
     *
     * @see org.globus.gridshib.gt.authorization.BasePDP
     * @see org.globus.gridshib.gt.authorization.GridShibConfigParams
     */
    public static class PushPullPDPImpl extends BasePDP
                                     implements GridShibConfigParams {

        public void initialize(HashMap configs, String name) throws Exception {

            logger.info("GS4GT.PushPullPDPImpl initializing in scope " + name);

            if (configs == null) {
                String msg = "HashMap configs is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("configs = " + configs.toString());

            if (this.isInitialized(name)) {
                logger.info("GS4GT.PushPullPDPImpl previously initialized " +
                            "in scope " + name);
                logger.info(PUSH_PULL_INITIALIZATION_MSG);
                return;
            }

            try {
                prePolicyPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize GS4GT.PrePolicyPDPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                samlPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLAttributePDPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                pullPDP.initialize(configs, name);
            } catch (Exception e) {
                logger.error("Unable to initialize GS4GT.PullPDPImpl", e);
                throw e;
            }

            super.initialize(configs, name);
            logger.info(PUSH_PULL_INITIALIZATION_MSG);
        }

        /**
         * This PDP implementation orchestrates the
         * following specific workflow:
         * <ol>
         *   <li>Invoke <code>GS4GT.PrePolicyPDPImpl</code>
         *     <ul>
         *       <li>If <code>PERMIT</code> or <code>DENY</code>,
         *       return <code>PERMIT</code> or <code>DENY</code>
         *       (resp.); otherwise, continue</li>
         *     </ul>
         *   <li>Invoke <code>SAMLAttributePDPImpl</code>
         *     <ul>
         *       <li>If <code>PERMIT</code>, return <code>PERMIT</code>;
         *       otherwise, continue</li>
         *     </ul>
         *   <li>Invoke <code>GS4GT.PullPDPImpl</code>
         *   <li>Return <code>GS4GT.PullPDPImpl</code> decision
         * </ol>
         * If this PDP reaches the end of the authorization chain
         * without returning <code>PERMIT</code> or <code>DENY</code>,
         * it returns whatever <code>GS4GT.PullPDPImpl</code>
         * returns.
         */
        public int isPermitted(Subject peerSubject,
                               MessageContext context,
                               QName operation) throws Exception {

            if (peerSubject == null) {
                String msg = "Subject is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("peerSubject = " + peerSubject.toString());
            if (context == null) {
                String msg = "MessageContext is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("context = " + context.toString());
            logger.debug("operation = " + operation.toString());

            String peerIdentity =
                GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
            logger.info("GS4GT.PushPullPDPImpl making decision " +
                        "for identity " + peerIdentity);

            int decisionCode;

            // invoke GS4GT.PrePolicyPDPImpl:
            try {
                decisionCode =
                    prePolicyPDP.isPermitted(peerSubject, context, operation);
            } catch (Exception e) {
                String msg = "GS4GT.PrePolicyPDPImpl decision making failed";
                logger.error(msg, e);
                return PDPDecision.INDETERMINATE;
            }
            if (decisionCode == PDPDecision.PERMIT) {
                logger.info("GS4GT.PrePolicyPDPImpl returned PERMIT");
                logger.info("GS4GT.PushPullPDPImpl short-circuited");
                logger.info(PUSH_PULL_COMPLETION_MSG);
                return PDPDecision.PERMIT;
            } else if (decisionCode == PDPDecision.DENY) {
                logger.info("GS4GT.PrePolicyPDPImpl returned DENY");
                logger.info("GS4GT.PushPullPDPImpl short-circuited");
                logger.info(PUSH_PULL_COMPLETION_MSG);
                return PDPDecision.DENY;
            }

            // invoke SAMLAttributePDPImpl:
            try {
                decisionCode =
                    samlPDP.isPermitted(peerSubject, context, operation);
            } catch (Exception e) {
                String msg = "SAMLAttributePDPImpl decision making failed";
                logger.error(msg, e);
                return PDPDecision.INDETERMINATE;
            }
            if (decisionCode == PDPDecision.PERMIT) {
                logger.info("SAMLAttributePDPImpl returned PERMIT");
                logger.info("GS4GT.PushPullPDPImpl short-circuited");
                logger.info(PUSH_PULL_COMPLETION_MSG);
                return PDPDecision.PERMIT;
            }

            // invoke GS4GT.PullPDPImpl:
            assert (pullPDP != null);
            try {
                decisionCode =
                    pullPDP.isPermitted(peerSubject, context, operation);
            } catch (Exception e) {
                logger.error("GridShibPullPDP decision making failed", e);
                return PDPDecision.INDETERMINATE;
            }

            logger.info("GS4GT.PushPullPDPImpl returned " +
                        PDPDecision.getString(decisionCode) + " decision");
            logger.info(PUSH_PULL_COMPLETION_MSG);

            return decisionCode;
        }
    }

    /**
     * Combined interceptor for pre-policy GridShib attribute
     * push, that is, this interceptor does not invoke
     * <code>SAMLAttributePDPImpl</code> at the end of the
     * authorization chain.
     *
     * @see org.globus.gridshib.gt.authorization.BasePDP
     * @see org.globus.gridshib.gt.authorization.GridShibConfigParams
     */
    private static class PrePolicyPDPImpl extends BasePDP
                                       implements GridShibConfigParams {

        private static GridMap gridmap = null;

        public void initialize(HashMap configs, String name) throws Exception {

            logger.info("GS4GT.PrePolicyPDPImpl initializing in scope " + name);

            if (configs == null) {
                String msg = "HashMap configs is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("configs = " + configs.toString());

            if (this.isInitialized(name)) {
                logger.info("GS4GT.PrePolicyPDPImpl previously initialized " +
                            "in scope " + name);
                logger.info(PRE_POLICY_INITIALIZATION_MSG);
                return;
            }

            try {
                pushPIP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLAssertionPushPIPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                aapPIP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize AttributeAcceptancePIPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                blPDP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLBlacklistPDPImpl";
                logger.error(msg, e);
                throw e;
            }

            try {
                mapPIP.initialize(configs, name);
            } catch (Exception e) {
                String msg = "Unable to initialize SAMLMapPIPImpl";
                logger.error(msg, e);
                throw e;
            }

            Object useGridmap = configs.get(CONSULT_GRIDMAP_KEY);
            if (useGridmap == null) {
                logger.info("Gridmap short-circuiting is disabled by default");
            } else {
                if (((String)useGridmap).trim().equalsIgnoreCase("true")) {
                    String msg = "Gridmap short-circuiting is enabled";
                    Object gridmapObj = configs.get(DEFAULT_GRIDMAP);
                    if (gridmapObj == null) {
                        msg += " but no gridmap file is configured";
                        logger.warn(msg);
                    } else {
                        try {
                            gridmap = (GridMap)gridmapObj;
                            logger.info(msg);
                        } catch (ClassCastException e) {
                            msg += " but the object provided is not of " +
                                   "type GridMap (programmer error)";
                            logger.error(msg, e);
                            throw e;
                        }
                    }
                } else {
                    logger.info("Gridmap short-circuiting is disabled");
                }
            }

            super.initialize(configs, name);
            logger.info(PRE_POLICY_INITIALIZATION_MSG);
        }

        /**
         * This PDP implementation orchestrates the
         * following specific workflow:
         * <ol>
         *   <li>Invoke <code>SAMLAssertionPushPIPImpl</code>
         *   <li>Invoke <code>AttributeAcceptancePIPImpl</code>
         *   <li>Invoke <code>SAMLBlacklistPDPImpl</code>
         *     <ul>
         *       <li>If <code>DENY</code>, return <code>DENY</code>;
         *       otherwise, continue</li>
         *     </ul>
         *   <li>Invoke <code>SAMLMapPIPImpl</code>
         *   <li>Invoke GridMap short-circuiting
         *     <ul>
         *       <li>If <code>PERMIT</code>, return <code>PERMIT</code>;
         *       otherwise, continue</li>
         *     </ul>
         *   <li>Invoke VOMS short-circuiting
         *     <ul>
         *       <li>If <code>PERMIT</code>, return <code>PERMIT</code>;
         *       otherwise, continue</li>
         *    </ul>
         * </ol>
         * If this PDP reaches the end of the authorization chain
         * without returning <code>PERMIT</code> or <code>DENY</code>,
         * it returns <code>INDETERMINATE</code>.
         */
        public int isPermitted(Subject peerSubject,
                               MessageContext context,
                               QName operation) throws Exception {

            if (peerSubject == null) {
                String msg = "Subject is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("peerSubject = " + peerSubject.toString());
            if (context == null) {
                String msg = "MessageContext is null";
                throw new IllegalArgumentException(msg);
            }
            logger.debug("context = " + context.toString());
            logger.debug("operation = " + operation.toString());

            String peerIdentity =
                GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
            logger.info("GS4GT.PrePolicyPDPImpl making decision " +
                        "for identity " + peerIdentity);

            // TODO: the next two PIP invocations are an atomic operation

            // invoke SAMLAssertionPushPIP:
            assert (pushPIP != null);
            try {
                pushPIP.collectAttributes(peerSubject, context);
            } catch (Exception e) {
                logger.error("SAMLAssertionPushPIP attribute collection " +
                             "failed", e);
                return PDPDecision.INDETERMINATE;
            }
            GS4GTSecurityContextLogger.log(peerSubject);

            // invoke AttributeAcceptancePIP:
            assert (aapPIP != null);
            try {
                aapPIP.collectAttributes(peerSubject, context);
            } catch (Exception e) {
                logger.error("AttributeAcceptancePIP attribute collection " +
                             "failed", e);
                return PDPDecision.INDETERMINATE;
            }
            GS4GTSecurityContextLogger.log(peerSubject);

            int decisionCode;

            // invoke SAMLBlacklistPDP:
            assert (blPDP != null);
            try {
                decisionCode =
                    blPDP.isPermitted(peerSubject, context, operation);
            } catch (Exception e) {
                logger.error("SAMLBlacklistPDP decision making failed", e);
                return PDPDecision.INDETERMINATE;
            }
            assert (decisionCode != PDPDecision.PERMIT);
            if (decisionCode == PDPDecision.DENY) {
                logger.info("SAMLBlacklistPDP returned DENY");
                logger.info("GS4GT.PrePolicyPDPImpl short-circuited");
                logger.info(PRE_POLICY_COMPLETION_MSG);
                return PDPDecision.DENY;
            }

            // invoke SAMLMapPIP:
            assert (mapPIP != null);
            try {
                mapPIP.collectAttributes(peerSubject, context);
            } catch (Exception e) {
                logger.error("error collecting attributes", e);
                return PDPDecision.INDETERMINATE;
            }
            GS4GTSecurityContextLogger.log(peerSubject);

            // TODO: covert this code block into a unit test
            //
            // Config parameter requireAuthzMap has been removed.
            //if (this.requireAuthzMap) {
            //    String[] usernames = null;
            //    try {
            //        usernames = SecurityManager.getManager().getLocalUsernames();
            //    } catch (SecurityException e) {
            //        logger.error("Unable to get local usernames", e);
            //        return PDPDecision.INDETERMINATE;
            //    }
            //    if ((usernames == null) || (usernames.length < 1)) {
            //        logger.error("No user name mappings found");
            //        return PDPDecision.INDETERMINATE;
            //    }
            //}

            GS4GTSecurityContext secCtx =
                GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
            assert (secCtx != null);

            if (gridmap != null) {
                logger.debug("GridMap short-circuiting invoked");
                String[] accounts = gridmap.getUserIDs(peerIdentity);
                // at least one user mapping is required:
                if (accounts != null && accounts.length > 0) {
                    logger.info("GridMap short-circuiting succeeded");
                    for (int i = 0; i < accounts.length; i++) {
                        logger.debug("Adding user account " + accounts[i] +
                                     " to security context");
                        secCtx.addPrincipal(new UserNamePrincipal(accounts[i]));
                    }
                    GS4GTSecurityContextLogger.log(peerSubject);
                    return PDPDecision.PERMIT;
                } else {
                    logger.warn("GridMap short-circuiting failed: " +
                                "no mapped usernames");
                }
            }

            // TODO: implement VOMS short-circuiting

            logger.info(PRE_POLICY_COMPLETION_MSG);
            return PDPDecision.INDETERMINATE;
        }
    }
}
