/*
 * Copyright 2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

public interface GridShibConfigParams {

    /**
     * The absolute path to a directory of metadata files.
     * <p>
     * In class <code>SAMLAssertionPushPIPImpl</code>,
     * the term <em>metadata file</em> refers to a
     * directory of one or more name mapping files.
     * At initialization time, the metadata path is
     * passed to the <code>GridShibEntityMapper</code>.
     * <p>
     * In class <code>SAMLQueryPIPImpl</code>, the term
     * <em>metadata file</em> refers to a SAML metadata
     * file.  At initialization time, all the metadata
     * files are preprocessed for any
     * <code>AttributeAuthority</code> entities they
     * might contain.
     * <p>
     * This configuration parameter is required.  It has
     * no default value.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     * @see org.globus.gridshib.gt.authorization.SAMLAssertionPushPIPImpl
     * @see org.globus.gridshib.gt.authorization.attributes.AttributeAuthority
     * @see org.globus.gridshib.common.mapper.GridShibEntityMapper
     */
    public static final String METADATA_PATH_KEY =
        "metadataPath";

    /**
     * Since this configuration parameter is confined to
     * <code>ClassicGridShibPDP</code>, it is no longer
     * exposed to end users.
     *
     * @see org.globus.gridshib.gt.authorization.ClassicGridShibPDPImpl
     *
     * @deprecated As of v0.6.0, this configuration property,
     *             which is exclusive to the
     *             <code>ClassicGridShibPDP</code> class, is
     *             deprecated.
     */
    public static final String RESPECT_KEY =
        "respectMDAttributes";

    /**
     * A <code>GridMap</code> object used for Gridmap
     * short-circuiting, which is set in the
     * <code>BasePDP</code> abstract class.
     * <p>
     * In the specific case of GT&nbsp;4.0 PDPs, the GridShib
     * authorization framework provides a <code>GridMap</code>
     * object obtained from the underlying GT&nbsp;4.0
     * configuration.  This <code>GridMap</code> object is
     * made available to PDP implementations via an internal
     * (i.e., not exposed) configuration parameter.
     * <p>
     * Since this parameter corresponds to a Java
     * object, it can't be set in a configuration
     * file and therefore should not be exposed to
     * end users.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.GS4GT.PrePolicyPDPImpl
     * @see org.globus.gridshib.gt.authorization.ClassicGridShibPDPImpl
     * @see org.globus.gridshib.BasePDP
     * @see org.globus.security.gridmap.GridMap
     */
    public static final String DEFAULT_GRIDMAP =
        "defaultGridmap";

    /**
     * A boolean parameter that enables gridmap
     * short-circuiting in GT&nbsp;4.0 deployments.
     * Gridmap short-circuiting is enabled if and only
     * if this parameter is set to true.  Gridmap
     * short-circuiting is disabled by default.
     * <p>
     * This parameter has no effect in GT&nbsp;4.2.
     * If this parameter is configured in a GT&nbsp;4.2
     * deployment, a warning is logged.
     * For GT&nbsp;4.2 deployments, you can configure a
     * gridmap authorization module ahead of the GridShib
     * interceptors to achieve gridmap short-circuiting.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.GS4GT.PrePolicyPDPImpl
     * @see org.globus.gridshib.gt.authorization.ClassicGridShibPDPImpl
     */
    public static final String CONSULT_GRIDMAP_KEY =
        "enableGridmap";

    /**
     * This configuration parameter is not yet
     * implemented in GS4GT&nbsp;v0.6.0.
     *
     * @see org.globus.gridshib.gt.authorization.GS4GT.PrePolicyPDPImpl
     */
    public static final String USE_VOMS =
        "useVOMS";

    /**
     * A boolean parameter that indicates whether
     * or not blacklisting is enabled.  This parameter
     * is programmatically set to true if and only if
     * there is at least one configured blacklist file.
     * <p>
     * If this parameter is set in the configuration
     * file, it is ignored.  Therefore, this parameter
     * should not be exposed to end users..
     *
     * @see org.globus.gridshib.gt.authorization.SAMLBlacklistPDPImpl
     *
     * @since 0.6.0
     */
    public static final String ENABLE_BLACKLISTING_KEY =
        "enableBlacklisting";

    /**
     * The absolute path to a file of blacklisted IP addresses.
     * If blacklisting is not enabled, this configuration
     * parameter has no effect.
     * <p>
     * At least one of <code>BLACKLIST_IP_ADDRESSES_FILE_KEY</code>
     * or <code>BLACKLIST_NAME_IDS_FILE_KEY</code> is required.
     * This configuration parameter has no default value.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLBlacklistPDPImpl
     *
     * @since 0.6.0
     */
    public static final String BLACKLIST_IP_ADDRESSES_FILE_KEY =
        "blacklistIPAddressesFile";

    /**
     * The absolute path to a file of blacklisted name identifiers.
     * If blacklisting is not enabled, this configuration
     * parameter has no effect.
     * <p>
     * At least one of <code>BLACKLIST_IP_ADDRESSES_FILE_KEY</code>
     * or <code>BLACKLIST_NAME_IDS_FILE_KEY</code> is required.
     * This configuration parameter has no default value.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLBlacklistPDPImpl
     *
     * @since 0.6.0
     */
    public static final String BLACKLIST_NAME_IDS_FILE_KEY =
        "blacklistNameIdentifiersFile";

    /**
     * A boolean parameter that indicates whether
     * or not attribute-based mapping is enabled.
     * This parameter is programmatically set to
     * true if and only if a policy file is configured.
     * <p>
     * If this parameter is set in the configuration
     * file, it is ignored.  Therefore, this parameter
     * should not be exposed to end users..
     *
     * @see org.globus.gridshib.gt.authorization.SAMLMapPIPImpl
     *
     * @since 0.6.0
     */
    public static final String ENABLE_ATTRIBUTE_MAPPING_KEY =
        "enableAttributeMapping";

    /**
     * This configuration parameter corresponds to
     * a <code>ShibbolethPDPPolicy</code> instance.
     * If this instance is non-null, initialization
     * of the <code>SAMLMapPIPImpl</code> interceptor
     * is completely short-circuited.
     * <p>
     * Since this parameter corresponds to a Java
     * object, it can't be set in a configuration
     * file and therefore should not be exposed to
     * end users.  Use of this configuration parameter
     * is optional.
     * <p>
     * Note: This parameter appears to be unused at
     * this time.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLMapPIPImpl
     * @see org.globus.gridshib.gt.authorization.ShibbolethPDPPolicy
     */
    public static final String SAML_MAP_POLICY_KEY =
        "samlMapPolicy";

    /**
     * The absolute path to a GridShib policy file.
     * This policy file is used by
     * <code>SAMLMapPIPImpl</code> to map attributes
     * to usernames.
     * <p>
     * If a <code>ShibbolethPDPPolicy</code> instance
     * is provided via <code>SAML_MAP_POLICY_KEY</code>,
     * the check for a policy file is short-circuited.
     * Otherwise the policy file itself gives rise to a
     * <code>ShibbolethPDPPolicy</code> object.
     * <p>
     * Use of this configuration parameter is optional.
     * If it is omitted, the implementation falls back
     * on <code>ATTR_FILE_KEY</code>.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLMapPIPImpl
     * @see org.globus.gridshib.gt.authorization.ShibbolethPDPPolicy
     */
    public static final String MAP_FILE_KEY =
        "attributeMappingPolicyFile";

    /**
     * A boolean parameter that indicates whether
     * or not attribute-based authorization is enabled.
     * This parameter is programmatically set to
     * true if and only if an authz policy file is
     * configured.
     * <p>
     * If this parameter is set in the configuration
     * file, it is ignored.  Therefore, this parameter
     * should not be exposed to end users..
     *
     * @see org.globus.gridshib.gt.authorization.SAMLAttributePDPImpl
     *
     * @since 0.6.0
     */
    public static final String ENABLE_ATTRIBUTE_AUTHZ_KEY =
        "enableAttributeAuthz";

    /**
     * This configuration parameter corresponds to
     * a <code>ShibbolethPDPPolicy</code> object.
     * If this instance is non-null, further
     * initialization of the interceptor
     * is short-circuited.  Basically what that
     * means is that the interceptor does not
     * attempt to load a GridShib policy file,
     * regardless of whether or not one is
     * specified in the configuration file.
     * <p>
     * Since this parameter corresponds to a Java
     * object, it can't be set in a configuration
     * file and therefore should not be exposed to
     * end users.  Use of this configuration parameter
     * is optional.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLAttributePDPImpl
     * @see org.globus.gridshib.gt.authorization.SAMLMapPIPImpl
     * @see org.globus.gridshib.gt.authorization.ClassicGridShibPDPImpl
     */
    public static final String SHIB_PDP_POLICY_KEY =
        "gridshibPDPPolicy";

    /**
     * The absolute path to a GridShib policy file.
     * This policy file is used by
     * <code>SAMLAttributePDPImpl</code> to render
     * an access control decision.  It also used by
     * <code>SAMLMapPIPImpl</code> in the event
     * that <code>MAP_FILE_KEY</code> is unspecified.
     * In both cases (and only these cases), use of
     * this configuration parameter is required.
     * It has no default value.
     * <p>
     * If a <code>ShibbolethPDPPolicy</code> instance
     * is provided via <code>SHIB_PDP_POLICY_KEY</code>,
     * the check for a policy file is short-circuited.
     * Otherwise the policy file itself gives rise to a
     * <code>ShibbolethPDPPolicy</code> object.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLAttributePDPImpl
     * @see org.globus.gridshib.gt.authorization.SAMLMapPIPImpl
     * @see org.globus.gridshib.gt.authorization.ClassicGridShibPDPImpl
     * @see org.globus.gridshib.gt.authorization.ShibbolethPDPPolicy
     */
    public static final String ATTR_FILE_KEY =
        "attributeAuthzPolicyFile";

    /**
     * A boolean parameter that determines whether
     * or not an attribute query is enabled.  Query
     * is enabled if and only if this parameter is
     * set to true, that is, attribute query is
     * disabled by default.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     *
     * @since 0.6.0
     */
    public static final String ENABLE_ATTRIBUTE_QUERY_KEY =
        "enableAttributeQuery";

    /**
     * This configuration parameter corresponds to
     * an <code>AttributeAuthority</code> object.
     * If this instance is non-null, the
     * <code>SAMLQueryPIPImpl</code> does not attempt
     * to load a metadata file, that is, it ignores
     * the value of <code>METADATA_PATH_KEY</code>
     * (if any).
     * <p>
     * If an <code>AttributeAuthority</code> instance
     * is provided via <code>AA_INSTANCE_KEY</code>,
     * the check for SAML metadata is short-circuited.
     * Otherwise the metadata files give rise to an
     * array of <code>AttributeAuthority</code> objects.
     * <p>
     * Since this parameter corresponds to a Java
     * object, it can't be set in a configuration
     * file and therefore should not be exposed to
     * end users.  Use of this configuration parameter
     * is optional.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     * @see org.globus.gridshib.gt.authorization.attributes.AttributeAuthority
     *
     * @since 0.6.0
     */
    public static final String AA_INSTANCE_KEY =
        "AAInstance";

    /**
     * The <code>entityID</code> of the IdP to query by default.
     * The <code>SAMLQueryPIPImpl</code> will attempt to conduct
     * a default query if there is insufficient information in
     * the security context on which to base a query.
     * Use of this configuration parameter is optional.
     * <p>
     * After the <code>AttributeAuthority</code> objects have
     * been obtained, either by <code>AA_INSTANCE_KEY</code>
     * or <code>METADATA_PATH_KEY</code>, a specific
     * <code>AttributeAuthority</code> object corresponding to
     * <code>entityID</code> is determined and may be later
     * used to conduct a default query.
     * If <code>IDP_PROVIDER_ID_KEY</code> is null or no
     * <code>AttributeAuthority</code> object corresponding to
     * the <code>entityID</code> is found, the
     * <code>SAMLQueryPIPImpl</code> will not be able to
     * conduct a default query, that is, the
     * <code>SAMLQueryPIPImpl</code> will only query based on
     * information found in the security context.
     * <p>
     * This configuration parameter is used in the system tests.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     */
    public static final String IDP_PROVIDER_ID_KEY =
        "defaultIDPEntityID";

    /**
     * The <code>entityID</code> of this attribute requester.
     * If attribute query is enabled at the container level,
     * this is a globally unique identifier for the container
     * as well as every service running in the container.
     * If attribute query is enabled at the service level,
     * this is a globally unique identifier for the service.
     * If attribute query is enabled at both the container
     * level and the service level, the <code>entityID</code>
     * of the latter takes precedence.
     * <p>
     * This configuration parameter is required.  It has
     * no default value.
     * <p>
     * This configuration parameter is used in the system tests.
     * <p>
     * Note: This configuration parameter is incorrectly
     * focused on the attribute requester component of the
     * Grid SP.  A future version of GS4GT will require an
     * <code>entityID</code> for each and every instance of
     * the <code>GridShibPDP</code>.  Moreover, every
     * instance of the <code>GridShibPDP</code> will have
     * SAML metadata associated with it.  Thus the
     * <code>entityID</code> MUST correspond to an entity
     * in metadata.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     */
    public static final String SP_PROVIDER_ID_KEY =
        "gridSPEntityID";

    /**
     * A boolean parameter that determines whether
     * or not host-based authorization is performed.
     * <p>
     * This configuration parameter is optional.  If omitted,
     * it defaults to false, that is, host-based authorization
     * is <strong>not</strong> performed.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     *
     * @deprecated this configuration parameter may be removed
     *             in a future version of GS4GT
     */
    public static final String AUTHZ_HOST_KEY =
        "AAHostAuthz";

    /**
     * Note: Caching in <code>SAMLMapPIPImpl</code> is
     * broken, so the only utilization of caching is
     * in <code>SAMLQueryPIPImpl</code> and therefore
     * this parameter need not be exposed in
     * GS4GT&nbsp;v0.6.0.
     *
     * @see org.globus.gridshib.gt.authorization.SAMLQueryPIPImpl
     * @see org.globus.gridshib.gt.authorization.SAMLMapPIPImpl
     * @see org.globus.gridshib.gt.authorization.ClassicGridShibPDPImpl
     */
    public static final String CACHE_KEY =
        "cacheAttributes";

    /**
     * This array of <code>String</code> is used for
     * config translation to Hashtable in PDP proxies,
     * which allows us to adapt the code to multiple
     * authorization framework implementations.
     * Anytime you add a configuration key to this
     * interface, make sure to add it to this array as well.
     */
    public static final String[] ALL_CONFIG_KEYS = {
        METADATA_PATH_KEY,
        RESPECT_KEY,
        DEFAULT_GRIDMAP,
        CONSULT_GRIDMAP_KEY,
        USE_VOMS,
        ENABLE_BLACKLISTING_KEY,
        BLACKLIST_IP_ADDRESSES_FILE_KEY,
        BLACKLIST_NAME_IDS_FILE_KEY,
        ENABLE_ATTRIBUTE_MAPPING_KEY,
        SAML_MAP_POLICY_KEY,
        MAP_FILE_KEY,
        ENABLE_ATTRIBUTE_AUTHZ_KEY,
        SHIB_PDP_POLICY_KEY,
        ATTR_FILE_KEY,
        ENABLE_ATTRIBUTE_QUERY_KEY,
        AA_INSTANCE_KEY,
        IDP_PROVIDER_ID_KEY,
        SP_PROVIDER_ID_KEY,
        AUTHZ_HOST_KEY,
        CACHE_KEY
    };
}

