/*
 * Copyright 2007-2008 University of Illinois
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization;

import java.io.File;
import java.security.Principal;
import java.util.HashMap;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gridshib.common.LoadException;
import org.globus.gridshib.common.StringSetFile;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.gt.security.SAMLPrincipal;
import org.globus.gridshib.gt.security.util.GridShibSecurityUtil;
import org.globus.gridshib.security.SAMLAuthnContext;

import org.globus.gsi.jaas.GlobusPrincipal;

/**
 * A blacklisting PDP implementation that backs
 * <code>SAMLBlacklistPDP</code>.
 *
 * @see org.globus.gridshib.SAMLBlacklistPDP
 */
public class SAMLBlacklistPDPImpl extends BasePDP
                               implements GridShibConfigParams {

    static Log logger =
        LogFactory.getLog(SAMLBlacklistPDPImpl.class.getName());

    private static final String INITIALIZATION_MSG =
        "SAMLBlacklistPDPImpl initialization complete";
    private static final String COMPLETION_MSG =
        "SAMLBlacklistPDPImpl decision making complete";

    private boolean isBlacklistingEnabled;
    private StringSetFile blacklistIPAddrs = null;
    private StringSetFile blacklistNameIDs = null;

    /**
     * Loads one or more blacklists.  If no blacklist
     * is configured, this method disables blacklisting.
     *
     * @param configs a (non-null) map of configuration
     *        parameters
     * @param name the scope prefix associated with this
     *        PDP implementation
     *
     * @exception java.lang.IllegalArgumentException
     *            If the <code>configs</code> argument is null
     * @exception java.lang.RuntimeException
     *            If blacklisting is enabled but no blacklist
     *            is configured or the path to the blacklist
     *            is not an absolute path
     * @exception org.globus.gridshib.common.LoadException
     *            If unable to get a <code>StringSetFile</code>
     *            instance, presumably because the file resource
     *            could not be loaded
     */
    public void initialize(HashMap configs, String name) throws Exception {

        logger.info("SAMLBlacklistPDPImpl initializing in scope " + name);

        if (configs == null) {
            String msg = "HashMap configs is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("configs = " + configs.toString());

        if (this.isInitialized(name)) {
            logger.info("SAMLBlacklistPDPImpl previously initialized " +
                        "in scope " + name);
            logger.info(INITIALIZATION_MSG);
            return;
        }

        String blacklistPath = null;

        // process blacklist of IP addresses:
        blacklistPath =
            (String)configs.get(BLACKLIST_IP_ADDRESSES_FILE_KEY);
        if (blacklistPath != null) {
            logger.info("IP address blacklist: " + blacklistPath);
            File blacklistFile = new File(blacklistPath);
            if (!blacklistFile.isAbsolute()) {
                String msg =
                    "Path (" + blacklistPath + ") is not absolute";
                logger.error(msg);
                throw new RuntimeException(msg);
            }
            this.blacklistIPAddrs =
                StringSetFile.getInstance(blacklistPath);
        } else {
            logger.info("No IP address blacklist found");
        }

        // process blacklist of name identifiers:
        blacklistPath =
            (String)configs.get(BLACKLIST_NAME_IDS_FILE_KEY);
        if (blacklistPath != null) {
            logger.info("Name ID blacklist: " + blacklistPath);
            File blacklistFile = new File(blacklistPath);
            if (!blacklistFile.isAbsolute()) {
                String msg =
                    "Path (" + blacklistPath + ") is not absolute";
                logger.error(msg);
                throw new RuntimeException(msg);
            }
            this.blacklistNameIDs =
                StringSetFile.getInstance(blacklistPath);
        } else {
            logger.info("No name ID blacklist found");
        }

        if (blacklistIPAddrs == null && blacklistNameIDs == null) {
            this.isBlacklistingEnabled = false;
            configs.put(ENABLE_BLACKLISTING_KEY, "false");
            logger.info("Blacklisting is implicitly disabled");
        } else {
            this.isBlacklistingEnabled = true;
            configs.put(ENABLE_BLACKLISTING_KEY, "true");
            logger.info("Blacklisting is implicitly enabled");
        }

        super.initialize(configs, name);
        logger.info(INITIALIZATION_MSG);
    }

    /**
     * If blacklisting is not enabled, this method short circuits
     * and returns <code>PDPDecision.NOT_APPLICABLE</code>.
     * If blacklisting is enabled, this method returns
     * <code>PDPDecision.DENY</code> if any of the following
     * conditions are met:
     * <ul>
     *   <li>the subject's security context contains a
     *   <code>SAMLAuthnContext</code> instance that asserts an
     *   IP address that is on the blacklist of IP addresses</li>
     *   <li>the subject's security context contains a
     *   <code>SAMLPrincipal</code> instance
     *   that is on the blacklist of name identifiers</li>
     *   <li>the subject's security context contains a
     *   <code>GlobusPrincipal</code> instance
     *   that is on the blacklist of name identifiers</li>
     * </ul>
     * If none of the above conditions is satisfied, this
     * method returns <code>PDPDecision.INDETERMINATE</code>.
     * Note that this method <strong>never</strong> returns
     * <code>PDPDecision.PERMIT</code>.
     * <p>
     * A blacklist is a dynamic file-based resource.
     * Every time the file is accessed, this method determines
     * if the file was modified since it was last loaded.
     * If the file is dirty, this method reloads the file
     * before accessing it.  If the file can not be reloaded
     * (because of a syntax error, e.g.), this method logs
     * an error message and returns
     * <code>PDPDecision.INDETERMINATE</code>.
     *
     * @param peerSubject the authenticated subject
     * @param context the message context
     * @param operation the requested operation
     *
     * @exception java.lang.IllegalArgumentException
     *            If either the <code>peerSubject</code> argument
     *            or the <code>context</code> argument is null
     */
    public int isPermitted(Subject peerSubject,
                           MessageContext context,
                           QName operation) throws Exception {

        if (peerSubject == null) {
            String msg = "Subject is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("peerSubject = " + peerSubject.toString());
        if (context == null) {
            String msg = "MessageContext is null";
            throw new IllegalArgumentException(msg);
        }
        logger.debug("context = " + context.toString());
        logger.debug("operation = " + operation.toString());

        String peerIdentity =
            GridShibSecurityUtil.getIdentityFromSubject(peerSubject);
        logger.info("SAMLBlacklistPDPImpl making decision " +
                    "for identity " + peerIdentity);

        if (this.isBlacklistingEnabled) {
            logger.info("Blacklisting is enabled");
        } else {
            logger.info("Blacklisting is disabled");
            return PDPDecision.NOT_APPLICABLE;
        }
        assert (this.isBlacklistingEnabled);

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);

        boolean ipAddressFound = false;
        if (blacklistIPAddrs != null) {
            logger.debug("Searching blacklist of IP addresses");
            SAMLAuthnContext[] authnCtxs = secCtx.getSAMLAuthnContexts();
            assert (authnCtxs != null);
            logger.debug("Found " + authnCtxs.length + " authn context" +
                         ((authnCtxs.length == 1) ? "" : "s"));
            try {
                for (int i = 0; i < authnCtxs.length; i++) {
                    String ipAddress = authnCtxs[i].getIPAddress();
                    if (this.blacklistIPAddrs.contains(ipAddress)) {
                        logger.info("Blacklisted IP address found: " +
                                    ipAddress);
                        ipAddressFound = true;
                        break;
                    } else {
                        logger.info("IP address " + ipAddress +
                                    " not blacklisted");
                    }
                }
            } catch (LoadException e) {
                String msg = "Unable to reload IP address blacklist";
                logger.error(msg, e);
                return PDPDecision.INDETERMINATE;
            }
        }

        boolean nameIDFound = false;
        if (blacklistNameIDs != null) {
            logger.debug("Searching blacklist of principal name identifiers");
            Principal[] principals = secCtx.getSAMLPrincipals();
            assert (principals != null);
            logger.debug("Found " + principals.length + " SAML principal" +
                         ((principals.length == 1) ? "" : "s"));
            try {
                for (int i = 0; i < principals.length; i++) {
                    String name = principals[i].getName();
                    if (this.blacklistNameIDs.contains(name)) {
                        logger.info("Blacklisted SAML principal name " +
                                    "found: " + name);
                        nameIDFound = true;
                        break;
                    } else {
                        logger.info("SAML principal name " + name +
                                    " not blacklisted");
                    }
                }
                if (!nameIDFound && secCtx.getGlobusPrincipal() != null) {
                    String name = secCtx.getGlobusPrincipal().getName();
                    if (this.blacklistNameIDs.contains(name)) {
                        logger.info("Blacklisted Globus principal name " +
                                    "found: " + name);
                        nameIDFound = true;
                    } else {
                        logger.info("Globus principal name " + name +
                                    " not blacklisted");
                    }
                }
            } catch (LoadException e) {
                String msg = "Unable to reload principal name blacklist";
                logger.error(msg, e);
                return PDPDecision.INDETERMINATE;
            }
        }

        int decisionCode;
        if (ipAddressFound || nameIDFound) {
            logger.info("SAMLBlacklistPDPImpl returned DENY");
            decisionCode = PDPDecision.DENY;
        } else {
            logger.info("SAMLBlacklistPDPImpl returned INDETERMINATE");
            decisionCode = PDPDecision.INDETERMINATE;
        }
        logger.info(COMPLETION_MSG);

        return decisionCode;
    }
}
