/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization.attributes;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;

// SAMLAttribute + time validity for storage
public class SAMLAttribute {
    private String name;  // <saml:Attribute AttributeName
    private String namespace; // <saml:Attribute AttributeNamespace
    private String[] values; // <saml:AttributeValue (Scope+)values

    private AttributeAuthority issuer;
    private Calendar notBefore;
    private Calendar notOnOrAfter;

    private static Log logger =
        LogFactory.getLog(SAMLAttribute.class.getName());

    //no no-arg constructor

    public SAMLAttribute(String name,
                         String namespace,
                         String[] value,
                         AttributeAuthority issuer,
                         Calendar notBefore,
                         Calendar notOnOrAfter) throws Exception {

        //required in the SAML attribute schema
        if (name != null) {
            this.name = name;
        } else {
            throw new Exception("name cannot be null");
        }

        //required in the SAML attribute schema
        if (namespace != null) {
            this.namespace = namespace;
        } else {
            throw new Exception("namespace cannot be null");
        }

        //required in the SAML attribute schema
        if (value != null) {
            if (value.length == 0) {
                // SAML attribute requires at least one AttributeValue
                throw new Exception("values[] cannot be empty");
            }
            this.values = value;
        } else {
            throw new Exception("values[] cannot be null");
        }

        //optional
        this.issuer = issuer;
        this.notBefore = notBefore;
        this.notOnOrAfter = notOnOrAfter;
    }

    /**
     * optional
     * @return attribute issuer
     */
    public AttributeAuthority getIssuer() {
        return issuer;
    }

    /**
     * optional
     * @return notBefore timestamp
     */
    public Calendar getNotBefore() {
        return notBefore;
    }

    /**
     * optional
     * @return notOnOrAfter timestamp
     */
    public Calendar getNotOnOrAfter() {
        return notOnOrAfter;
    }

    public String getName() {
        return name;
    }

    public String getNamespace() {
        return namespace;
    }

    public String[] getValue() {
        return values;
    }

    public String toString() {
        SimpleDateFormat formatter =
                new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
        formatter.setTimeZone(TimeZone.getTimeZone("GMT"));

        StringBuffer buf = new StringBuffer("SAMLAttribute {\n");
        buf.append("  name='").
            append(name).
            append("'\n  namespace='").
            append(namespace);
        for (int i = 0; i < values.length; i++) {
            buf.append("'\n  value #").
                append(i+1).
                append("='").
                append(values[i]);
        }
        buf.append("'\n  issuer='").
            append(issuer != null ? issuer.getAAProviderId() : null).
            append("'\n  notBefore='").
            append(notBefore != null
                    ? formatter.format(notBefore.getTime()) : null).
            append("'\n  notOnOrAfter='").
            append(notOnOrAfter != null
                    ? formatter.format(notOnOrAfter.getTime()) : null).
            append("'\n}");

        return buf.toString();
    }

    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof SAMLAttribute)) return false;

        final SAMLAttribute samlAttribute = (SAMLAttribute) o;

        if (!name.equals(samlAttribute.name)) return false;
        if (!namespace.equals(samlAttribute.namespace)) return false;

        // TODO: document this; if either is null, it's a wildcard
        if ((issuer != null) && (samlAttribute.issuer !=null)) {
            if (!issuer.equals(samlAttribute.issuer)) return false;
        }

        // logical OR any multi-valued attributes
        for (int i = 0; i < values.length; i++) {
            for (int j = 0; j < samlAttribute.values.length; j++) {
                if (values[i].equals(samlAttribute.values[j])) {
                    logger.debug(values[i] + " = " + samlAttribute.values[j]);
                    return true;
                }
            }
        }
        return false;
    }

    public int hashCode() {
        int result;
        result = name.hashCode();
        result = 29 * result + namespace.hashCode();
        for (int i = 0; i < values.length; i++) {
            result = 29 * result + values[i].hashCode();
        }
        return result;
    }

    public boolean isValid() {
        if (notBefore == null && notOnOrAfter == null) {
            return true;
        }

        Calendar now = Calendar.getInstance();

        if (notOnOrAfter == null) {
            if (notBefore.before(now)) {
                return true;
            }
        } else if (notBefore == null) {
            if (notOnOrAfter.after(now)) {
                return true;
            }
        } else {
            if (notBefore.before(now) && notOnOrAfter.after(now)) {
                return true;
            }
        }

        return false;
    }

}
