/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization.attributes.policy;

import org.apache.axis.message.MessageElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.gridshib.policy.AttributePolicy;
import org.globus.gridshib.policy.Entry;
import org.globus.gridshib.policy.ListOfAttributes;
import org.globus.gridshib.policy.ListOfUsernames;
import org.globus.gridshib.policy.saml.AttributeType;
import org.globus.gridshib.policy.saml.AttributeValueType;
import org.globus.util.I18n;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.gridshib.gt.authorization.attributes.MessageInterface;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.xml.sax.InputSource;

import java.io.File;
import java.io.FileInputStream;
import java.util.Vector;

public class SimpleSAMLAttributePolicy {

    private static Log logger =
        LogFactory.getLog(SimpleSAMLAttributePolicy.class.getName());

    private static I18n i18n =
        I18n.getI18n(
            MessageInterface.RESOURCE,
            SimpleSAMLAttributePolicy.class.getClassLoader()
        );

    public static AttributeMapping[] parseAttrFile(File policyFile)
            throws Exception {

        AttributePolicy policy = null;
        FileInputStream in = null;
        try {
            in = new FileInputStream(policyFile);
            policy = (AttributePolicy)ObjectDeserializer
                .deserialize(new InputSource(in),
                             AttributePolicy.class);
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (Exception e) {
                    logger.error("couldn't close stream", e);
                }
            }
        }

        Entry[] entries = policy.getEntry();
        if (entries == null) {
            return null;
        }

        AttributeMapping[] attrMappings =
            new AttributeMapping[entries.length];

        for (int i=0; i < entries.length; i++) {

            ListOfAttributes attributeList = entries[i].getListOfAttributes();
            if (attributeList == null) {
                continue;
            }

            AttributeType[] samlAttrs = attributeList.getAttribute();

            Vector attrs  = getSAMLAttributes(samlAttrs);

            if (attrs == null) {
                continue;
            }
            if (logger.isDebugEnabled()) {
                logger.debug("# policy SAML attrs: " + attrs.size());
            }

            String[] usernames = null;
            ListOfUsernames usernameList = entries[i].getListOfUsernames();
            if (usernameList != null) {
                usernames = usernameList.getUsername();
            }

            attrMappings[i] = new AttributeMapping(attrs, usernames);
        }

        return attrMappings;
    }

    private static Vector getSAMLAttributes(AttributeType[] samlAttrs)
            throws Exception {

        if (samlAttrs == null) {
            return null;
        }

        //normalize...Storing as vector since individual element
        //compare in loop can be avoided.
        Vector attrs = new Vector(samlAttrs.length);
        for (int i = 0; i < samlAttrs.length; i++) {

            // Below would be the exact same code as in the PIP
            // if the SAML and shib namespaces agreed.  In fact,
            // at that point, this normalization wouldn't make
            // sense.  At that point, a collection of attributes
            // with associated times would be best, and xml:any
            // comparisons would be straightforward

            AttributeType attribute = samlAttrs[i];
            if (attribute == null) {
                String err = i18n.getMessage("attrStmtNoAttr");
                throw new Exception(err);
            }

            String attributeName = attribute.getAttributeName();
            if (attributeName == null) {
                String err = i18n.getMessage("attrStmtNoAttrName");
                throw new Exception(err);
            }

            String attributeNamespace =
                attribute.getAttributeNamespace().toString();
            if (attributeNamespace == null) {
                String err = i18n.getMessage("attrStmtNoAttrNamespace");
                throw new Exception(err);
            }


            logger.debug("found an attribute\n\n" +
                         "attribute name  = " + attributeName +
                         "\nattribute namespace  = " + attributeNamespace);

            AttributeValueType[] attrVals = attribute.getAttributeValue();

            if (attrVals == null) {
                String err = i18n.getMessage("attrStmtNoAttrValue");
                throw new Exception(err);
            }

            String[] values = new String[attrVals.length];

            for (int k = 0; k < attrVals.length; k++) {
                MessageElement[] elms = attrVals[k].get_any();
                String attributeValue = elms[0].getValue();
                String attributeScope = attrVals[k].getScope();

                String oldval = null; //for debug output
                if (attributeScope != null) {
                    oldval = attributeValue;
                    // see Shibboleth profiles
                    attributeValue = attributeValue +
                        "@" + attributeScope;
                }

                if (logger.isDebugEnabled()) {
                    if (attributeScope != null) {
                        logger.debug("\nattribute value = " + oldval +
                                     "\nattribute scope = " + attributeScope +
                                     "\nnew attribute value = "
                                     + attributeValue + "\n");
                    } else {
                        logger.debug("\nattribute value = " +
                                     attributeValue + "\n");
                    }
                }

                values[k] = attributeValue;
            }

            attrs.add(new SAMLAttribute(attributeName, attributeNamespace,
                                             values, null, null, null));
        }

        return attrs;
    }
}
