/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.samlpiputils;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.globus.common.CoGProperties;

import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.io.File;

/**
 * For pre-4.2 Globus installations, program to allow user to add
 * attribute authority certificates to Globus' trusted certificate
 * directory.
 */
public class CertificateSetup {

    public static final int COMMAND_LINE_ERROR = 1;
    public static final int APPLICATION_ERROR = 2;

    protected Options options = new Options();

    protected boolean debugMode = false;

    protected String mdFilePath;
    protected String mdDirPath;
    protected String removeDirPath;
    protected String[] trustedCertDirPaths;

    protected Integer action = CertificateSetupRun.NOTSET;

    /* Options --------------------------------------------------- */

    public static final Option HELP_OPT =
        OptionBuilder.withDescription("Displays help")
        .withLongOpt("help")
        .create("h");

    // see main()
    private static final String debugFlag = "-z";
    private static final String debugLongFlag = "--debug";
    public static final Option DEBUG_OPT =
        OptionBuilder.withDescription("Enables debug mode")
        .withLongOpt("debug")
        .create("z");

    public static final Option MD_DIR_OPT =
        OptionBuilder.withArgName( "directory" )
        .hasArg()
        .withDescription("A directory with many metadata files in it")
        .withLongOpt("mdDir")
        .create("d");

    public static final Option MD_FILE_OPT =
        OptionBuilder.withArgName( "file" )
        .hasArg()
        .withDescription("A metadata file")
        .withLongOpt("mdFile")
        .create("f");

    public static final Option STATUS_OPT =
        OptionBuilder.withDescription("Print status and exit (default action)")
        .withLongOpt("status")
        .create("s");

    public static final Option INSTALL_OPT =
        OptionBuilder.withDescription("Install certificates")
        .withLongOpt("install")
        .create("i");

//    public static final Option FORCE_OPT =
//        OptionBuilder.withDescription("Install certificates, overwrite without asking")
//        .withLongOpt("force")
//        .create("x");

    public static final Option REMOVE_OPT =
        OptionBuilder.withArgName("directory")
        .hasArg()
        .withDescription("Remove certificates, moving them to specified directory")
        .withLongOpt("remove")
        .create("r");

    public static final Option OVERRIDE_OPT =
        OptionBuilder.withArgName("directory")
        .hasArg()
        .withDescription("Manually set trusted certificate directory")
        .withLongOpt("override")
        .create("o");

    protected CertificateSetup() {
        addOptions();
    }

    protected void addOptions() {
        options.addOption(HELP_OPT);
        options.addOption(DEBUG_OPT);
        options.addOption(MD_DIR_OPT);
        options.addOption(MD_FILE_OPT);
        options.addOption(STATUS_OPT);
        options.addOption(INSTALL_OPT);
        //options.addOption(FORCE_OPT);
        options.addOption(REMOVE_OPT);
        options.addOption(OVERRIDE_OPT);
    }

    protected static Properties defaults() {
        // none yet
        // Properties defaultOptions = new Properties();
        //defaultOptions.put(AA_URL.getOpt(), DEFAULT_AAURL);
        return new Properties();
    }


    /* Get/set --------------------------------------------------- */

    public boolean isDebugMode() {
        return debugMode;
    }

    public void setDebugMode(boolean debugMode) {
        this.debugMode = debugMode;
    }

    public String getMdFilePath() {
        return mdFilePath;
    }

    public void setMdFilePath(String mdFilePath) {
        this.mdFilePath = mdFilePath;
    }

    public String getMdDirPath() {
        return mdDirPath;
    }

    public void setMdDirPath(String mdDirPath) {
        this.mdDirPath = mdDirPath;
    }

    public Integer getAction() {
        return action;
    }

    public void setAction(Integer action) {
        this.action = action;
    }

    public String getRemoveDirPath() {
        return removeDirPath;
    }

    public void setRemoveDirPath(String removeDirPath) {
        this.removeDirPath = removeDirPath;
    }

    public String[] getTrustedCertDirPaths() {
        return trustedCertDirPaths;
    }

    public void setTrustedCertDirPath(String[] trustedCertDirPaths) {
        this.trustedCertDirPaths = trustedCertDirPaths;
    }

    /* Run ------------------------------------------------------- */

    public static void main(String[] args) {
        CertificateSetup client = new CertificateSetup();
        Properties defaults = defaults();

        try {
            client.parse(args, defaults);
        } catch (ParseException e) {
            // can't client.isDebugMode() yet...
            boolean debug = false;
            for (int i = 0; i < args.length; i++) {
                if ((args[i].trim().equals(debugFlag)) ||
                        (args[i].trim().equals(debugLongFlag))) {
                    debug = true;
                }
            }
            System.err.println("Problem with arguments: ");
            System.err.println(e.getMessage());
            if (debug) {
                System.err.println(
                        "\n\n--------------- STACKTRACE ---------------");
                e.printStackTrace();
                System.err.println(
                        "------------------------------------------\n");
            }
            System.exit(COMMAND_LINE_ERROR);
        }


        try {
            if (client.isDebugMode()) {
                CertificateSetupRun.run(client, System.out, System.err);
            } else {
                CertificateSetupRun.run(client, System.out, null);
            }
        } catch (Exception e) {
            System.err.println("Problem: ");
            System.err.println(e.getMessage());
            if (client.isDebugMode()) {
                System.err.println(
                        "\n\n--------------- STACKTRACE ---------------");
                e.printStackTrace();
                System.err.println(
                        "------------------------------------------\n");
            }
            System.exit(APPLICATION_ERROR);
        }
    }

    public void displayUsage() {
        StringBuffer usage = new StringBuffer("Set either metadata file " +
                "or metadata directory to read in the AA certificates:");

        usage.append("\n    Either file:      [-").
              append(MD_FILE_OPT.getOpt()).
              append(" <path>]").
              append("\n    Or directory:     [-").
              append(MD_DIR_OPT.getOpt()).
              append(" <directory>]");

        usage.append("\n\nPick an action:");

        usage.append("\n    Either status [-").
              append(STATUS_OPT.getOpt()).
              //append("], force [-").
              //append(FORCE_OPT.getOpt()).
              append("], install [-").
              append(INSTALL_OPT.getOpt()).
              append("], or remove [-").
              append(REMOVE_OPT.getOpt()).
              append(" <dir>]");

        // passing usage to printHelp makes it look bad
        System.out.println(usage.toString());

        String header = "\nOptions:";
        HelpFormatter formatter = new HelpFormatter();

        // cli package is bugged, cannot set this to whitespace
        formatter.setSyntaxPrefix("__");
        formatter.printHelp(" ", header, options, null, false);
    }


    protected CommandLine parse(String [] args) throws ParseException {
        return parse(args, null);
    }

    protected CommandLine parse(String [] args, Properties defaultOptions)
        throws ParseException {

        CommandLineParser parser = new PosixParser();
        CommandLine line = parser.parse(this.options, args, defaultOptions);

        if (defaultOptions != null) {
            Iterator iter = defaultOptions.entrySet().iterator();
            while(iter.hasNext()) {
                Map.Entry entry = (Map.Entry)iter.next();
                Option opt = this.options.getOption((String)entry.getKey());
                if (opt != null) {
                    String desc = opt.getDescription();
                    desc += " (Default '" + entry.getValue() + "')";
                    opt.setDescription(desc);
                }
            }
        }

        if (args.length == 0) {
            System.out.println("For help message, use -" + HELP_OPT.getOpt());
            System.exit(0);
        }

        if (line.hasOption(HELP_OPT.getOpt())) {
            displayUsage();
            System.exit(0);
        }

        this.debugMode = line.hasOption(DEBUG_OPT.getOpt());

        if (line.hasOption(MD_DIR_OPT.getOpt())) {
            mdDirPath = line.getOptionValue(MD_DIR_OPT.getOpt());
            if (mdDirPath == null) {
                throw new ParseException("Metadata directory flag (" +
                        MD_DIR_OPT.getOpt() + ") " +
                        "supplied, but argument is empty");
            }
            if (this.debugMode) {
                System.err.println("Metadata directory path: " + mdDirPath);
            }
        }

        if (line.hasOption(MD_FILE_OPT.getOpt())) {
            mdFilePath = line.getOptionValue(MD_FILE_OPT.getOpt());
            if (mdFilePath == null) {
                throw new ParseException("Metadata directory flag (" +
                        MD_DIR_OPT.getOpt() + ") " +
                        "supplied, but argument is empty");
            }
            if (this.debugMode) {
                System.err.println("Metadata file path: " + mdFilePath);
            }
        }

        if ((mdDirPath != null) && (mdFilePath != null)) {
            throw new ParseException("Cannot specify both metadata " +
                    "file and metadata directory");
        }

        if ((mdDirPath == null) && (mdFilePath == null)) {
            throw new ParseException("Must specify either metadata " +
                    "file or metadata directory");
        }

        if (line.hasOption(STATUS_OPT.getOpt())) {
            if (this.debugMode) {
                System.err.println("Action: status");
            }
            action = CertificateSetupRun.STATUS;
        }

        if (line.hasOption(INSTALL_OPT.getOpt())) {
            if (this.debugMode) {
                System.err.println("Action: install");
            }
            if (action != CertificateSetupRun.NOTSET) {
                throw new ParseException("Can only specify one action");
            }
            action = CertificateSetupRun.INSTALL;
        }

//        if (line.hasOption(FORCE_OPT.getOpt())) {
//            if (this.debugMode) {
//                System.err.println("Action: install (forced)");
//            }
//            if (action != CertificateSetupRun.NOTSET) {
//                throw new ParseException("Can only specify one action");
//            }
//            action = CertificateSetupRun.FORCE;
//        }

        if (line.hasOption(REMOVE_OPT.getOpt())) {
            if (this.debugMode) {
                System.err.println("Action: remove");
            }
            if (action != CertificateSetupRun.NOTSET) {
                throw new ParseException("Can only specify one action");
            }
            removeDirPath = line.getOptionValue(REMOVE_OPT.getOpt());
            if (removeDirPath == null) {
                throw new ParseException("Override trusted certificate" +
                        " directory flag (" +
                        REMOVE_OPT.getOpt() + ") " +
                        "supplied, but argument is empty");
            }

            File dir = new File(removeDirPath);
            if (!dir.isDirectory()) {
                throw new ParseException("Not a directory: "
                                                     + dir.getAbsolutePath());
            }

            if (!dir.canRead()) {
                System.err.println("\n********\n" +
                        "WARNING: can not read remove's target directory "
                                     + dir.getAbsolutePath() + "\n********");
            }

            if (!dir.canWrite()) {
                System.err.println("\n********\n" +
                     "WARNING: can not write to remove's target directory "
                                     + dir.getAbsolutePath() + "\n********");
            }

            action = CertificateSetupRun.REMOVE;
        }

        if (action == CertificateSetupRun.NOTSET) {
            System.out.println("No action given, assuming 'status'");
            action = CertificateSetupRun.STATUS;
        }

        if (line.hasOption(OVERRIDE_OPT.getOpt())) {
            String dir = line.getOptionValue(OVERRIDE_OPT.getOpt());
            if (dir == null) {
                throw new ParseException("Override trusted certificate" +
                        " directory flag (" +
                        OVERRIDE_OPT.getOpt() + ") " +
                        "supplied, but argument is empty");
            }
            trustedCertDirPaths = new String[1];
            trustedCertDirPaths[0] = dir;
        } else {
            // these can be multiple paths
            String dirsStr =
                    CoGProperties.getDefault().getCaCertLocations();
            StringTokenizer tok = new StringTokenizer(dirsStr, ",");
            System.out.println("Found " + tok.countTokens() +
                                                    " CA cert locations");
            ArrayList dirs = new ArrayList();
            while (tok.hasMoreTokens()) {
                String dir = tok.nextToken();
                if (debugMode) {
                    System.err.println("Found a directory: " + dir);
                }
                dirs.add(dir);
            }

            trustedCertDirPaths = (String[]) dirs.toArray(new String[0]);
        }

        for (int i = 0; i < trustedCertDirPaths.length; i++) {
            File dir = new File(trustedCertDirPaths[i]);
            trustedCertDirPaths[i] = dir.getAbsolutePath();

            System.out.println("Trusted certificate directory: "
                                                     + trustedCertDirPaths[i]);

            if (!dir.isDirectory()) {
                throw new ParseException("Not a directory: "
                                                     + trustedCertDirPaths[i]);
            }

            if (!dir.canRead()) {
                System.err.println("\n********\n" +
                        "WARNING: can not read trusted certificate directory "
                                     + trustedCertDirPaths[i] + "\n********");
            }

            if (!dir.canWrite()) {
                System.err.println("\n********\n" +
                     "WARNING: can not write to trusted certificate directory "
                                     + trustedCertDirPaths[i] + "\n********");
            }
        }

        return line;
    }
}


