/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.samlpiputils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.keys.keyresolver.KeyResolverException;
import org.globus.opensaml11.md.metadata.AttributeAuthorityDescriptor;
import org.globus.opensaml11.md.metadata.EntitiesDescriptor;
import org.globus.opensaml11.md.metadata.EntityDescriptor;
import org.globus.opensaml11.md.metadata.ExtendedEntitiesDescriptor;
import org.globus.opensaml11.md.metadata.ExtendedEntityDescriptor;
import org.globus.opensaml11.md.metadata.KeyAuthority;
import org.globus.opensaml11.md.metadata.KeyDescriptor;
import org.globus.opensaml11.md.metadata.provider.XMLMetadata;
import org.globus.opensaml11.md.xml.Parser;
import org.globus.opensaml11.saml.SAMLException;
import org.globus.util.I18n;
import org.globus.wsrf.config.ConfigException;
import org.globus.gridshib.gt.authorization.ShibbolethConstants;
import org.globus.gridshib.gt.authorization.attributes.MessageInterface;
import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Iterator;

public class MetadataUtil {

    private static I18n i18n =
        I18n.getI18n(
            MessageInterface.RESOURCE,
            MetadataUtil.class.getClassLoader()
        );

    static Log logger = LogFactory.getLog(MetadataUtil.class.getName());


    /**
     * Given a directory of metadata files, find all Attribute Authorities.
     * Does not load the attributes.
     * Not being able to read files is NOT an exception.
     * @param path
     * @return Attribute Authorities
     * @throws ConfigException
     */
    public static AttributeAuthority[] loadMetadataDir(String path)
                            throws ConfigException {
        return loadMetadataDir(path, false);
    }

    /**
     * Given a metadata file, find all Attribute Authorities.
     * Does not load the attributes.
     * Not being able to read files IS an exception.
     * @param mdpath
     * @return Attribute Authorities
     */
    public static AttributeAuthority[] loadMetadata(String mdpath)
                    throws IOException, ConfigException {
        return loadMetadata(mdpath, false);
    }

    /**
     * Given a directory of metadata files, find all Attribute Authorities
     * Not being able to read files is NOT an exception.
     * @param path
     * @param respectAttributes if true, loads in attributes in metadata
     * @return Attribute Authorities
     * @throws ConfigException
     */
    public static AttributeAuthority[] loadMetadataDir(
                                        String path, boolean respectAttributes)
                            throws ConfigException {


        File dir = new File(path);
        if (!dir.isDirectory()) {
            String err = i18n.getMessage("metadataDirNotDir");
            logger.fatal(err);
            throw new ConfigException(err);
        }

        File[] files = dir.listFiles();
        if (files == null) {
            return new AttributeAuthority[0];
        }

        if (logger.isDebugEnabled()) {
            for (int i = 0; i < files.length; i++) {
                logger.debug("found metadata file: " + files[i].getPath());
            }
        }

        ArrayList aaList = new ArrayList();

        for (int i = 0; i < files.length; i++) {
            try {
                ArrayList aas = findAAs(
                      loadMetadataPath(files[i].getPath()), respectAttributes);
                aaList.addAll(aas);
            } catch (Exception e) {
                logger.debug("",e);
                logger.error(e.getClass().getName() + ": " + e.getMessage());
            }
        }

        return (AttributeAuthority[])
                     aaList.toArray(new AttributeAuthority[0]);
    }

    /**
     * Given a metadata file, find all Attribute Authorities
     * Not being able to read files IS an exception.
     * @param mdpath
     * @param respectAttributes if true, loads in attributes in metadata
     * @return Attribute Authorities
     */
    public static AttributeAuthority[] loadMetadata(String mdpath,
                                                    boolean respectAttributes)
            throws IOException, ConfigException {

        ArrayList aaList;
        try {
            aaList = findAAs(loadMetadataPath(mdpath),respectAttributes);
        } catch (SAXException e) {
            throw new ConfigException(e);
        } catch (SAMLException e) {
            throw new ConfigException(e);
        }

        return (AttributeAuthority[])
                     aaList.toArray(new AttributeAuthority[0]);
    }

    private static XMLMetadata loadMetadataPath(String mdpath)
            throws IOException, SAXException, SAMLException {

        // Parse file and verify root element.
        URL url;
        try {
            url = new URL(mdpath);
        } catch (Exception e) {
            url = new URL(new URL("file:"), mdpath);
        }

        Parser.init();
        Document doc = Parser.loadDom(url, true);

        if (doc == null) {
            String err =
                    "Metadata error: unable to read in file: " + mdpath;
            throw new IOException(err);
        }

        logger.debug("file is valid XML: " + mdpath);

        Element el = doc.getDocumentElement();
        XMLMetadata md = new XMLMetadata(el);

        logger.debug("successfully read in metadata from file: " + mdpath);

        return md; // this cannot be null (see XMLMetadata)
    }

    private static ArrayList findAAs(XMLMetadata md,
                                     boolean respectAttributes)
            throws ConfigException {

        // md cannot be null since it only comes from loadMetadata

        EntityDescriptor ed = md.getRootEntity();
        EntitiesDescriptor eds = md.getRootEntities();

        if ((ed == null) && (eds == null)) {
            throw new ConfigException("no EntityDescriptor(s) in metadata");
        }

        if (ed != null) {
            logger.debug("found lone EntityDescriptor in metadata");
            ArrayList ret = new ArrayList();
            ret.add(convertED(ed,respectAttributes,null));
            return ret;
        }

        // eds == null already checked
        return convertEDs(eds,respectAttributes,null);
    }

    /**
     * it's recursive...
     *
     *  <choice minOccurs="1" maxOccurs="unbounded">
     *      <element ref="md:EntityDescriptor"/>
     *      <element ref="md:EntitiesDescriptor"/>
     *  </choice>
     *
     * @param eds
     * @return ArrayList of AttributeAuthority
     */
    private static ArrayList convertEDs(EntitiesDescriptor eds,
                                        boolean respectAttributes,
                                        ArrayList edsKeyAuths) {
        logger.debug("found EntitiesDescriptor in metadata");

        if (eds instanceof ExtendedEntitiesDescriptor) {
            Iterator keyAuths =
                    ((ExtendedEntitiesDescriptor) eds).getKeyAuthorities();
            ArrayList temp = new ArrayList();
            addKeyAuthorities(keyAuths, temp);
            if (edsKeyAuths == null) {
                edsKeyAuths = temp;
            } else {
                edsKeyAuths.addAll(temp);
            }
        }

        ArrayList ret = new ArrayList();

        Iterator edsIter = eds.getEntitiesDescriptors();
        Iterator edIter = eds.getEntityDescriptors();

        while (edsIter.hasNext()) {
            ret.addAll(convertEDs((EntitiesDescriptor)edsIter.next(),
                                  respectAttributes, edsKeyAuths));
        }

        while (edIter.hasNext()) {
            // passing edsKeyAuths because the top level EntitiesDescriptor
            // can contain authorities for each entity in the document
            AttributeAuthority aa =
                    convertED((EntityDescriptor) edIter.next(),
                              respectAttributes,
                              edsKeyAuths);

            if (aa != null) {
                ret.add(aa);
            }
        }

        return ret;
    }

    private static AttributeAuthority convertED(EntityDescriptor ed,
                                                boolean respectAttributes,
                                                ArrayList edsKeyAuths) {

        ExtendedEntityDescriptor eed = null;
        try {
             eed = (ExtendedEntityDescriptor) ed;
        } catch (ClassCastException e) {
            // this is ok
            logger.debug("EntityDescriptor does not have " +
                    "the KeyAuthority metadata extension: ",e);
        }

        AttributeAuthorityDescriptor aad =
             ed.getAttributeAuthorityDescriptor(ShibbolethConstants.SAMLP_1_1);

        if (aad == null) {
            // this is ok
            logger.debug("AttributeAuthorityDescriptor not present in " +
                    "EntityDescriptor");
            return null;
        }

        String IdPproviderId = ed.getId();

        // TODO: there could be multiple endpoints, not supporting
        String AAurl =
           aad.getAttributeServiceManager().getDefaultEndpoint().getLocation();

        if ((AAurl == null) || (IdPproviderId == null)) {
            logger.error("Attribute authority or IdPproviderId missing" +
                    " from EntityDescriptor");
            return null;
        }

        // todo: could check against the requirement
        // which is urn:oasis:names:tc:SAML:1.1:nameid-format:X509SubjectName
        // Iterator nameidIter = aad.getNameIDFormats();


        /* metadata may contain list of possible attributes that
           would be returned from a query and so the PDP can
           optimize (user configurable) if ACL does not rely on
           anything the AA can give */

        SAMLAttribute[] attrs = null;

        if (respectAttributes) {

            Iterator attributes = aad.getAttributes();
            ArrayList normalized = new ArrayList();
            while (attributes.hasNext()) {
                org.globus.opensaml11.saml.SAMLAttribute attr =
                  (org.globus.opensaml11.saml.SAMLAttribute) attributes.next();
                if (attr != null) {
                    SAMLAttribute norm = null;
                    try {
                         norm = SAMLutil.parseAttribute(attr,
                                                        null,
                                                        null,
                                                        null);

                    } catch (Exception e) {
                        logger.error("",e);

                    }
                    normalized.add(norm);
                }
            }

            attrs = (SAMLAttribute[]) normalized.toArray(new SAMLAttribute[0]);

            if (attrs.length == 0) {
                attrs = null;
            } else if (logger.isDebugEnabled()) {
                logger.debug("found " + attrs.length +
                                              " attributes in IdP metadata:");
                for (int i = 0; i < attrs.length; i++) {
                    logger.debug(attrs[i]);
                }
            }
        }

        ArrayList keyauths = new ArrayList();
        if (eed != null) {
            Iterator keyauthsIter = eed.getKeyAuthorities();
            addKeyAuthorities(keyauthsIter, keyauths);
        }

        if (edsKeyAuths != null) {
            keyauths.addAll(edsKeyAuths);
        }

        X509Certificate key = null;
        Iterator keys = aad.getKeyDescriptors();
        if (keys.hasNext()) {
            KeyDescriptor kd = (KeyDescriptor)keys.next();
            try {
                key = kd.getKeyInfo().getX509Certificate();
                if (isSelfSigned(key)) {
                    // it MAY or MAY NOT already be part of the keyauthorities
                    // list -- it is optional to add it there
                    if (!keyauths.contains(key)) {
                        keyauths.add(key);
                    }
                }
            } catch (KeyResolverException e) {
                logger.error("",e);
            }
        }

        return new AttributeAuthority(AAurl,
                                      IdPproviderId,
                                      IdPproviderId,
                                      attrs,
                                      key,
                                      keyauths);
    }

    private static void addKeyAuthorities(Iterator keyauthsIter, ArrayList keyauths) {
        while (keyauthsIter.hasNext()) {
            // disregard depth and use flag... todo?
            KeyAuthority ka = (KeyAuthority) keyauthsIter.next();
            Iterator kaIter = ka.getKeyInfos();
            while (kaIter.hasNext()) {
                KeyInfo ki = (KeyInfo)kaIter.next();
                try {
                    keyauths.add(ki.getX509Certificate());
                } catch (KeyResolverException e) {
                    logger.error("",e);
                }
            }
        }
    }

    /**
     * @param key this will not be modified
     * @return true if issuer DN matches subject DN
     */
    public static boolean isSelfSigned(X509Certificate key) {
        if (key != null) {
            if (key.getSubjectDN().equals(key.getIssuerDN())) {
                return true;
            }
        }
        return false;
    }
}