/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package org.globus.gridshib.gt.authorization.samlpiputils;

import _0._1.shibboleth.mace.AttributeValueType;

import assertion._0._1.SAML.tc.names.oasis.AssertionType;
import assertion._0._1.SAML.tc.names.oasis.AttributeStatementType;
import assertion._0._1.SAML.tc.names.oasis.AttributeType;
import assertion._0._1.SAML.tc.names.oasis.AudienceRestrictionConditionType;
import assertion._0._1.SAML.tc.names.oasis.ConditionsType;
import assertion._0._1.SAML.tc.names.oasis.NameIdentifierType;
import assertion._0._1.SAML.tc.names.oasis.SubjectType;

import org.apache.axis.message.MessageElement;
import org.apache.axis.types.URI;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;
import org.globus.gridshib.gt.authorization.attributes.AttributeAuthority;
import org.globus.gridshib.gt.authorization.attributes.MessageInterface;
import org.globus.gridshib.gt.authorization.attributes.SAMLAttribute;
import org.globus.gridshib.gt.security.GS4GTSecurityContext;
import org.globus.gridshib.security.SAMLIdentity;
import org.globus.gridshib.security.util.CertUtil;
import org.globus.opensaml11.md.common.Constants;
import org.globus.opensaml11.saml.SAMLNameIdentifier;

import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TimeZone;

import javax.security.auth.Subject;
import javax.security.auth.x500.X500Principal;

public class SAMLutil {

    private static I18n i18n =
        I18n.getI18n(
            MessageInterface.RESOURCE,
            SAMLutil.class.getClassLoader()
        );

    static Log logger = LogFactory.getLog(SAMLutil.class.getName());



    /* ------------------------------------------------ */
    /*     Subject                                      */
    /* ------------------------------------------------ */

    public static SubjectType constructSubjectFromX509Identity(Subject peerSubject)
                                                        throws Exception {

        assert (peerSubject != null);

        String name = getX509SubjectName(peerSubject);
        String format = SAMLNameIdentifier.FORMAT_X509;
        logger.debug("name = " + name);
        logger.debug("format = " + format);

        URI formatAsURI;
        try {
            formatAsURI = new URI(format);
        } catch (URI.MalformedURIException e) {
            String msg = "Unable to convert NameIdentifier format " +
                         "to URI (this should never happen!)";
            logger.error(msg, e);
            throw new RuntimeException(msg);
        }
        logger.debug("formatAsURI = " + formatAsURI.toString());

        return constructSubjectType(name, formatAsURI, null);
    }

    /*
     * Finds X509 SAML subject name.
     *
     * @return Proper SAML subject.
     */
    private static String getX509SubjectName(Subject peerSubject)
                                      throws Exception {

        X509Certificate cert = CertUtil.getEEC(peerSubject);
        if (cert == null) {
            String err = i18n.getMessage("nopeer");
            logger.debug(err);
            throw new Exception(err);
        }

        X500Principal certSubject = cert.getSubjectX500Principal();
        String subjectName = certSubject.getName(X500Principal.RFC2253);
        //subjectName = subjectName.toLowerCase();
        logger.debug("x500 subjectName: " + subjectName);
        return subjectName;
    }

    public static SubjectType constructSubjectFromSAMLIdentity(Subject peerSubject) {

        assert (peerSubject != null);

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);
        SAMLIdentity identity = secCtx.getCurrentSAMLIdentity();
        if (identity == null) {
            String msg = "Current SAML identity does not exist";
            logger.error(msg);
            throw new IllegalStateException(msg);
        }

        SAMLNameIdentifier nameid = identity.getNameID();
        String name = nameid.getName();
        String format = nameid.getFormat();
        String nameQualifier = nameid.getNameQualifier();
        logger.debug("name = " + name);
        logger.debug("format = " + format);
        logger.debug("nameQualifier = " + nameQualifier);

        URI formatAsURI;
        try {
            formatAsURI = new URI(format);
        } catch (URI.MalformedURIException e) {
            String msg = "Unable to convert NameIdentifier format " +
                         "to URI (this should never happen!)";
            logger.error(msg, e);
            throw new RuntimeException(msg);
        }
        logger.debug("formatAsURI = " + formatAsURI.toString());

        return constructSubjectType(name, formatAsURI, nameQualifier);
    }

    /**
     * Get the issuer and the nameID from the
     * <code>currentIdentity</code> and use these to compute
     * the AA as follows:
     * <ol>
     *   <li> If the nameID does not contain a NameQualifier,
     *   skip the next step.
     *   <li> Use the NameQualifier to resolve the AA (from
     *   metadata); if that fails, continue with the next step.
     *   <li> Use the issuer to resolve the AA (from metadata);
     *   if that fails, return null.
     * </ol>
     */
    public static AttributeAuthority chooseAA(AttributeAuthority[] authorities,
                                              Subject peerSubject)
                                       throws Exception {

        assert (authorities != null && authorities.length > 0);
        assert (peerSubject != null);

        GS4GTSecurityContext secCtx =
            GS4GTSecurityContext.getGS4GTSecurityContext(peerSubject);
        assert (secCtx != null);
        SAMLIdentity identity = secCtx.getCurrentSAMLIdentity();
        if (identity == null) {
            String msg = "Current SAML identity does not exist";
            logger.error(msg);
            throw new IllegalStateException(msg);
        }

        AttributeAuthority aa = null;

        String entityID = identity.getNameID().getNameQualifier();
        if (entityID == null) {
            entityID = identity.getIssuer();
            logger.debug("Using Issuer as entityID: " + entityID);
            aa = AttributeAuthority.findAA(entityID, authorities);
        } else {
            logger.debug("Using NameQualifier as entityID: " + entityID);
            aa = AttributeAuthority.findAA(entityID, authorities);
            if (aa == null) {
                logger.debug("NameQualifier not found in metadata");
                entityID = identity.getIssuer();
                logger.debug("Using Issuer as entityID: " + entityID);
                aa = AttributeAuthority.findAA(entityID, authorities);
            }
        }

        if (aa == null) {
            logger.debug("entityID not found in metadata");
        } else {
            logger.debug("entityID found in metadata: " + aa.toString());
        }

        return aa;
    }

    private static SubjectType constructSubjectType(String subjectName,
                                                    URI formatElement,
                                                    String nameQualifer) {

        NameIdentifierType nameIdentifier = new NameIdentifierType();
        nameIdentifier.setFormat(formatElement);
        nameIdentifier.setNameQualifier(nameQualifer);
        nameIdentifier.set_value(subjectName);
        return new SubjectType(nameIdentifier, null);
    }




    /* ------------------------------------------------ */
    /*     Assertions                                   */
    /* ------------------------------------------------ */


    public static AttributeType[] parseAssertion(AssertionType assertion,
                                                 SubjectType subject,
                                                 GridSP sp,
                                                 AttributeAuthority aa)
            throws Exception {

        SimpleDateFormat formatter =
                    new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
            formatter.setTimeZone(TimeZone.getTimeZone("GMT"));

        if (logger.isDebugEnabled()) {
            Calendar instant = assertion.getIssueInstant();
            logger.debug("assertion issue instant: " +
                    formatter.format(instant.getTime()));
        }

        String issuer = assertion.getIssuer();
        if (!issuer.equals(aa.getAAProviderId())) {
            // we only trust assertions from the intended IdP
            String err = i18n.getMessage("assertionWrongIssuer",
                    new Object[] {aa.getAAProviderId(), issuer} );
            throw new Exception(err);
        }

        ConditionsType conditions = assertion.getConditions();
        Calendar notBefore = null;
        Calendar notOnOrAfter = null;
        if (conditions != null) {
            notBefore = conditions.getNotBefore();
            notOnOrAfter = conditions.getNotOnOrAfter();
        }
        if ((conditions == null)
                || (notBefore == null)
                || (notOnOrAfter == null)) {

            // times are optional in schema, but required by us
            String err = i18n.getMessage("assertionNoTimeConditions");
            throw new Exception(err);
        }

        AudienceRestrictionConditionType aud =
                conditions.getAudienceRestrictionCondition();

        // only accepting assertions with this SP providerId in
        // the audience restriction. Sort of pedantic, but if the
        // providerId was different, right now this signals
        // something wrong on the IdP side
        if (aud != null) {
            URI[] uris = aud.getAudience();
            boolean found = false;
            for (int j = 0; j < uris.length; j++) {
                if (uris[j].equals(sp.getResourceURI())) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                StringBuffer buf = new StringBuffer();
                for (int j = 0; j < uris.length; j++) {
                    buf.append(uris[j].toString());
                }
                String err = i18n.getMessage("assertionWrongAudience",
                  new Object[] {sp.getResourceURI(),
                          buf.toString()});
                throw new Exception(err);
            }
        }

        // if the assertion is not valid anymore, ignore it
        Calendar currentTime = Calendar.getInstance();
        if (currentTime.after(notOnOrAfter)) {
            String err = i18n.getMessage("assertionExpired",
                    formatter.format(notOnOrAfter.getTime()));
            logger.warn(err);
            return null;
        }

        // TODO: look at spec to verify, see TODO below
        AttributeStatementType attrStatement =
                assertion.getAttributeStatement();
        if (attrStatement == null) {
            // AttributeStatement is optional in schema,
            // but required by us
            String err = i18n.getMessage("assertionNoAttrStatement");
            throw new Exception(err);
        }
        SubjectType attrSubject = attrStatement.getSubject();
        if (attrSubject == null) {
            String err = i18n.getMessage("assertionNoSubject");
            throw new Exception(err);
        }

        // only accepting assertions about queried subject
        if (!attrSubject.equals(subject)) {
            String err = i18n.getMessage("assertionWrongSubject",
                  new Object[] {subject.getNameIdentifier().get_value(),
                          attrSubject.getNameIdentifier().get_value()});
            throw new Exception(err);
        }

        AttributeType[] attributes = attrStatement.getAttribute();

        // TODO look into 0 attribute message, possibly no exception
        // here;  currently, DENY will be returned anyhow, mostly a
        // question of what to pass to the client.  That is really
        // is not up to the PIP, but the PDP. So, this TODO is mainly
        // about what to tell the PDP.  When the authz chain changes
        // this question should come up again since the PIP will stand
        // alone in the chain.
        if ((attributes == null) || (attributes.length == 0)) {
            String err = i18n.getMessage("attrStmtNoAttrs");
            throw new Exception(err);
        }

        return attributes;
    }

    /**
     * Good attribute conversion or exception, but never null.
     *
     * @param attribute
     * @param notBefore
     * @param notOnOrAfter
     * @param issuer
     * @return normalized attribute, not null
     * @throws Exception
     */
    public static SAMLAttribute parseAttribute(AttributeType attribute,
                                               Calendar notBefore,
                                               Calendar notOnOrAfter,
                                               AttributeAuthority issuer)
                                        throws Exception {

        if (attribute == null) {
            String err = i18n.getMessage("attrStmtNoAttr");
            throw new Exception(err);
        }

        String attributeName = attribute.getAttributeName();
        if (attributeName == null) {
            //it's required by schema...
            String err = i18n.getMessage("attrStmtNoAttrName");
            throw new Exception(err);
        }

        String attributeNamespace =
                attribute.getAttributeNamespace().toString();
        if (attributeNamespace == null) {
            //it's required by schema...
            String err = i18n.getMessage("attrStmtNoAttrNamespace");
            throw new Exception(err);
        }

        logger.debug("found an attribute\n\n" +
                  "attribute name  = " + attributeName +
                  "\nattribute namespace  = " + attributeNamespace);

        AttributeValueType[] attr = attribute.getAttributeValue();

        String[] values = new String[attr.length];

        for (int k = 0; k < attr.length; k++) {

            MessageElement[] elms = attr[k].get_any();
            String attributeValue = elms[0].getValue();
            String attributeScope = attr[k].getScope();

            String oldval = null; //for debug output
            if (attributeScope != null) {
                oldval = attributeValue;
                // see Shibboleth profiles
                attributeValue = attributeValue +
                        "@" + attributeScope;
            }

            if (logger.isDebugEnabled()) {
                if (attributeScope != null) {
                    logger.debug("\nattribute value = " + oldval +
                       "\nattribute scope = " + attributeScope +
                       "\nnew attribute value = " + attributeValue +
                       "\n");
                } else {
                    logger.debug("\nattribute value = " +
                            attributeValue + "\n");
                }
            }

            values[k] = attributeValue;
        }

        return new SAMLAttribute(attributeName,
                                 attributeNamespace,
                                 values,
                                 issuer,
                                 notBefore,
                                 notOnOrAfter);
    }

    /**
     * Good attribute conversion or exception, but never null.
     *
     * @param attribute
     * @param notBefore
     * @param notOnOrAfter
     * @return normalized attribute, not null
     * @throws Exception
     */
    public static org.globus.opensaml11.saml.SAMLAttribute
                                parseAttribute(AttributeType attribute,
                                               Calendar notBefore,
                                               Calendar notOnOrAfter)
                                        throws Exception {

        if (attribute == null) {
            String err = i18n.getMessage("attrStmtNoAttr");
            throw new Exception(err);
        }

        String attributeName = attribute.getAttributeName();
        if (attributeName == null) {
            //it's required by schema...
            String err = i18n.getMessage("attrStmtNoAttrName");
            throw new Exception(err);
        }

        String attributeNamespace =
                attribute.getAttributeNamespace().toString();
        if (attributeNamespace == null) {
            //it's required by schema...
            String err = i18n.getMessage("attrStmtNoAttrNamespace");
            throw new Exception(err);
        }

        logger.debug("found an attribute\n\n" +
                  "attribute name  = " + attributeName +
                  "\nattribute namespace  = " + attributeNamespace);

        AttributeValueType[] attr = attribute.getAttributeValue();

        Set values = new HashSet();

        for (int k = 0; k < attr.length; k++) {

            MessageElement[] elms = attr[k].get_any();
            String attributeValue = elms[0].getValue();
            String attributeScope = attr[k].getScope();

            String oldval = null; //for debug output
            if (attributeScope != null) {
                oldval = attributeValue;
                // see Shibboleth profiles
                attributeValue = attributeValue +
                        "@" + attributeScope;
            }

            if (logger.isDebugEnabled()) {
                if (attributeScope != null) {
                    logger.debug("\nattribute value = " + oldval +
                       "\nattribute scope = " + attributeScope +
                       "\nnew attribute value = " + attributeValue +
                       "\n");
                } else {
                    logger.debug("\nattribute value = " +
                            attributeValue + "\n");
                }
            }

            values.add(attributeValue);
        }

        org.globus.opensaml11.saml.SAMLAttribute a;
        a = new org.globus.opensaml11.saml.SAMLAttribute();
        a.setName(attributeName);
        a.setNamespace(attributeNamespace);
        a.setType(null);
        if (notBefore == null || notOnOrAfter == null) {
            a.setLifetime(0);
        } else {
            // compute lifetime:
            long t1 = notBefore.getTimeInMillis();
            long t2 = notOnOrAfter.getTimeInMillis();
            a.setLifetime(t2 - t1);
        }
        a.setValues(values);

        return a;
    }

    /**
     * Parse SAML Attribute.
     *
     * @param attribute
     * @param notBefore
     * @param notOnOrAfter
     * @param issuer
     * @return normalized attribute, not null
     * @throws Exception
     */
    public static SAMLAttribute parseAttribute(
                                    org.globus.opensaml11.saml.SAMLAttribute attribute,
                                    Calendar notBefore,
                                    Calendar notOnOrAfter,
                                    AttributeAuthority issuer)
                             throws Exception {

        String name = attribute.getName();
        String namespace = attribute.getNamespace();

        Iterator valuesIter = attribute.getValues();
        ArrayList valuesList = new ArrayList();
        while (valuesIter.hasNext()) {
            try {
                // do not need to worry about scope here, but
                // do need to check each is just a string
                String value = (String)valuesIter.next();
                valuesList.add(value);
            } catch (ClassCastException e) {
                throw new Exception("only strings are supported" +
                        " for attribute values");
            }

        }

        String[] values = (String[]) valuesList.toArray(new String[0]);

        return new SAMLAttribute(name,
                                 namespace,
                                 values,
                                 issuer,
                                 notBefore,
                                 notOnOrAfter);
    }

    /**
     * Parse SAML Subject.
     *
     * @param subject
     * @param notBefore
     * @param notOnOrAfter
     * @param issuer
     * @return normalized attribute, not null
     * @throws Exception
     */
    public static SAMLAttribute parseSubject(  // UNUSED
                                    org.globus.opensaml11.saml.SAMLSubject subject,
                                    Calendar notBefore,
                                    Calendar notOnOrAfter,
                                    AttributeAuthority issuer)
                             throws Exception {

        SAMLNameIdentifier nameid = subject.getNameIdentifier();

        String name = nameid.getFormat();
        String namespace = Constants.SHIB_ATTRIBUTE_NAMESPACE_URI;

        ArrayList valuesList = new ArrayList();
        valuesList.add(nameid.toString());
        String[] values = (String[]) valuesList.toArray(new String[0]);

        return new SAMLAttribute(name,
                                 namespace,
                                 values,
                                 issuer,
                                 notBefore,
                                 notOnOrAfter);
    }
}


