/*
 * Copyright 1999-2008 University of Chicago
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package org.globus.gridshib.gt.authorization.attributes.tests;

import junit.framework.Test;
import junit.framework.TestSuite;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.Text;
import org.apache.axis.types.URI;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.gridshib.gt.authorization.ShibbolethConstants;
import org.globus.gridshib.gt.authorization.attributes.tests.aaservice.generated.AttributeType;
import org.globus.gridshib.gt.authorization.attributes.tests.aaservice.generated.AttributeValueType;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.ListOfAttributes;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.ListOfUsernames;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.NoAuthRequest;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.Policy;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.PolicyDNs;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.PolicyEntries;
import org.globus.gridshib.gt.authorization.attributes.tests.service.generated.ShibbolethTestPortType;
import org.globus.gridshib.gt.authorization.attributes.tests.util.CompatibleUtil;

import org.globus.gsi.GlobusCredential;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.NoAuthorization;
import org.globus.wsrf.test.GSITestContainer;
import org.globus.wsrf.test.GridTestCase;
import org.globus.wsrf.test.TestContainer;

import javax.xml.rpc.Stub;
import java.io.File;
import java.text.DateFormat;
import java.util.Calendar;

public class TestShibbolethCallout extends GridTestCase {

    static Log logger =
        LogFactory.getLog(TestShibbolethCallout.class.getName());

    Object authz = null;
    String testServiceAddress = null;
    String attributeServiceUrl = null;
    String attributeServiceIdentity = null;

    private DateFormat localFormat = DateFormat.getDateTimeInstance();

    private TestContainer testContainer;

    public TestShibbolethCallout(String name) throws Exception {
        super(name);

        //authz = TestConstants.getConfiguredClientAuthz();
        //logger.debug("Authz is " + authz.getClass().getName());

        authz = NoAuthorization.getInstance();
    }

    public static Test suite() {
        return new TestSuite(TestShibbolethCallout.class);
    }

    public static void main(String[] args) {
        junit.textui.TestRunner.run(suite());
    }

    protected void setUp() throws Exception {
        this.testContainer = new GSITestContainer();
        this.testContainer.start();
        lognobullet("\n---- starting test ----\n");
        lognobullet(localFormat.format(Calendar.getInstance().getTime()));
    }

    protected void tearDown() throws Exception {
        this.testContainer.stop();
        super.tearDown();
        lognobullet(localFormat.format(Calendar.getInstance().getTime()));
        lognobullet("\n----  ending test  ----\n");
    }


    /* ********** TESTS ************ */


    // would be better to test finer grained denials, but the PDP is
    // set up to return DENY for most errors

    public void testPositiveAttributeAccessTransport() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testPositiveAttributeAccessTransport";
        String description =
                    "tests for "
                    + "allowed access based on one attribute using \n"
                    + "GSITransport. A valid assertion with one attribute \n"
                    + "is returned from the test AA and the same attribute \n"
                    + "is in the configured PDP policy.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress, authz);

        NoAuthRequest defaultControl =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        defaultControl.setTestname(name);
        defaultControl.setTestdescription(description);

        log("Calling the setup operation with default settings.");
        testPort.noAuth(defaultControl);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {
            log("FAIL: exception while calling createResource(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");

    }

    public void testPositiveAttributeAccessSecMsg() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testPositiveAttributeAccessSecMsg";
        String description =
                    "tests for"
                    + "allowed access based on one\n attribute using "
                    + "GSISecureMessage. A valid assertion with one "
                    + "attribute\n is returned from the test AA and the same "
                    + "attribute is in the configured PDP\n policy.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress, authz);

        NoAuthRequest defaultControl =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        defaultControl.setTestname(name);
        defaultControl.setTestdescription(description);

        log("Calling the setup operation with default settings.");
        testPort.noAuth(defaultControl);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_SEC_MSG, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {
            log("FAIL: exception while calling createResource(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");

    }

    public void testPositiveAttributeAccessSecConv() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testPositiveAttributeAccessSecConv";
        String description = "tests for "
                    + "allowed access based on one\n attribute using "
                    + "GSISecureConversation. A valid assertion with one "
                    + "attribute\n is returned from the test AA and the same "
                    + "attribute is in the configured PDP\n  policy.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress, authz);

        NoAuthRequest defaultControl =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        defaultControl.setTestname(name);
        defaultControl.setTestdescription(description);

        log("Calling the setup operation with default settings.");
        testPort.noAuth(defaultControl);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_SEC_CONV, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {
            log("FAIL: exception while calling createResource(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");

    }

    public void testNegativeNoAttribute() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testNegativeNoAttribute";
        String description = "tests for denied "
                    + "access.\n A valid assertion with zero attributes is "
                    + "returned\n from the test AA and authorization should "
                    + "fail.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        // what we would like to test here is more fine-grained, that the
        // PIP rejects the empty assertion, but the PDP is set up to return
        // DENY for most errors

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);

        log("Configuring control to signal empty assertion");
        control.setAaEmptyAssertion(true);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testNegativeWithAttribute() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testNegativeWithAttribute";
        String description = "tests for denied "
                    + "access.\n A valid assertion with the wrong attribute "
                    + "is returned\n from the test AA and authorization should "
                    + "fail.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);


        log("Configuring control to send back wrong attribute");
        control.setAaMungeAttributes(true);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            log("Exception thrown, need to check if expected");
            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            } else {
                log(e.getMessage());
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testPositiveDNAccess() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testPositiveDNAccess";
        String description = "tests for allowed access based on DN.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);


        log("Adding DN to policy and removing attribute(s).");

        PolicyEntries policyEntries = new PolicyEntries(null, null);
        GlobusCredential defaultCredential =
                GlobusCredential.getDefaultCredential();
        String dn = defaultCredential.getIdentity();
        PolicyDNs policydns = null;
        if (dn != null) {
            log("Adding DN: " + dn);
            String[] dns = {dn};
            policydns = new PolicyDNs(dns);
        } else {
            log("FAIL: No DN.");
            assertTrue(false);
        }
        control.setPolicy(new Policy(policydns, policyEntries));

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {
            log("FAIL: exception while calling createResource(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");
    }

    public void testNegativeDNAccess() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testNegativeDNAccess";
        String description = "tests for denied access "
                    + "based on the\n wrong DN in the policy.  A valid "
                    + "assertion with the wrong\n attribute is returned from "
                    + "the test AA and authorization\n should fail "
                    + "overall.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);


        log("Adding wrong DN to policy and removing attribute(s).");

        PolicyEntries policyEntries = new PolicyEntries(null, null);
        GlobusCredential defaultCredential =
                GlobusCredential.getDefaultCredential();
        String dn = defaultCredential.getIdentity();

        PolicyDNs policydns;
        if (dn != null) {
            log("Correct DN: " + dn);
            String[] dns = {"/wrongDN="+dn+"/wrongDN"};
            policydns = new PolicyDNs(dns);
        } else {
            log("No correct DN?");
            String[] dns = {"/wrongDN=thisone"};
            policydns = new PolicyDNs(dns);
        }
        log("Wrong DN: " + policydns.getDN()[0]);
        control.setPolicy(new Policy(policydns, policyEntries));

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testAAerror() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testAAerror";
        String description = "tests for denied access because the "
                    + "AA had\n an error.  No assertion is returned from the "
                    + "test AA and\n (attribute based) authorization should "
                    + "fail.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);


        log("Configuring control to send an AA error");
        control.setAaError(true);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testBadResource() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testBadResource";
        String description = "tests for denied access because "
                    + "the AA\n returns an assertion with the wrong resource "
                    + "which the PIP should\n reject and authorization should "
                    + "fail.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);



        log("Configuring control to send wrong resource attribute");
        control.setAaMungeResource(true);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            } else {
                log(e.getMessage());
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testPositiveMultipleAttributeValueAccess() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testPositiveMultipleAttributeValueAccess";
        String description = "tests for allowed access\n" +
                    "      based on an attribute with multiple values. A valid assertion\n" +
                    "      with all of the attribute values is returned from the test AA and\n" +
                    "      authorization should succeed based on one of the values.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);



        log("Adding another value to the attribute in default policy.");

        PolicyEntries policyEntries = control.getPolicy().getPolicyEntries();
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        //GlobusCredential defaultCredential =
        //       GlobusCredential.getDefaultCredential();

        AttributeValueType[] attrvals = new AttributeValueType[2];
        attrvals[1] = loa.getAttribute(0).getAttributeValue(0);
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny("anothervalue"));
        loa.getAttribute(0).setAttributeValue(attrvals);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {
            log("FAIL: exception while calling createResource(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");
    }

    public void testNegativeMultipleAttributeValueAccess() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testNegativeMultipleAttributeValueAccess";
        String description = "tests for denied\n" +
                    "      access based on an attribute with multiple values. An assertion\n" +
                    "      with none of the matching attribute values is returned from the\n" +
                    "      test AA and authorization should fail.\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);


        log("Adding another value to the attribute in default policy.");

        PolicyEntries policyEntries = control.getPolicy().getPolicyEntries();
        //GlobusCredential defaultCredential =
        //      GlobusCredential.getDefaultCredential();

        AttributeValueType[] attrvals = new AttributeValueType[2];
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        attrvals[1] = loa.getAttribute(0).getAttributeValue(0);
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny("anothervalue"));
        loa.getAttribute(0).setAttributeValue(attrvals);

        control.setAaMungeAttributes(true);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            } else {
                log(e.getMessage());
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testPositiveMultipleAttributesAccess() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String testName = "testPositiveMultipleAttributesAccess";
        String description = "tests for allowed access\n" +
                    "      based on an assertions with multiple attributes. A valid assertion\n" +
                    "      with all of the attributes is returned from the test AA and\n" +
                    "      authorization should succeed based on one of the values.\n";

        lognobullet("TEST: \"" + testName + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(testName);
        control.setTestdescription(description);


        log("Adding another attribute to the default policy.");

        String value = "member2";
        String name = "urn:mace:dir:attribute-def:eduPersonAffiliation2";
        String namespace = "urn:mace:shibboleth:1.0:attributeNamespace:uri2";
        AttributeValueType[] attrvals = new AttributeValueType[1];
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny(value));
        AttributeType[] attrs = new AttributeType[2];
        attrs[0] = new AttributeType();
        attrs[0].setAttributeValue(attrvals);
        attrs[0].setAttributeName(name);
        attrs[0].setAttributeNamespace(new URI(namespace));

        PolicyEntries policyEntries = control.getPolicy().getPolicyEntries();
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        attrs[1] = loa.getAttribute(0);
        loa.setAttribute(attrs);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {
            log("FAIL: exception while calling createResource(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");
    }

    public void testNegativeMultipleAttributesAccess() throws Exception {
        assertTrue(this.testContainer != null);
        setDefaults();

        String testName = "testNegativeMultipleAttributesAccess";
        String description = "tests for denied\n" +
                    "      access based on an assertion with multiple attributes. An assertion\n" +
                    "      with none of the matching attribute values is returned from the\n" +
                    "      test AA and authorization should fail.\n";

        lognobullet("TEST: \"" + testName + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(testName);
        control.setTestdescription(description);


        log("Adding another attribute to the default policy.");

        String value = "member2";
        String name = "urn:mace:dir:attribute-def:eduPersonAffiliation2";
        String namespace = "urn:mace:shibboleth:1.0:attributeNamespace:uri2";
        AttributeValueType[] attrvals = new AttributeValueType[1];
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny(value));
        AttributeType[] attrs = new AttributeType[2];
        attrs[0] = new AttributeType();
        attrs[0].setAttributeValue(attrvals);
        attrs[0].setAttributeName(name);
        attrs[0].setAttributeNamespace(new URI(namespace));

        PolicyEntries policyEntries = control.getPolicy().getPolicyEntries();
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        attrs[1] = loa.getAttribute(0);
        policyEntries.setListOfAttributes(new ListOfAttributes(attrs));

        control.setAaMungeAttributes(true);

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.createResource(true);
        } catch (Exception e) {

            if (e.getMessage().matches(".*not authorized.*")) {
                log("PASS: exception with 'not authorized'");
                exp = true;
            } else {
                log(e.getMessage());
            }
        }
        if (!exp) {
            log("FAIL: Returned without proper exception.");
        }
        assertTrue(exp);
    }

    public void testUsernamePresence() throws Exception {

        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testUsernamePresence";
        String description = "tests for presence of system account mappings\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);


        PolicyEntries policyEntries = control.getPolicy().getPolicyEntries();
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        //GlobusCredential defaultCredential =
        //      GlobusCredential.getDefaultCredential();

        AttributeValueType[] attrvals = new AttributeValueType[2];
        attrvals[1] = loa.getAttribute(0).getAttributeValue(0);
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny("anothervalue"));
        loa.getAttribute(0).setAttributeValue(attrvals);
        ListOfUsernames lou =
            new ListOfUsernames(AttributeTestConstants.USERNAME_LIST);
        policyEntries.setListOfUsernames(lou);
        control.setPolicy(new Policy(null, policyEntries));

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.setValue(AttributeTestConstants.USERNAME_TEST_CODE);
        } catch (Exception e) {
            log("FAIL: exception while calling setValue(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");
    }

    public void testUsernameAbsence() throws Exception {

        assertTrue(this.testContainer != null);
        setDefaults();

        String name = "testUsernameAbsence";
        String description = "tests for absence of system account mappings\n";

        lognobullet("TEST: \"" + name + "\" " + description);

        ShibbolethTestPortType testPort =
            CompatibleUtil.getTestPort(testServiceAddress,authz);

        NoAuthRequest control =
                getDefaultNoAuth(attributeServiceUrl,attributeServiceIdentity);

        control.setTestname(name);
        control.setTestdescription(description);

        PolicyEntries policyEntries = control.getPolicy().getPolicyEntries();
        ListOfAttributes loa = policyEntries.getListOfAttributes();
        //GlobusCredential defaultCredential =
        //      GlobusCredential.getDefaultCredential();

        AttributeValueType[] attrvals = new AttributeValueType[2];
        attrvals[1] = loa.getAttribute(0).getAttributeValue(0);
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny("anothervalue"));
        loa.getAttribute(0).setAttributeValue(attrvals);
        control.setPolicy(new Policy(null, policyEntries));

        log("Calling the setup operation.");
        testPort.noAuth(control);
        log("Setup operation complete.");

        boolean exp = false;
        try {
            log("Calling the protected operation.");
            ((Stub)testPort)._setProperty(
                    Constants.GSI_TRANSPORT, Constants.SIGNATURE);
            testPort.setValue(-1);
        } catch (Exception e) {
            log("FAIL: exception while calling setValue(): "
                    + e.getMessage());
            exp = true;
        }
        assertFalse(exp);
        log("PASS: Returned without exception.");

    }

    /* ********************** UTILITIES ********************************* */

    private void setDefaults() {
        attributeServiceUrl = this.testContainer.getBaseURL() +
            AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY;
        attributeServiceIdentity = this.testContainer.getBaseURL()
            + AttributeTestConstants.TEST_ATTRIBUTE_AUTHORITY_IDENTITY;
        testServiceAddress = this.testContainer.getBaseURL() +
            AttributeTestConstants.SHIB_PROTECTED_SERVICE_PATH;
    }

    private static NoAuthRequest
        getDefaultNoAuth(String attrServiceUrl,
                         String attrServiceIdentity)
        throws Exception {

        log("Constructing default control object.");
        NoAuthRequest noauth = new NoAuthRequest();

        // find $GLOBUS_LOCATION
        String gl = System.getProperty("gridshib.globuslocation");

        // find $GL etc directory
        String servicedir = System.getProperty("gridshib.servicetestdir");

        if (gl == null || servicedir == null) {
            throw new Exception("Cannot resolve test service " +
                    "directory, parameters empty.");
        }

        String etcdir = gl + File.separator + "etc" +
                        File.separator + servicedir;

        File path = new File(etcdir);

        if (!path.exists()) {
            throw new Exception("Cannot resolve test service " +
                   "directory, path does not exist: '" + etcdir + "'");
        }

        log("Test service etc dir: '" + etcdir + "'");
        noauth.setEtcdir(etcdir);

        log("Attribute service URL " + attrServiceUrl);
        noauth.setAttributeService(attrServiceUrl);

        noauth.setAttributeServiceIdentity(attrServiceIdentity);

        noauth.setAaEmptyAssertion(false);
        noauth.setAaError(false);
        noauth.setAaMungeAssertion(false);
        noauth.setAaMungeAttributes(false);
        noauth.setAaValidityMinutes(20);
        noauth.setAaMungeIssuer(false);
        noauth.setAaIssueTimeOffset(0);

        noauth.setPdpCache(false);
        noauth.setPdpUseAttributes(true);
        noauth.setPdpUseDNs(false);

        noauth.setPipIdPProviderId(ShibbolethConstants.DEFAULT_IDP_PROVIDER_ID);
        noauth.setPipSPProviderId(ShibbolethConstants.DEFAULT_SP_PROVIDER_ID);

        log("IdP ProviderId: " + ShibbolethConstants.DEFAULT_IDP_PROVIDER_ID);
        log("SP  ProviderId: " + ShibbolethConstants.DEFAULT_SP_PROVIDER_ID);

        String value = "member";
        String name = "urn:mace:dir:attribute-def:eduPersonAffiliation";
        String namespace = "urn:mace:shibboleth:1.0:attributeNamespace:uri";

        AttributeValueType[] attrvals = new AttributeValueType[1];
        attrvals[0] = new AttributeValueType();
        attrvals[0].set_any(getAsAny(value));
        AttributeType[] attrs = new AttributeType[1];
        attrs[0] = new AttributeType();
        attrs[0].setAttributeValue(attrvals);
        attrs[0].setAttributeName(name);
        attrs[0].setAttributeNamespace(new URI(namespace));

//        log("Policy - one attribute:");
//        lognobullet("  Name: " + name);
//        lognobullet("  Namespace: " + namespace);
//        lognobullet("  Value: " + value);

        PolicyEntries policyEntries = new PolicyEntries();
        policyEntries.setListOfAttributes(new ListOfAttributes(attrs));
        String[] dns = {};
        PolicyDNs policydns = new PolicyDNs(dns);
        noauth.setPolicy(new Policy(policydns, policyEntries));

        return noauth;
    }

    // todo: spend time to get logger to work, right now, only stdout goes
    // to the test xml output
    private static void log(String logstring) {
        System.out.println("- " + logstring);
    }
    private static void lognobullet(String logstring) {
        System.out.println(logstring);
    }

    // Shibboleth hacks the SAML schema
    private static MessageElement[] getAsAny(String attr) {
        MessageElement[] mesEl = new MessageElement[1];
        mesEl[0] = new MessageElement(new Text(attr));
        return mesEl;
    }

}
